/* Implement classes and message passing for Objective C.
   Copyright (C) 1990 Free Software Foundation, Inc.
   Author: Steve Naroff.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
 *      Purpose: This module implements the Objective-C 4.0 language.
 *
 *      compatibility issues (with the Stepstone translator):
 *
 *	- does not recognize the following 3.3 constructs.
 *	  @requires, @classes, @messages, = (...)
 *	- methods with variable arguments must conform to ANSI standard.
 *	- tagged structure definitions that appear in BOTH the interface
 *	  and implementation are not allowed.
 *      - public/private: all instance variables are public within the
 *        context of the implementation...I consider this to be a bug in
 *        the translator.
 *      - statically allocated objects are not supported. the user will
 *        receive an error if this service is requested.
 *
 *      code generation `options':
 *
 *      - OBJC_SELS_R_INTS, OBJC_SELS_R_STRUCT_PTRS
 */

#include <stdio.h>
#include "config.h"
#include "tree.h"
#include "flags.h"
#include "input.h"
#define NULLT	(tree) 0

#ifdef OBJCPLUS

#include "cplus-parse.h"
#include "cplus-tree.h"
extern tree make_anon_name ();

/* Hacks to simulate start_struct() and finish_struct(). */

static int cplus_struct_hack = 0;

tree objcplus_start_struct (enum tree_code code, tree name) 
{ 
  tree s = xref_tag (record_type_node, name ? name : make_anon_name (), 0);
  
  /* simulate `LC' production */
  int temp = allocation_temporary_p ();
  int momentary = suspend_momentary ();

  if (temp)
    end_temporary_allocation ();
  cplus_struct_hack = (momentary << 1) | temp;
  pushclass (s, 0); 

  return s;	
}

static tree objcplus_finish_struct (tree t, tree fieldlist)
{
  tree s = finish_struct (t,  
	      build_tree_list ((tree) visibility_default, fieldlist), 
	      1, 0 /* dont warn for untagged struct templates */);

  if (cplus_struct_hack & 1)
    resume_temporary_allocation ();
  if (cplus_struct_hack & 2)
    resume_momentary (1);

  return s;
}

/* Hacks to simulate push_parm_decl() and objcplus_get_parm_info(). */

static tree objcplus_parmlist = NULLT;

tree objcplus_push_parm_decl (tree parm) 
{
  if (objcplus_parmlist)
    objcplus_parmlist = chainon (objcplus_parmlist, build_tree_list (0, parm));
  else
    objcplus_parmlist = build_tree_list (0, parm);

  return objcplus_parmlist;
}

static tree objcplus_get_parm_info (int void_at_end) 
{
  tree parm_info = objcplus_parmlist;
  
  TREE_PARMLIST (parm_info) = 1;

  if (void_at_end)
    chainon (parm_info, void_list_node);

  objcplus_parmlist = NULLT;

  return parm_info;
}

static tree objcplus_type_name (tree type)
{
  if (TYPE_NAME (type) && TREE_CODE (TYPE_NAME (type)) == TYPE_DECL)
    return DECL_NAME (TYPE_NAME (type));
  else
    return TYPE_NAME (type);
}

static tree objcplus_type_size (tree type)
{
  int x = TREE_INT_CST_LOW (TYPE_SIZE (type)) * TYPE_SIZE_UNIT (type);
  
  return build_int_2 (x, 0);
}

/* Macros to cover functions with changed interfaces. */

#define start_struct(code, name) objcplus_start_struct (code, name)

#define finish_struct(code, name) objcplus_finish_struct (code, name)

#define start_function(declspecs, declarator, nested) \
  start_function (declspecs, declarator, NULLT, 0)

#define xref_tag(code, name) xref_tag (record_type_node, name, 0)

#define pushlevel(tag_transparent)

#define poplevel(keep, reverse, functionbody)

#define push_parm_decl(parm) objcplus_push_parm_decl (parm)

#define get_parm_info(void_at_end) objcplus_get_parm_info (void_at_end)

#define grokfield(filename, line, declarator, declspecs, width) \
  grokfield (declarator, declspecs, width, 0, 0)

#define build_component_ref(datum, component) \
  build_component_ref (datum, component, NULLT, 1)

#define comptypes(type1, type2) comptypes (type1, type2, 0)

#define start_decl(declarator, declspecs, spspecs) \
  start_decl (declarator, declspecs, spspecs, NULLT)

#define TREE_CONSTANT(NODE) TREE_READONLY (NODE)

#define TYPE_VOLATILE(NODE) TREE_VOLATILE (NODE)

#undef TYPE_NAME
#define TYPE_NAME(NODE) objcplus_type_name (NODE)

#undef TYPE_SIZE
#define TYPE_SIZE(NODE) objcplus_type_size (NODE)

#else /* OBJCPLUS */

#include "c-parse.h"
#include "c-tree.h"

#endif /* OBJCPLUS */

#include "objc-actions.h"

int doing_objc_thang;

/* Define the special tree codes that we use.  */

/* Table indexed by tree code giving a string containing a character
   classifying the tree code.  Possibilities are
   t, d, s, c, r, <, 1 and 2.  See objc-tree.def for details.  */

#define DEFTREECODE(SYM, NAME, TYPE, LENGTH) TYPE,

char *objc_tree_code_type[] = {
  "x",
#include "objc-tree.def"
};
#undef DEFTREECODE

/* Table indexed by tree code giving number of expression
   operands beyond the fixed part of the node structure.
   Not used for types or decls.  */

#define DEFTREECODE(SYM, NAME, TYPE, LENGTH) LENGTH,

int objc_tree_code_length[] = {
  0,
#include "objc-tree.def"
};
#undef DEFTREECODE

/* Names of tree components.
   Used for printing out the tree and error messages.  */
#define DEFTREECODE(SYM, NAME, TYPE, LEN) NAME,

char *objc_tree_code_name[] = {
  "@@dummy",
#include "objc-tree.def"
};
#undef DEFTREECODE

/* for encode_method_def */
#include "rtl.h"

#define OBJC_VERSION	5
#define PROTOCOL_VERSION 2

#define OBJC_SELS_R_STRUCT_PTRS

#define OBJC_ENCODE_INLINE_DEFS 	0
#define OBJC_ENCODE_DONT_INLINE_DEFS	1

#define OBJC_CLASS_REFS

/*** Private Interface (procedures) ***/

/* code generation */

static void synth_module_prologue ();
static void build_module_descriptor ();
static tree   init_module_descriptor ();
static void build_module_entry ();
static void generate_strings ();
static void build_selector_translation_table ();
static tree build_ivar_chain ();

static tree build_ivar_template ();
static tree build_method_template ();
static tree build_private_template ();
static void build_class_template ();
static void build_category_template ();
static tree build_super_template ();
static void build_class_ref_template ();
static tree build_category_initializer ();
static tree build_protocol_initializer ();

static void synth_forward_declarations ();
static void generate_ivar_lists ();
static void generate_dispatch_tables ();
static void generate_shared_structures ();
static tree generate_protocol_list ();
static void generate_forward_declaration_to_string_table ();
static void build_protocol_reference ();

static tree init_selector ();
static tree build_keword_selector ();
static tree synth_id_with_class_suffix ();

/* misc. bookkeeping */

typedef struct hashedEntry 	*hash;
typedef struct hashedAttribute  *attr;

struct hashedAttribute {
        attr    next;
        tree    value;
};
struct hashedEntry {
        attr    list;
	hash	next;
	tree 	key;
};
static void hash_init ();
static void hash_enter ();
static hash hash_lookup ();
static void hash_add_attr ();
static tree lookup_method ();
static tree lookup_instance_method_static ();
static tree lookup_class_method_static ();
static tree add_class ();
static void add_class_reference ();
enum stringSection {
	class_names,		/* class, category, protocol, module names */
	meth_var_names,		/* method and variable names */
	meth_var_types,		/* method and variable type descriptors */
};
static tree add_objc_string ();
static tree build_objc_string_decl ();
static tree build_selector_reference_decl ();

/* protocol additions */

static tree add_protocol ();
static tree lookup_protocol ();
static void lookup_and_install_protocols ();

/* type encoding */

static void encode_aggregate ();
static void encode_bitfield ();
static void encode_type ();
static void encode_field_decl ();
static void encode_type_qualifiers ();

static void really_start_method ();
static int  comp_method_with_proto ();
static int  comp_proto_with_proto ();
static tree get_arg_type_list ();
static tree expr_last ();

/* utilities for debugging and error diagnostics: */

static void warn_with_method ();
static void error_with_method ();
static void error_with_ivar ();
static char *gen_method_decl ();
static char *gen_declaration ();
static char *gen_declarator ();
static int is_complex_decl ();
static void adorn_decl ();
static void dump_interfaces ();

/*** Private Interface (data) ***/

/* reserved tag definitions: */

#define TYPE_ID			"id"
#define TAG_OBJECT		"objc_object"
#define TAG_CLASS		"objc_class"
#define TAG_SUPER		"objc_super"
#define TAG_SELECTOR		"objc_selector"

#define _TAG_CLASS		"_objc_class"
#define _TAG_IVAR		"_objc_ivar"
#define _TAG_IVAR_LIST		"_objc_ivar_list"
#define _TAG_METHOD		"_objc_method"
#define _TAG_METHOD_LIST	"_objc_method_list"
#define _TAG_CATEGORY		"_objc_category"
#define _TAG_MODULE		"_objc_module"
#define _TAG_SYMTAB		"_objc_symtab"
#define _TAG_SUPER		"_objc_super"

#define _TAG_PROTOCOL		     "_objc_protocol"
#define _TAG_PROTOCOL_LIST	     "_objc_protocol_list"
#define _TAG_METHOD_PROTOTYPE        "_objc_method_prototype"
#define _TAG_METHOD_PROTOTYPE_LIST   "_objc_method_prototype_list"

#define STRING_OBJECT_CLASS_NAME	"NXConstantString"
#define PROTOCOL_OBJECT_CLASS_NAME	"Protocol"

/* set by `continue_class ()' and checked by `is_public ()' */

#define TREE_STATIC_TEMPLATE(record_type) (TREE_PUBLIC(record_type))
#define TYPED_OBJECT(type) \
       (TREE_CODE (type) == RECORD_TYPE && TREE_STATIC_TEMPLATE (type))

/* some commonly used instances of "identifier_node". */

static tree self_id, _cmd_id, _msg_id, _msgSuper_id;
static tree objc_getClass_id, objc_getMetaClass_id;

static tree self_decl, _msg_decl, _msgSuper_decl;
static tree objc_getClass_decl, objc_getMetaClass_decl;

static tree super_type, _selector_type, id_type, objc_class_type;
static tree instance_type, protocol_type;

static tree class_chain = NULLT;
static tree alias_chain = NULLT;
static tree interface_chain = NULLT;
static tree protocol_chain = NULLT;

/* chains to manage selectors that are referenced and defined in the module */

static tree cls_ref_chain = NULLT;	/* classes referenced */
static tree sel_ref_chain = NULLT;	/* selectors referenced */

/* chains to manage uniquing of strings */

static tree class_names_chain = NULLT;
static tree meth_var_names_chain = NULLT;
static tree meth_var_types_chain = NULLT;

/* hash tables to manage the global pool of method prototypes */

static hash *nst_method_hash_list = 0;
static hash *cls_method_hash_list = 0;

/* the following are used when compiling a class implementation.
 *
 * implementation_template will normally be an anInterface, however if
 * none exists this will be equal to implementation_context...it is
 * set in start_class.
 */

/* backend data declarations */

static tree _OBJC_SYMBOLS_decl;
static tree 	_OBJC_INSTANCE_VARIABLES_decl, _OBJC_CLASS_VARIABLES_decl;
static tree 	_OBJC_INSTANCE_METHODS_decl, _OBJC_CLASS_METHODS_decl;
static tree 	_OBJC_CLASS_decl, _OBJC_METACLASS_decl;
static tree _OBJC_MODULES_decl;
static tree _OBJC_STRINGS_decl;

static tree implementation_context = NULLT,
	    implementation_template = NULLT;

struct imp_entry {
	struct imp_entry *next;
	tree imp_context;
	tree imp_template;
	tree class_decl; 	/* _OBJC_CLASS_<my_name>; */
	tree meta_decl;  	/* _OBJC_METACLASS_<my_name>; */
};
static struct imp_entry *imp_list = 0;
static int imp_count = 0;	/* `@implementation' */
static int cat_count = 0;	/* `@category' */

static tree objc_class_template, objc_category_template, _PRIVATE_record;
static tree objc_protocol_template;
#ifdef OBJC_CLASS_REFS
static tree objc_class_ref_template;
#endif /* OBJC_CLASS_REFS */
static tree _clsSuper_ref, __clsSuper_ref;

static tree objc_method_template, objc_ivar_template;
static tree objc_symtab_template, objc_module_template;
static tree objc_super_template, objc_object_reference;

 static tree objc_object_id, objc_class_id, objc_id_id;
static tree constant_string_id;
static tree constant_string_type;
static tree _OBJC_SUPER_decl;

static tree method_context = NULLT;
static int  method_slot = 0;	/* used by start_method_def */

#define BUFSIZE		1024

static char *errbuf;	/* a buffer for error diagnostics */
static char *utlbuf;	/* a buffer for general utility */

/* services imported from "varasm.c" */

extern void assemble_asm ();

/* services imported from "final.c" */

extern void app_disable ();

extern char *strcpy (),*strcat ();

extern tree groktypename_in_parm_context ();

extern struct obstack permanent_obstack, *current_obstack,  *rtl_obstack;

/* data imported from toplev.c  */

extern char *dump_base_name;

/* Open and close the file for outputting class declarations, if requested.  */

int flag_gen_declaration = 0;

FILE *gen_declaration_file;

/* Warn if multiple methods are seen for the same selector, but with
   different argument types. */

int warn_selector = 0;

void
lang_init ()
{
  /* the beginning of the file is a new line; check for # */
  /* With luck, we discover the real source file's name from that
     and put it in input_filename.  */
  ungetc (check_newline (), finput);

  /* If gen_declaration desired, open the output file.  */
  if (flag_gen_declaration)
    {
      int dump_base_name_length = strlen (dump_base_name);
      register char *dumpname = (char *) xmalloc (dump_base_name_length + 7);
      strcpy (dumpname, dump_base_name);
      strcat (dumpname, ".decl");
      gen_declaration_file = fopen (dumpname, "w");
      if (gen_declaration_file == 0)
	pfatal_with_name (dumpname);
    }

  if (doing_objc_thang)
    init_objc ();
}

static void
objc_fatal ()
{
#ifdef OBJCPLUS
  fatal ("Objective-C text in C++ source file: use -ObjC++");
#else /* OBJCPLUS */
  fatal ("Objective-C text in C++ source file: use -ObjC");
#endif /* OBJCPLUS */
}

void
objc_finish ()
{
  if (doing_objc_thang)
    finish_objc ();		/* Objective-C finalization */

  if (gen_declaration_file)
    fclose (gen_declaration_file);
}

void
lang_finish ()
{
}

int
lang_decode_option (p)
     char *p;
{
  if (!strcmp (p, "-lang-objc"))
    doing_objc_thang = 1;
  else if (!strcmp (p, "-gen-decls"))
    flag_gen_declaration = 1;
  else if (!strcmp (p, "-Wselector"))
    warn_selector = 1;
  else
#ifdef OBJCPLUS
    return cplus_decode_option (p);
#else /* OBJCPLUS */
    return c_decode_option (p);
#endif /* OBJCPLUS */

  return 1;
}

static tree
define_decl (declarator, declspecs)
     tree declarator;
     tree declspecs;
{
  tree decl = start_decl (declarator, declspecs, 0);
  finish_decl (decl, NULLT, NULLT);
  return decl;
}

/*
 * rules for statically typed objects...called from `c-typeck.comptypes'.
 *
 * an assignment of the form `a' = `b' is permitted if:
 *
 *   - `a' is of type "id".
 *   - `a' and `b' are the same class type.
 *   - `a' and `b' are of class types A and B such that B is a descendant
 *     of A.
 */

int
maybe_objc_comptypes (lhs, rhs)
     tree lhs, rhs;
{
  if (doing_objc_thang)
    return objc_comptypes (lhs, rhs);
  return 0;
}

static tree
lookup_method_in_protocol_list (rproto_list, sel_name, class_meth)
   tree rproto_list;
   tree sel_name;
   int class_meth;
{
   tree rproto, p;
   tree fnd = 0;
    
   for (rproto = rproto_list; rproto; rproto = TREE_CHAIN(rproto))
     {
        p = TREE_VALUE(rproto);
	  
	if (TREE_CODE(p) == PROTOCOL_INTERFACE)
	  {
	  if (fnd = lookup_method (class_meth ? PROTOCOL_CLS_METHODS (p) : PROTOCOL_NST_METHODS (p), sel_name))
            ;    
	  else if (PROTOCOL_LIST(rproto))
	    fnd = lookup_method_in_protocol_list(PROTOCOL_LIST(p), sel_name, class_meth);
	  }
	else 
	  ; /* an identifier...if we could not find a protocol.  */
	  
	if (fnd)
	  return fnd;
     }
   return 0;
}

static tree
lookup_protocol_in_reflist (rproto_list, lproto)
   tree rproto_list;
   tree lproto;
{
   tree rproto, p;
    
   /* make sure the protocol is support by the object on the rhs */
   if (TREE_CODE(lproto) == PROTOCOL_INTERFACE)
     {
     tree fnd = 0;
     for (rproto = rproto_list; rproto; rproto = TREE_CHAIN(rproto))
       {
          p = TREE_VALUE(rproto);
	  
	  if (TREE_CODE(p) == PROTOCOL_INTERFACE)
	    {
	    if (lproto == p)
	      fnd = lproto;
	    
	    else if (PROTOCOL_LIST(rproto))
	      fnd = lookup_protocol_in_reflist(PROTOCOL_LIST(p), lproto);
	    }
	  
	  if (fnd)
	    return fnd;
       }
     }
   else 
     ; /* an identifier...if we could not find a protocol. */

   return 0;
}

int
objc_comptypes (lhs, rhs)
     tree lhs;
     tree rhs;
{
  /* new clause for protocols */
  
  if ((TREE_CODE (lhs) == POINTER_TYPE
       && TREE_CODE (TREE_TYPE (lhs)) == RECORD_TYPE)
      && (TREE_CODE (rhs) == POINTER_TYPE
       && TREE_CODE (TREE_TYPE (rhs)) == RECORD_TYPE))
    {
      int lhs_is_proto = (TYPE_MAIN_VARIANT (lhs) == id_type
			  && lhs != id_type);
      int rhs_is_proto = (TYPE_MAIN_VARIANT (rhs) == id_type
			  && rhs != id_type);

      if (lhs_is_proto)
        {
         tree lproto, lproto_list = TYPE_PROTOCOL_LIST (lhs);
         tree rproto, rproto_list;
         tree p;
      
         if (rhs_is_proto)
           {
             rproto_list = TYPE_PROTOCOL_LIST (rhs);

             /* make sure the protocol is supported by the object on the rhs */
             for (lproto = lproto_list; lproto; lproto = TREE_CHAIN(lproto))
	       {
                 p = TREE_VALUE(lproto);
	         rproto = lookup_protocol_in_reflist(rproto_list, p); 
	  
	         if (!rproto)
                   warning("object does not conform to the `%s' protocol",
	  	           IDENTIFIER_POINTER(PROTOCOL_NAME(p)));
	       }
	   }
         else if (TYPED_OBJECT (TREE_TYPE(rhs)))
           {
             tree rname = TYPE_NAME (TREE_TYPE(rhs));
	     tree rinter;

             /* make sure the protocol is supported by the object on the rhs */
             for (lproto = lproto_list; lproto; lproto = TREE_CHAIN(lproto))
	       {
                 p = TREE_VALUE(lproto);
	         rproto = 0;
	         rinter = lookup_interface (rname);
	  
	         while (rinter && !rproto)
	           {
		   tree cat;

	           rproto_list = CLASS_PROTOCOL_LIST(rinter);
	           rproto = lookup_protocol_in_reflist(rproto_list, p);

/* NEW!!! */
		   /* Check for protocols adopted by categories. */
		   cat = CLASS_CATEGORY_LIST (rinter);
		   while (cat && !rproto)
		     {
		       rproto_list = CLASS_PROTOCOL_LIST (cat);
		       rproto = lookup_protocol_in_reflist (rproto_list, p);

		       cat = CLASS_CATEGORY_LIST (cat);
		     }

	           rinter = lookup_interface (CLASS_SUPER_NAME (rinter));
	           }
	         if (!rproto)
	           warning("class `%s' does not implement the `%s' protocol",
	                     IDENTIFIER_POINTER(TYPE_NAME (TREE_TYPE(rhs))),
		             IDENTIFIER_POINTER(PROTOCOL_NAME(p)));
               }
	   }
	
          return 1; /* may change...based on whether there was any mismatch */
        }
      else if (rhs_is_proto)
        {
	/* lhs is not a protocol...warn if it is statically typed */
	
	if (TYPED_OBJECT (TREE_TYPE(lhs)))
	  return 0;
	else
	  return 1;	/* one of the types is a protocol */
	}
      else
	return 2;	/* defer to comptypes */      
    }
  else if (TREE_CODE (lhs) == RECORD_TYPE && TREE_CODE (rhs) == RECORD_TYPE)
    ; /* fall thru...this is the case we have been handling all along */
  else
    return 2;	/* defer to comptypes */

  /* End of new protocol support.  */

  /* `id' = `<class> *', `<class> *' = `id' */

  if ((TYPE_NAME (lhs) == objc_object_id && TYPED_OBJECT (rhs))
      || (TYPED_OBJECT (lhs) && TYPE_NAME (rhs) == objc_object_id))
    return 1;

  /* `id' = `Class', `Class' = `id' */

  else if ((TYPE_NAME (lhs) == objc_object_id &&
	    TYPE_NAME (rhs) == objc_class_id) ||
	   (TYPE_NAME (lhs) == objc_class_id &&
	    TYPE_NAME (rhs) == objc_object_id))
    return 1;

  /* `<class> *' = `<class> *' */

  else if (TYPED_OBJECT (lhs) && TYPED_OBJECT (rhs))
    {
      tree lname = TYPE_NAME (lhs), rname = TYPE_NAME (rhs);

      if (lname == rname)
	return 1;
      else
	{
	  /* if the left hand side is a super class of the right hand side,
	     allow it...
	     */
	  tree rinter = lookup_interface (rname);

	  while (rinter)
	    {
	      if (lname == CLASS_SUPER_NAME (rinter))
		return 1;

	      rinter = lookup_interface (CLASS_SUPER_NAME (rinter));
	    }

	  return 0;
	}
    }
  else
    return 0;
}

/* Called from c-decl.c before all calls to rest_of_decl_compilation.  */

void
maybe_objc_check_decl (decl)
     tree decl;
{
  if (doing_objc_thang)
    objc_check_decl (decl);
}

void
objc_check_decl (decl)
     tree decl;
{
  tree type = TREE_TYPE (decl);
  static int alreadyWarned = 0;

  if (TREE_CODE (type) == RECORD_TYPE && TREE_STATIC_TEMPLATE (type)
      && type != constant_string_type)
    {
      if (!alreadyWarned)
	{
	  error ("GNU compiler does not support statically allocated objects");
	  alreadyWarned = 1;
	}
      error_with_decl (decl, "`%s' cannot be statically allocated");
    }
}

/* implement static typing. at this point, we know we have an interface... */

tree
get_static_reference (ident, protocols)
     tree ident;
     tree protocols;
{
     tree type = xref_tag (RECORD_TYPE, ident);

     if (protocols)
       {
	  tree proto, t, m = TYPE_MAIN_VARIANT (type);
	  struct obstack *ambient_obstack = current_obstack;
	  
	  current_obstack = &permanent_obstack;
	  t = copy_node (type);
	  
	  /* Add this type to the chain of variants of TYPE.  */
	  TYPE_NEXT_VARIANT (t) = TYPE_NEXT_VARIANT (m);
	  TYPE_NEXT_VARIANT (m) = t;
	  
	  current_obstack = ambient_obstack;
  
	  /* look up protocols...and install in lang specific list */
	  
	  TYPE_PROTOCOL_LIST (t) = protocols;
	  lookup_and_install_protocols (protocols);
	    
	  /* this forces a new pointer type to be created later 
	   * (in build_pointer_type())...so that the new template
	   * we just created will actually be used...what a hack!
	   */
	  if (TYPE_POINTER_TO (t))
	    TYPE_POINTER_TO(t) = NULL;

	  type = t;
       }

     return type;
}

tree
get_object_reference (protocols)
     tree protocols;
{
     tree type_decl = lookup_name (objc_id_id);
     tree type;
     
     if (type_decl && (TREE_CODE(type_decl) == TYPE_DECL))
       {
       type = TREE_TYPE (type_decl); 
       if (TYPE_MAIN_VARIANT(type) != id_type)
         warning("Unexpected type for `id' (%s)", 
			gen_declaration(type,errbuf));
       }
     else
       {
       error ("Undefined type `id', please import <objc/objc.h>");
       exit(1);
       }

     /* this clause creates a new pointer type that is qualified with
        the protocol specification...this info is used later to do more
	elaborate type checking.
      */
     if (protocols)
	{
	  tree proto, t, m = TYPE_MAIN_VARIANT (type);
	  struct obstack *ambient_obstack = current_obstack;
	  
	  current_obstack = &permanent_obstack;
	  t = copy_node (type);
	  
	  /* Add this type to the chain of variants of TYPE.  */
	  TYPE_NEXT_VARIANT (t) = TYPE_NEXT_VARIANT (m);
	  TYPE_NEXT_VARIANT (m) = t;
	  
	  current_obstack = ambient_obstack;
  
	  /* look up protocols...and install in lang specific list */
	  
	  TYPE_PROTOCOL_LIST (t) = protocols;
	  lookup_and_install_protocols (protocols);
	    
	  /* this forces a new pointer type to be created later 
	   * (in build_pointer_type())...so that the new template
	   * we just created will actually be used...what a hack!
	   */
	  if (TYPE_POINTER_TO (t))
	    TYPE_POINTER_TO(t) = NULL;

	  type = t;
	}
     return type;
}

static void 
lookup_and_install_protocols (protocols)
     tree protocols;
{ 
     tree proto;
       
     for (proto = protocols; proto; proto = TREE_CHAIN (proto))
       {
	  tree ident = TREE_VALUE (proto);
	  tree p = lookup_protocol (ident);
  
	  if (!p)
	    error ("Cannot find protocol declaration for `%s'",
		   IDENTIFIER_POINTER (ident));
	  else /* replace identifier with actual protocol node */
	    TREE_VALUE(proto) = p;
       }
}

/*
 *	purpose: "play" parser, creating/installing representations
 *		 of the declarations that are required by Objective-C.
 *
 *	model:
 *
 *		type_spec--------->sc_spec
 *		(tree_list)        (tree_list)
 *		    |                  |
 *		    |                  |
 *		identifier_node    identifier_node
 */
static void
synth_module_prologue ()
{
  tree sc_spec, type_spec, decl_specs, expr_decl, parms, record;

#ifdef OBJCPLUS
  push_lang_context (lang_name_c);
#endif /* OBJCPLUS */

  /* defined in `objc.h' */
  objc_object_id = get_identifier (TAG_OBJECT);

  objc_object_reference = xref_tag (RECORD_TYPE, objc_object_id);

  id_type = groktypename (build_tree_list (
					   build_tree_list (NULLT, objc_object_reference),
					   build1 (INDIRECT_REF, NULLT, NULLT)));

  objc_id_id = get_identifier(TYPE_ID);
  objc_class_id = get_identifier (TAG_CLASS);
  
  objc_class_type = groktypename (build_tree_list
			     (build_tree_list
			      (NULLT, xref_tag (RECORD_TYPE, objc_class_id)),
			      build1 (INDIRECT_REF, NULLT, NULLT)));

  protocol_type = groktypename (build_tree_list
			    (build_tree_list 
                             (NULLT, xref_tag (RECORD_TYPE,
			  	get_identifier (PROTOCOL_OBJECT_CLASS_NAME))), 
			        build1 (INDIRECT_REF, NULLT, NULLT)));

/* Declare SEL type before prototypes for objc_msgSend(), or else those
   struct tags are considered local to the prototype and won't match the one
   in <objc/objc-runtime.h>. */

#ifdef OBJC_SELS_R_INTS
  /* `unsigned int' */
  _selector_type = unsigned_type_node;
#endif

#ifdef OBJC_SELS_R_STRUCT_PTRS
  /* `struct objc_selector *' */
  _selector_type = groktypename (build_tree_list (
	     build_tree_list (NULLT,
			      xref_tag (RECORD_TYPE,
					get_identifier (TAG_SELECTOR))),
	     build1 (INDIRECT_REF, NULLT, NULLT)));
#endif

  /* forward declare type...or else the prototype for `super' will bitch */
  groktypename (build_tree_list (build_tree_list (NULLT,
						  xref_tag (RECORD_TYPE, get_identifier (TAG_SUPER))),
				 build1 (INDIRECT_REF, NULLT, NULLT)));

  _msg_id = get_identifier ("objc_msgSend");
  _msgSuper_id = get_identifier ("objc_msgSendSuper");
  objc_getClass_id = get_identifier ("objc_getClass");
  objc_getMetaClass_id = get_identifier ("objc_getMetaClass");

  /* struct objc_object *objc_msgSend (id, SEL, ...); */
  pushlevel (0);
  decl_specs = build_tree_list (NULLT, objc_object_reference);
  push_parm_decl (build_tree_list (decl_specs,
				   build1 (INDIRECT_REF, NULLT, NULLT)));

#ifdef OBJC_SELS_R_INTS
  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_UNSIGNED]);
  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_INT], decl_specs);
  expr_decl = NULLT;
#endif

#ifdef OBJC_SELS_R_STRUCT_PTRS
  decl_specs = build_tree_list (NULLT,
				xref_tag (RECORD_TYPE,
					  get_identifier (TAG_SELECTOR)));
  expr_decl = build1 (INDIRECT_REF, NULLT, NULLT);
#endif

  push_parm_decl (build_tree_list (decl_specs, expr_decl));
  parms = get_parm_info (0);
  poplevel (0, 0, 0);

  decl_specs = build_tree_list (NULLT, objc_object_reference);
  expr_decl = build_nt (CALL_EXPR, _msg_id, parms, NULLT);
  expr_decl = build1 (INDIRECT_REF, NULLT, expr_decl);

  _msg_decl = define_decl (expr_decl, decl_specs);

  /* struct objc_object *objc_msgSendSuper (struct objc_super *, SEL, ...); */
  pushlevel (0);
  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
						 get_identifier (TAG_SUPER)));
  push_parm_decl (build_tree_list (decl_specs,
				   build1 (INDIRECT_REF, NULLT, NULLT)));

#ifdef OBJC_SELS_R_INTS
  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_UNSIGNED]);
  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_INT], decl_specs);
  expr_decl = NULLT;
#endif

#ifdef OBJC_SELS_R_STRUCT_PTRS
  decl_specs = build_tree_list (NULLT,
				xref_tag (RECORD_TYPE,
					  get_identifier (TAG_SELECTOR)));
  expr_decl = build1 (INDIRECT_REF, NULLT, NULLT);
#endif

  push_parm_decl (build_tree_list (decl_specs, expr_decl));
  parms = get_parm_info (0);
  poplevel (0, 0, 0);

  decl_specs = build_tree_list (NULLT, objc_object_reference);
  expr_decl = build_nt (CALL_EXPR, _msgSuper_id, parms, NULLT);
  expr_decl = build1 (INDIRECT_REF, NULLT, expr_decl);

  _msgSuper_decl = define_decl (expr_decl, decl_specs);

  /* id objc_getClass (const char *); */
  pushlevel (0);
  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_CONST]);
  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_CHAR], decl_specs);
  push_parm_decl (build_tree_list (decl_specs,
				   build1 (INDIRECT_REF, NULLT, NULLT)));
  parms = get_parm_info (1);	
  poplevel (0, 0, 0);
  decl_specs = build_tree_list (NULLT, objc_object_reference);
  expr_decl = build_nt (CALL_EXPR, objc_getClass_id, parms, NULLT);
  expr_decl = build1 (INDIRECT_REF, NULLT, expr_decl);

  objc_getClass_decl = define_decl (expr_decl, decl_specs);

  /* id objc_getMetaClass (const char *); */
  pushlevel (0);
  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_CONST]);
  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_CHAR], decl_specs);
  push_parm_decl (build_tree_list (decl_specs, 
			 	   build1 (INDIRECT_REF, NULLT, NULLT)));
  parms = get_parm_info (1);	
  poplevel (0, 0, 0);
  decl_specs = build_tree_list (NULLT, objc_object_reference);
  expr_decl = build_nt (CALL_EXPR, objc_getMetaClass_id, parms, NULLT);
  expr_decl = build1 (INDIRECT_REF, NULLT, expr_decl);

  objc_getMetaClass_decl = define_decl (expr_decl, decl_specs);

  generate_forward_declaration_to_string_table ();

/* Forward declare constant_string_id and constant_string_type.  */
  constant_string_id = get_identifier (STRING_OBJECT_CLASS_NAME);
  constant_string_type = xref_tag (RECORD_TYPE, constant_string_id);

#ifdef OBJC_CLASS_REFS
  build_class_ref_template ();
#endif /* OBJC_CLASS_REFS */

#ifdef OBJCPLUS
  pop_lang_context ();
#endif /* OBJCPLUS */
}

/*
 * custom "build_string ()" which sets TREE_TYPE!
 */
static tree
my_build_string (len, str)
     int len;
     char *str;
{
  int wide_flag = 0;
  tree aString = build_string (len, str);
  /*
   *  some code from "combine_strings ()", which is local to c-parse.y.
   */
  if (TREE_TYPE (aString) == int_array_type_node)
    wide_flag = 1;

  TREE_TYPE (aString) =
    build_array_type (wide_flag ? integer_type_node : char_type_node,
		      build_index_type (build_int_2 (len - 1, 0)));

  TREE_CONSTANT (aString) = 1;	/* puts string in the ".text" segment */
  TREE_STATIC (aString) = 1;

  return aString;
}

/* Return a newly constructed OBJC_STRING_CST node whose value is
   the LEN characters at STR.
   The TREE_TYPE is not initialized.  */

tree
build_objc_string (len, str)
     int len;
     char *str;
{
  tree s = build_string (len, str);
  
  TREE_CODE (s) = OBJC_STRING_CST;
  return s;
}

/* Given a chain of OBJC_STRING_CST's, build a static instance of
   NXConstantString which points at the concatenation of those strings.
   We place the string object in the __string_objects section of the
   __OBJC segment.  The Objective-C runtime will initialize the isa
   pointers of the string objects to point at the NXConstantString
   class object.  */

tree
build_objc_string_object (strings)
     tree strings;
{
  tree string, initlist, constructor;
  int length;
  
  if (!doing_objc_thang)
    objc_fatal ();

  if (lookup_interface (constant_string_id) == NULLT)
    {
      error ("Cannot find interface declaration for `%s'",
	       IDENTIFIER_POINTER (constant_string_id));
      return error_mark_node;
    }

  add_class_reference (constant_string_id);

  /* combine_strings will work for OBJC_STRING_CST's too.  */
  string = combine_strings (strings);
  TREE_CODE (string) = STRING_CST;
  length = TREE_STRING_LENGTH (string) - 1;

  /* & ((NXConstantString) {0, string, length})  */

  initlist = build_tree_list (NULLT, build_int_2 (0, 0));
  initlist = tree_cons (NULLT, build_unary_op (ADDR_EXPR, string, 1),
			initlist);
  initlist = tree_cons (NULLT, build_int_2 (length, 0), initlist);
  constructor = build (CONSTRUCTOR, constant_string_type, NULLT,
		       nreverse (initlist));
  TREE_CONSTANT (constructor) = 1;
  TREE_STATIC (constructor) = 1;
  TREE_READONLY (constructor) = 1;

  return build_unary_op (ADDR_EXPR, constructor, 1);
}

/*
 *	struct objc_symtab {
 *		long sel_ref_cnt;
 *		char *refs;
 *		long cls_def_cnt;
 *		long cat_def_cnt;
 *		void *defs[cls_def_cnt + cat_def_cnt];
 *	};
 */
static void
build_objc_symtab_template ()
{
  tree decl_specs, field_decl, field_decl_chain;

  objc_symtab_template = start_struct (RECORD_TYPE, get_identifier (_TAG_SYMTAB));

  /* long sel_ref_cnt; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_LONG]);
  field_decl = get_identifier ("sel_ref_cnt");
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

#ifdef OBJC_SELS_R_INTS
  /* unsigned int *sel_ref; */
  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_UNSIGNED]);
  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_INT], decl_specs);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("refs"));
#endif

#ifdef OBJC_SELS_R_STRUCT_PTRS
  /* struct objc_selector **sel_ref; */
  decl_specs = build_tree_list (NULLT,
				xref_tag (RECORD_TYPE,
					  get_identifier (TAG_SELECTOR)));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("refs"));
  field_decl = build1 (INDIRECT_REF, NULLT, field_decl);
#endif

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* short cls_def_cnt; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_SHORT]);
  field_decl = get_identifier ("cls_def_cnt");
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* short cat_def_cnt; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_SHORT]);
  field_decl = get_identifier ("cat_def_cnt");
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* void *defs[cls_def_cnt + cat_def_cnt]; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_VOID]);
  field_decl = build_nt (ARRAY_REF, get_identifier ("defs"),
			 build_int_2 (imp_count + cat_count, 0));
  field_decl = build1 (INDIRECT_REF, NULLT, field_decl);
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  finish_struct (objc_symtab_template, field_decl_chain);
}

static tree
init_def_list ()
{
  tree expr, initlist = NULLT;
  struct imp_entry *impent;

  if (imp_count)
    for (impent = imp_list; impent; impent = impent->next)
      {
	if (TREE_CODE (impent->imp_context) == CLASS_IMPLEMENTATION)
	  {
	    expr = build_unary_op (ADDR_EXPR, impent->class_decl, 0);
	    initlist = tree_cons (NULLT, expr, initlist);
	  }
      }

  if (cat_count)
    for (impent = imp_list; impent; impent = impent->next)
      {
	if (TREE_CODE (impent->imp_context) == CATEGORY_IMPLEMENTATION)
	  {
	    expr = build_unary_op (ADDR_EXPR, impent->class_decl, 0);
	    initlist = tree_cons (NULLT, expr, initlist);
	  }
      }
  return build_nt (CONSTRUCTOR, NULLT, nreverse (initlist));
}

/*
 *	struct objc_symtab {
 *		long sel_ref_cnt;
 *		char *refs;
 *		long cls_def_cnt;
 *		long cat_def_cnt;
 *		void *defs[cls_def_cnt + cat_def_cnt];
 *	};
 */
static tree
init_objc_symtab ()
{
  tree initlist;

  /* sel_ref_cnt = { ..., 5, ... } */

  initlist = build_tree_list (NULLT, build_int_2 (0, 0));

  /* refs = { ..., _OBJC_SELECTOR_REFERENCES, ... } */

  initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);

  /* cls_def_cnt = { ..., 5, ... } */

  initlist = tree_cons (NULLT, build_int_2 (imp_count, 0), initlist);

  /* cat_def_cnt = { ..., 5, ... } */

  initlist = tree_cons (NULLT, build_int_2 (cat_count, 0), initlist);

  /* cls_def = { ..., { &Foo, &Bar, ...}, ... } */

  if (imp_count || cat_count)
    initlist = tree_cons (NULLT, init_def_list (), initlist);

  return build_nt (CONSTRUCTOR, NULLT, nreverse (initlist));
}

static void
forward_declare_categories ()
{
  struct imp_entry *impent;
  tree sav = implementation_context;
  for (impent = imp_list; impent; impent = impent->next)
    {
      if (TREE_CODE (impent->imp_context) == CATEGORY_IMPLEMENTATION)
	{
	  tree sc_spec, decl_specs, decl;

	  sc_spec = build_tree_list (NULLT, ridpointers[(int) RID_EXTERN]);
	  decl_specs = tree_cons (NULLT, objc_category_template, sc_spec);

	  implementation_context = impent->imp_context;
	  impent->class_decl = define_decl (
		synth_id_with_class_suffix ("_OBJC_CATEGORY", implementation_context),
		decl_specs);
	}
    }
  implementation_context = sav;
}

static void
generate_objc_symtab_decl ()
{
  tree sc_spec;

  if (!objc_category_template)
    build_category_template ();

  /* forward declare categories */
  if (cat_count)
    forward_declare_categories ();

  if (!objc_symtab_template)
    build_objc_symtab_template ();

  sc_spec = build_tree_list (NULLT, ridpointers[(int) RID_STATIC]);

  _OBJC_SYMBOLS_decl = start_decl (get_identifier ("_OBJC_SYMBOLS"),
				   tree_cons (NULLT, objc_symtab_template, sc_spec), 1);

  finish_decl (_OBJC_SYMBOLS_decl, init_objc_symtab (), NULLT);
}

/*
 *	tree_node------->tree_node----->...
 *          |                |
 *          | (value)        | (value)
 *          |                |
 *	  expr             expr
 */
static tree
init_module_descriptor ()
{
  tree initlist, expr;

  /* version = { 1, ... } */

  expr = build_int_2 (OBJC_VERSION, 0);
  initlist = build_tree_list (NULLT, expr);

  /* size = { ..., sizeof (struct objc_module), ... } */

  expr = build_int_2 (TREE_INT_CST_LOW (TYPE_SIZE (objc_module_template)) /
		      BITS_PER_UNIT, 0);
  initlist = tree_cons (NULLT, expr, initlist);

  /* name = { ..., "foo.m", ... } */

  expr = add_objc_string (get_identifier (input_filename), class_names);
  initlist = tree_cons (NULLT, expr, initlist);

  /* symtab = { ..., _OBJC_SYMBOLS, ... } */

  if (_OBJC_SYMBOLS_decl)
    expr = build_unary_op (ADDR_EXPR, _OBJC_SYMBOLS_decl, 0);
  else
    expr = build_int_2 (0, 0);
  initlist = tree_cons (NULLT, expr, initlist);

  return build_nt (CONSTRUCTOR, NULLT, nreverse (initlist));
}

/*
 *  struct objc_module { ... } _OBJC_MODULE = { ... };
 */
static void
build_module_descriptor ()
{
  tree decl_specs, field_decl, field_decl_chain;

  objc_module_template = start_struct (RECORD_TYPE, get_identifier (_TAG_MODULE));

  /* long version; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_LONG]);
  field_decl = get_identifier ("version");
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

  /* long  size; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_LONG]);
  field_decl = get_identifier ("size");
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* char  *name; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_CHAR]);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("name"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_symtab *symtab; */

  decl_specs = get_identifier (_TAG_SYMTAB);
  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE, decl_specs));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("symtab"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  finish_struct (objc_module_template, field_decl_chain);

  /* create an instance of "objc_module" */

  decl_specs = tree_cons (NULLT, objc_module_template,
			  build_tree_list (NULLT, ridpointers[(int) RID_STATIC]));

  _OBJC_MODULES_decl = start_decl (get_identifier ("_OBJC_MODULES"),
				   decl_specs, 1);

  finish_decl (_OBJC_MODULES_decl, init_module_descriptor (), NULLT);

  /* Mark the decl as used to avoid "defined but not used" warning. */
  TREE_USED (_OBJC_MODULES_decl) = 1;
}

/* extern const char _OBJC_STRINGS[]; */

static void
generate_forward_declaration_to_string_table ()
{
  tree sc_spec, decl_specs, expr_decl;

  sc_spec = tree_cons (NULLT, ridpointers[(int) RID_EXTERN], NULLT);
  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_CHAR], sc_spec);

  expr_decl = build_nt (ARRAY_REF, get_identifier ("_OBJC_STRINGS"), NULLT);

  _OBJC_STRINGS_decl = define_decl (expr_decl, decl_specs);
}

/* Output all strings. */

static void
generate_strings ()
{
  tree sc_spec, decl_specs, expr_decl;
  tree chain, string_expr;
  tree string, decl;
  
  for (chain = class_names_chain; chain; chain = TREE_CHAIN (chain))
    {
      string = TREE_VALUE (chain);
      decl = TREE_PURPOSE (chain);
      sc_spec = tree_cons (NULLT, ridpointers[(int) RID_STATIC], NULLT);
      decl_specs = tree_cons (NULLT, ridpointers[(int) RID_CHAR], sc_spec);
      expr_decl = build_nt (ARRAY_REF, DECL_NAME (decl), NULLT);
      decl = start_decl (expr_decl, decl_specs, 1);
      string_expr = my_build_string (IDENTIFIER_LENGTH (string) + 1,
				     IDENTIFIER_POINTER (string));
      finish_decl (decl, string_expr, NULLT);
    }
  
  for (chain = meth_var_names_chain; chain; chain = TREE_CHAIN (chain))
    {
      string = TREE_VALUE (chain);
      decl = TREE_PURPOSE (chain);
      sc_spec = tree_cons (NULLT, ridpointers[(int) RID_STATIC], NULLT);
      decl_specs = tree_cons (NULLT, ridpointers[(int) RID_CHAR], sc_spec);
      expr_decl = build_nt (ARRAY_REF, DECL_NAME (decl), NULLT);
      decl = start_decl (expr_decl, decl_specs, 1);
      string_expr = my_build_string (IDENTIFIER_LENGTH (string) + 1,
				     IDENTIFIER_POINTER (string));
      finish_decl (decl, string_expr, NULLT);
    }
  
  for (chain = meth_var_types_chain; chain; chain = TREE_CHAIN (chain))
    {
      string = TREE_VALUE (chain);
      decl = TREE_PURPOSE (chain);
      sc_spec = tree_cons (NULLT, ridpointers[(int) RID_STATIC], NULLT);
      decl_specs = tree_cons (NULLT, ridpointers[(int) RID_CHAR], sc_spec);
      expr_decl = build_nt (ARRAY_REF, DECL_NAME (decl), NULLT);
      decl = start_decl (expr_decl, decl_specs, 1);
      string_expr = my_build_string (IDENTIFIER_LENGTH (string) + 1,
				     IDENTIFIER_POINTER (string));
      finish_decl (decl, string_expr, NULLT);
    }
}

static tree
build_selector_reference_decl (name)
      tree name;
{
  tree decl, ident;
  char buf[256];
  struct obstack *save_current_obstack = current_obstack;
  struct obstack *save_rtl_obstack = rtl_obstack;
  static int idx = 0;
  
  sprintf (buf, "_OBJC_SELECTOR_REFERENCES_%d", idx++);
  
  /* new stuff */
  rtl_obstack = current_obstack = &permanent_obstack;
  ident = get_identifier (buf);
  
  decl = build_decl (VAR_DECL, ident, _selector_type);
  TREE_EXTERNAL (decl) = 1;
  TREE_PUBLIC (decl) = 1;
  TREE_USED (decl) = 1;
#ifndef OBJCPLUS
  TREE_READONLY (decl) = 1;
#endif /* not OBJCPLUS */
  
  make_decl_rtl (decl, 0, 1); /* usually called from `rest_of_decl_compilation' */
  pushdecl_top_level (decl);  /* our `extended/custom' pushdecl in c-decl.c */
  
  current_obstack = save_current_obstack;
  rtl_obstack = save_rtl_obstack;
  
  return decl;
}


/* Just a handy wrapper for add_objc_string() */

static tree
build_selector (ident)
     tree ident;
{
  tree expr = add_objc_string (ident, meth_var_names);
  
  return build_c_cast (_selector_type, expr);	/* cast! */
}

static void
build_selector_translation_table ()
{
  tree sc_spec, decl_specs, expr_decl;
  tree chain, initlist = NULLT;
  tree var_decl;
  tree expr, name, decl;

  for (chain = sel_ref_chain; chain; chain = TREE_CHAIN (chain))
    {
      expr = build_selector (TREE_VALUE (chain));
      name = DECL_NAME (TREE_PURPOSE (chain));

      sc_spec = build_tree_list (NULLT, ridpointers[RID_STATIC]);

#ifdef OBJC_SELS_R_INTS
      /* static unsigned int _OBJC_SELECTOR_REFERENCES_n = ...; */
      decl_specs = tree_cons (NULLT, ridpointers[(int) RID_UNSIGNED], sc_spec);
      decl_specs = tree_cons (NULLT, ridpointers[(int) RID_INT], decl_specs);
      var_decl = name;
#endif

#ifdef OBJC_SELS_R_STRUCT_PTRS
      /* static struct objc_selector *_OBJC_SELECTOR_REFERENCES_n = ...; */
      decl_specs = tree_cons (NULLT,
			      xref_tag (RECORD_TYPE,
					get_identifier (TAG_SELECTOR)),
			      sc_spec);
      var_decl = build1 (INDIRECT_REF, NULLT, name);
#endif

      /* the `decl' that is returned from start_decl is the one that we
	* forward declared in `build_selector_reference()'
	*/
      decl = start_decl (var_decl, decl_specs, 1); 

      finish_decl (decl, expr, NULLT);
    }
}

static void
add_class_reference (ident)
     tree ident;
{
  tree chain;

  if (chain = cls_ref_chain)
    {
      tree tail;
      do
        {
	  if (ident == TREE_VALUE (chain))
	    return;

	  tail = chain;
	  chain = TREE_CHAIN (chain);
        }
      while (chain);

      /* append to the end of the list */
      TREE_CHAIN (tail) = perm_tree_cons (NULLT, ident, NULLT);
    }
  else
    cls_ref_chain = perm_tree_cons (NULLT, ident, NULLT);
}

/*
 * sel_ref_chain is a list whose "value" fields will be instances of
 * identifier_node that represent the selector.
 */
static tree
build_selector_reference (ident)
     tree ident;
{
  tree *chain = &sel_ref_chain;
  tree decl;

  while (*chain)
    {
      if (TREE_VALUE (*chain) == ident)
	return TREE_PURPOSE (*chain);

      chain = &TREE_CHAIN (*chain);
    }
  
  decl = build_selector_reference_decl (ident);
  
  *chain = perm_tree_cons (decl, ident, NULLT);
  
  return decl;
}

/*
 * sel_refdef_chain is a list whose "value" fields will be instances of
 * identifier_node that represent the selector. It returns the offset of
 * the selector from the beginning of the _OBJC_STRINGS pool. This offset
 * is typically used by "init_selector ()" during code generation.
 */

/* For each string section we have a chain which maps identifier nodes to
   decls for the strings. */
static tree
add_objc_string (ident, section)
     tree ident;
     enum stringSection section;
{
  tree *chain, decl;

  if (section == class_names)
    chain = &class_names_chain;
  else if (section == meth_var_names)
    chain = &meth_var_names_chain;
  else if (section == meth_var_types)
    chain = &meth_var_types_chain;

  while (*chain)
    {
      if (TREE_VALUE (*chain) == ident)
        return TREE_PURPOSE (*chain);
      
      chain = &TREE_CHAIN (*chain);
    }
  
  decl = build_objc_string_decl (ident, section);
  
  *chain = perm_tree_cons (decl, ident, NULLT);
  
  return decl;
}

static tree
build_objc_string_decl (name, section)
     tree name;
     enum stringSection section;
{
  tree decl, ident;
  char buf[256];
  struct obstack *save_current_obstack = current_obstack;
  struct obstack *save_rtl_obstack = rtl_obstack;
  static int class_names_idx = 0;
  static int meth_var_names_idx = 0;
  static int meth_var_types_idx = 0;
  
  if (section == class_names)
    sprintf (buf, "_OBJC_CLASS_NAME_%d", class_names_idx++);
  else if (section == meth_var_names)
    sprintf (buf, "_OBJC_METH_VAR_NAME_%d", meth_var_names_idx++);
  else if (section == meth_var_types)
    sprintf (buf, "_OBJC_METH_VAR_TYPE_%d", meth_var_types_idx++);
  
  rtl_obstack = current_obstack = &permanent_obstack;
  ident = get_identifier (buf);
  
  decl = build_decl (VAR_DECL, ident, build_array_type (char_type_node, 0));
  TREE_EXTERNAL (decl) = 1;
  TREE_PUBLIC (decl) = 1;
  TREE_USED (decl) = 1;
#ifndef OBJCPLUS
  TREE_READONLY (decl) = 1;
  TREE_CONSTANT (decl) = 1;
#endif /* not OBJCPLUS */
  
  make_decl_rtl (decl, 0, 1); /* usually called from `rest_of_decl_compilation' */
  pushdecl_top_level (decl);  /* our `extended/custom' pushdecl in c-decl.c */
  
  current_obstack = save_current_obstack;
  rtl_obstack = save_rtl_obstack;
  
  return decl;
}


void
objc_declare_alias (alias_ident, class_ident)
     tree alias_ident;
     tree class_ident;
{
  if (!doing_objc_thang)
    objc_fatal ();

  if (is_class_name (class_ident) != class_ident)
    warning ("Cannot find class `%s'", IDENTIFIER_POINTER (class_ident));
  else if (is_class_name (alias_ident))
    warning ("Class `%s' already exists", IDENTIFIER_POINTER (alias_ident));
  else
    alias_chain = tree_cons (class_ident, alias_ident, alias_chain);
}

void
objc_declare_class (ident_list)
     tree ident_list;
{
  tree list;

  if (!doing_objc_thang)
    objc_fatal ();

  for (list = ident_list; list; list = TREE_CHAIN (list))
    {
      tree ident = TREE_VALUE (list);
      tree decl;

      if ((decl = lookup_name (ident)) != 0)
	{
	  error ("`%s' redeclared as different kind of symbol",
		  IDENTIFIER_POINTER (ident));
	  error_with_decl (decl, "previous declaration of `%s'");
	}

      if (! is_class_name (ident))
        {
#ifdef OBJCPLUS
	  tree record;

	  push_lang_context (lang_name_c);
	  record = xref_tag (RECORD_TYPE, ident);
	  pop_lang_context ();
#else /* OBJCPLUS */

	  tree record = xref_tag (RECORD_TYPE, ident);

#endif /* OBJCPLUS */
	  TREE_STATIC_TEMPLATE (record) = 1;
	  class_chain = tree_cons (NULLT, ident, class_chain);
	}
    }
}

tree
is_class_name (ident)
     tree ident;
{
  tree chain;

  if (lookup_interface (ident))
    return ident;

  for (chain = class_chain; chain; chain = TREE_CHAIN (chain))
    {
      if (ident == TREE_VALUE (chain))
	return ident;
    }

  for (chain = alias_chain; chain; chain = TREE_CHAIN (chain))
    {
      if (ident == TREE_VALUE (chain))
	return TREE_PURPOSE (chain);
    }

  return 0;
}

tree
lookup_interface (ident)
     tree ident;
{
  tree chain;

  for (chain = interface_chain; chain; chain = TREE_CHAIN (chain))
    {
      if (ident == CLASS_NAME (chain))
	return chain;
    }

  return NULLT;
}

static tree
objc_copy_list (list, head)
     tree list;
     tree *head;
{
  tree newlist = NULL_TREE, tail = NULL_TREE;

  while (list)
    {
      tail = copy_node (list);

      /* the following statement fixes a bug when inheriting instance
	 variables that are declared to be bitfields. finish_struct () expects
	 to find the width of the bitfield in DECL_INITIAL (), which it
	 nulls out after processing the decl of the super class...rather
	 than change the way finish_struct () works (which is risky),
	 I create the situation it expects...s.naroff (7/23/89).
	 */
#ifdef OBJCPLUS
      if (TREE_PACKED(tail) && DECL_SIZE_UNIT(tail) && DECL_INITIAL(tail) == 0)
	DECL_INITIAL(tail) = build_int_2(DECL_SIZE_UNIT(tail),0);
#else /* OBJCPLUS */
      if (DECL_BIT_FIELD (tail) && DECL_INITIAL (tail) == 0)
	DECL_INITIAL (tail) = build_int_2 (DECL_FRAME_SIZE (tail), 0);
#endif /* OBJCPLUS */

      newlist = chainon (newlist, tail);
      list = TREE_CHAIN (list);
    }
  *head = newlist;
  return tail;
}

/* used by:
 * build_private_template (), get_class_ivars (), and continue_class ().
 */
/* COPY is 1 when called from @defs.  In this case copy all fields.
   Otherwise don't copy leaf ivars since we rely on them being side-effected
   exactly once by finish_struct().  (mself 7/26/91) */
static tree
build_ivar_chain (interface, copy)
     tree interface;
     int copy;
{
  tree my_name, super_name, ivar_chain;

  my_name = CLASS_NAME (interface);
  super_name = CLASS_SUPER_NAME (interface);
  
  /* Possibly copy leaf ivars.  */
  if (copy)
    objc_copy_list (CLASS_IVARS (interface), &ivar_chain);
  else
    ivar_chain = CLASS_IVARS (interface);

  while (super_name)
    {
      tree op1;
      tree super_interface = lookup_interface (super_name);

      if (!super_interface)
        {
	  /* fatal did not work with 2 args...should fix */
	  error ("Cannot find interface declaration for `%s', superclass of `%s'",
		 IDENTIFIER_POINTER (super_name), IDENTIFIER_POINTER (my_name));
	  exit (34);
        }
      if (super_interface == interface)
        {
          fatal ("Circular inheritance in interface declaration for `%s'",
                 IDENTIFIER_POINTER (super_name));
        }
      interface = super_interface;
      my_name = CLASS_NAME (interface);
      super_name = CLASS_SUPER_NAME (interface);

      op1 = CLASS_IVARS (interface);
      if (op1)
        {
	  tree head, tail = objc_copy_list (op1, &head);

	  /* prepend super class ivars...make a copy of the list, we
	   * do not want to alter the original.
	   */
	  TREE_CHAIN (tail) = ivar_chain;
	  ivar_chain = head;
        }
    }
  return ivar_chain;
}

/*
 *  struct <classname> {
 *    struct objc_class *isa;
 *    ...
 *  };
 */
static tree
build_private_template (class)
     tree class;
{
  tree ivar_context;

  if (CLASS_STATIC_TEMPLATE (class))
    {
      _PRIVATE_record = CLASS_STATIC_TEMPLATE (class);
      ivar_context = TYPE_FIELDS (CLASS_STATIC_TEMPLATE (class));
    }
  else
    {
      _PRIVATE_record = start_struct (RECORD_TYPE, CLASS_NAME (class));

      ivar_context = build_ivar_chain (class, 0);

      finish_struct (_PRIVATE_record, ivar_context);

      CLASS_STATIC_TEMPLATE (class) = _PRIVATE_record;

      /* mark this record as class template - for class type checking */
      TREE_STATIC_TEMPLATE (_PRIVATE_record) = 1;
    }
  instance_type = groktypename (
				build_tree_list (build_tree_list (NULLT, _PRIVATE_record),
						 build1 (INDIRECT_REF, NULLT, NULLT)));
  return ivar_context;
}

/* begin code generation for protocols... */

/*
 *  struct objc_protocol {
 *    char *protocol_name;
 *    struct objc_protocol **protocol_list;
 *    struct objc_method_desc *instance_methods;
 *    struct objc_method_desc *class_methods;
 *  };
 */
static tree
build_protocol_template ()
{
  tree decl_specs, field_decl, field_decl_chain;
  tree template;

  template = start_struct (RECORD_TYPE, get_identifier (_TAG_PROTOCOL));

  /* struct objc_class *isa; */

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
					get_identifier (_TAG_CLASS)));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("isa"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

  /* char *protocol_name; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_CHAR]);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("protocol_name"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_protocol **protocol_list; */

  decl_specs = build_tree_list (NULLT, template);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("protocol_list"));
  field_decl = build1 (INDIRECT_REF, NULLT, field_decl);
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_method_list *instance_methods; */

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
					get_identifier (_TAG_METHOD_PROTOTYPE_LIST)));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("instance_methods"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_method_list *class_methods; */

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
					get_identifier (_TAG_METHOD_PROTOTYPE_LIST)));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("class_methods"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  return finish_struct (template, field_decl_chain);
}

static tree
build_descriptor_table_initializer (entries, size)
     tree entries;
     int *size;
{
  tree initlist = NULLT;

  do
    {
      initlist = tree_cons (NULLT, build_selector (METHOD_SEL_NAME (entries)), initlist);

      initlist = tree_cons (NULLT, add_objc_string (METHOD_ENCODING (entries), meth_var_types), initlist);

      (*size)++;
      entries = TREE_CHAIN (entries);
    }
  while (entries);

  return build_nt (CONSTRUCTOR, NULLT, nreverse (initlist));
}

/*
 *  struct objc_method_prototype_list {
 *    int count;
 *    struct objc_method_prototype {
 *	SEL name;
 *	char *types;
 *    } list[1];	
 *  };
 */
static tree
build_method_prototype_list_template (list_type, size)
     tree list_type;
     int size;
{
  tree objc_ivar_list_id, objc_ivar_list_record;
  tree decl_specs, field_decl, field_decl_chain;

  /* generate an unnamed struct definition */

  objc_ivar_list_record = start_struct (RECORD_TYPE, NULLT); 

  /* int method_count; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_INT]);
  field_decl = get_identifier ("method_count");

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

  /* struct objc_method method_list[]; */

  decl_specs = build_tree_list (NULLT, list_type);
  field_decl = build_nt (ARRAY_REF, get_identifier ("method_list"),
			 build_int_2 (size, 0));

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  finish_struct (objc_ivar_list_record, field_decl_chain);

  return objc_ivar_list_record;
}

static tree
build_method_prototype_template ()
{
  tree proto_record;
  tree decl_specs, field_decl, field_decl_chain, parms;

  proto_record = start_struct (RECORD_TYPE, get_identifier (_TAG_METHOD_PROTOTYPE));

#ifdef OBJC_SELS_R_INTS
  /* unsigned int _cmd; */
  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_UNSIGNED], NULLT);
  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_INT], decl_specs);
  field_decl = get_identifier ("_cmd");
#endif

#ifdef OBJC_SELS_R_STRUCT_PTRS
  /* struct objc_selector *_cmd; */
  decl_specs = tree_cons (NULLT, xref_tag (RECORD_TYPE,
		          get_identifier (TAG_SELECTOR)), NULLT);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("_cmd"));
#endif

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_CHAR], NULLT);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("method_types"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  finish_struct (proto_record, field_decl_chain);

  return proto_record;
}

static tree
encode_method_prototype (method_decl, func_decl)
      tree method_decl;
      tree func_decl;
{
  tree parms;
  int stack_size = 0, i;
  tree user_args;
  
  bzero (utlbuf, BUFSIZE);

  /* `oneway' and 'bycopy', for remote object are the only method qualifiers */
  encode_type_qualifiers(TREE_PURPOSE(TREE_TYPE(method_decl)), utlbuf);

  /* C type */  
  encode_type (TREE_TYPE (TREE_TYPE (func_decl)), utlbuf, 
  		utlbuf+strlen(utlbuf), OBJC_ENCODE_INLINE_DEFS);
	   
  /* stack size */
  for (parms = DECL_ARGUMENTS (func_decl); parms;
       parms = TREE_CHAIN (parms))
    stack_size += TREE_INT_CST_LOW (TYPE_SIZE (DECL_ARG_TYPE (parms)))
		  / BITS_PER_UNIT;

  sprintf (&utlbuf[strlen (utlbuf)], "%d", stack_size);

  user_args = METHOD_SEL_ARGS (method_decl);

  /* argument types */
  for (parms = DECL_ARGUMENTS (func_decl), i = 0; parms;
       parms = TREE_CHAIN (parms), i++)
    {
      int offset_in_bytes;
  
      /* process argument qualifiers for user supplied arguments */
      if (i > 1)
        {
        encode_type_qualifiers(TREE_PURPOSE(TREE_TYPE(user_args)), utlbuf);
	user_args = TREE_CHAIN(user_args);
 	}
	
      /* type */ 
      encode_type (TREE_TYPE (parms), utlbuf, utlbuf+strlen(utlbuf), 
      			OBJC_ENCODE_INLINE_DEFS);
  
      /* compute offset */
#ifdef OBJCPLUS
      if (DECL_OFFSET (parms) >= 0)
	{
	  offset_in_bytes = DECL_OFFSET (parms) / BITS_PER_UNIT;
#else /* OBJCPLUS */
      if (GET_CODE (DECL_INCOMING_RTL (parms)) == MEM)
        {
	  rtx addr = XEXP (DECL_INCOMING_RTL (parms), 0);
	  
	  /* ??? Here we assume that the parm address is indexed
	      off the frame pointer or arg pointer.
	      If that is not true, we produce meaningless results,
	      but do not crash.  */
	  if (GET_CODE (addr) == PLUS
	      && GET_CODE (XEXP (addr, 1)) == CONST_INT)
	    offset_in_bytes = INTVAL (XEXP (addr, 1));
	  else
	    offset_in_bytes = 0;
#endif /* OBJCPLUS */
	  
	  /* This is the case where the parm is passed as an int or double
	      and it is converted to a char, short or float and stored back
	      in the parmlist.  In this case, describe the parm
	      with the variable's declared type, and adjust the address
	      if the least significant bytes (which we are using) are not
	      the first ones.  */
#ifdef BYTES_BIG_ENDIAN
	  if (TREE_TYPE (parms) != DECL_ARG_TYPE (parms))
	    offset_in_bytes += (GET_MODE_SIZE (TYPE_MODE (DECL_ARG_TYPE (parms)))
				- GET_MODE_SIZE (GET_MODE (DECL_RTL (parms))));
#endif
	}
      else
        offset_in_bytes = 0;
      
#ifdef OBJCPLUS
      sprintf (&utlbuf[strlen (utlbuf)], "%d", offset_in_bytes);
#else /* OBJCPLUS */
      /* The "+ 8" is a total hack to account for the return pc and
         saved fp on the 68k.  We should redefine this format! */
      sprintf (&utlbuf[strlen (utlbuf)], "%d", offset_in_bytes + 8);
#endif /* OBJCPLUS */
    }

  return get_identifier (utlbuf);
}

static tree
generate_descriptor_table (type, name, size, list, proto)
     tree type;
     char *name;
     int size;
     tree list;
     tree proto;
{
  tree sc_spec, decl_specs, decl, initlist;

  sc_spec = tree_cons (NULLT, ridpointers[(int) RID_STATIC], NULLT);
  decl_specs = tree_cons (NULLT, type, sc_spec);

  decl = start_decl (synth_id_with_class_suffix (name, proto), 
				decl_specs, 1);

  initlist = build_tree_list (NULLT, build_int_2 (size, 0));
  initlist = tree_cons (NULLT, list, initlist);

  finish_decl (decl, build_nt (CONSTRUCTOR, NULLT, nreverse (initlist)), NULLT);

  return decl;
}

static void
generate_method_descriptors (protocol)	/* generate_dispatch_tables */
  tree protocol;
{
  static tree objc_method_prototype_template;
  tree initlist, chain, method_list_template;
  tree cast, variable_length_type;
  int size;

  if (!objc_method_prototype_template)
    objc_method_prototype_template = build_method_prototype_template ();

  cast = build_tree_list (build_tree_list (NULLT, xref_tag (RECORD_TYPE,
				get_identifier (_TAG_METHOD_PROTOTYPE_LIST))), NULLT);
  variable_length_type = groktypename (cast);

  chain = PROTOCOL_CLS_METHODS (protocol);
  if (chain)
    {
      size = 0;
      
      initlist = build_descriptor_table_initializer (chain, &size);

      method_list_template = build_method_prototype_list_template(
				objc_method_prototype_template, size);

      _OBJC_CLASS_METHODS_decl = 
	    generate_descriptor_table (method_list_template,
				     "_OBJC_PROTOCOL_CLASS_METHODS", 
				     size, initlist, protocol);
      /* cast! */
      TREE_TYPE (_OBJC_CLASS_METHODS_decl) = variable_length_type;
    }
  else
    _OBJC_CLASS_METHODS_decl = 0;

  chain = PROTOCOL_NST_METHODS (protocol);
  if (chain)
    {
      size = 0;
      initlist = build_descriptor_table_initializer (chain, &size);

      method_list_template = build_method_prototype_list_template(
				objc_method_prototype_template, size);

      _OBJC_INSTANCE_METHODS_decl = 
	    generate_descriptor_table (method_list_template,
				     "_OBJC_PROTOCOL_INSTANCE_METHODS", 
				     size, initlist, protocol);
      /* cast! */
      TREE_TYPE (_OBJC_INSTANCE_METHODS_decl) = variable_length_type;
    }
  else
    _OBJC_INSTANCE_METHODS_decl = 0;
}

static tree
build_tmp_function_decl()
{
  tree decl_specs, expr_decl, parms;
  
  /* struct objc_object *objc_xxx (id, SEL, ...); */
  pushlevel (0);
  decl_specs = build_tree_list (NULLT, objc_object_reference);
  push_parm_decl (build_tree_list (decl_specs,
				   build1 (INDIRECT_REF, NULLT, NULLT)));

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
					  get_identifier (TAG_SELECTOR)));
  expr_decl = build1 (INDIRECT_REF, NULLT, NULLT);

  push_parm_decl (build_tree_list (decl_specs, expr_decl));
  parms = get_parm_info (0);
  poplevel (0, 0, 0);

  decl_specs = build_tree_list (NULLT, objc_object_reference);
  expr_decl = build_nt (CALL_EXPR, get_identifier("objc_xxx"), parms, NULLT);
  expr_decl = build1 (INDIRECT_REF, NULLT, expr_decl);

  return define_decl (expr_decl, decl_specs);
}

static void
hack_method_prototype(nst_methods, _tmp_decl)
   tree nst_methods;
   tree _tmp_decl;
{
      tree parms;
#ifdef OBJCPLUS
      extern tree last_function_parms;
#endif /* OBJCPLUS */

      /* Hack to avoid problem with static typing of self arg. */
      TREE_CODE(nst_methods) = CLASS_METHOD_DECL;
      start_method_def(nst_methods);
      TREE_CODE(nst_methods) = INSTANCE_METHOD_DECL;

      if (METHOD_ADD_ARGS (nst_methods) == (tree)1)
        parms = get_parm_info (0); /* we have a `, ...' */
      else
        parms = get_parm_info (1); /* place a `void_at_end' */

      poplevel (0, 0, 0);	/* must be called BEFORE "start_function ()" */
  
      /* usually called from store_parm_decls()->init_function_start() */ 

      init_emit();	// needed to make assign_parms() work (with -O)

#ifdef OBJCPLUS
      /* usually called from start_function()->grokdeclarator(). */ 
      grokparms (parms, 1);
      /* usually done by store_parm_decls(). */
      DECL_ARGUMENTS(_tmp_decl) = last_function_parms;
#else /* OBJCPLUS */
      DECL_ARGUMENTS(_tmp_decl) = TREE_PURPOSE(parms);
#endif /* OBJCPLUS */
      
      /* install return type */
      TREE_TYPE (TREE_TYPE (_tmp_decl)) = groktypename (TREE_TYPE (nst_methods));

      {
      /* code take from start_function() */
      tree restype = TREE_TYPE (TREE_TYPE (_tmp_decl));
      /* Promote the value to int before returning it.  */
      if (TREE_CODE (restype) == INTEGER_TYPE
          && TYPE_PRECISION (restype) < TYPE_PRECISION (integer_type_node))
        restype = integer_type_node;
      DECL_RESULT (_tmp_decl) = build_decl (RESULT_DECL, 0, restype);
      }
      
      /* typically called from expand_function_start() for function definitions */
      assign_parms(_tmp_decl, 0);
      
      /* install return type */
      TREE_TYPE (TREE_TYPE (_tmp_decl)) = groktypename (TREE_TYPE (nst_methods));
}

static void 
generate_protocol_references (plist)
     tree plist;
{ 
     tree lproto;

     /* forward declare protocols referenced */
     for (lproto = plist; lproto; lproto = TREE_CHAIN(lproto)) {
        tree proto = TREE_VALUE(lproto);

        if ((TREE_CODE(proto) == PROTOCOL_INTERFACE) &&
              PROTOCOL_NAME(proto)) {

          if (!PROTOCOL_FORWARD_DECL (proto)) 
            build_protocol_reference(proto);

          if (PROTOCOL_LIST (proto))
            generate_protocol_references(PROTOCOL_LIST (proto));
        }
     }     
}

static void 
generate_protocols ()
{
  tree p, _tmp_decl, encoding;
  tree sc_spec, decl_specs, decl;
  tree initlist, protocol_name_expr, refs_decl, refs_expr;
  tree cast_type2 = 0, cast_type = 0;
  
  _tmp_decl = build_tmp_function_decl();

  if (!objc_protocol_template)
    objc_protocol_template = build_protocol_template ();

  /* if a protocol was directly referenced, pull in indirect references */
  for (p = protocol_chain; p; p = TREE_CHAIN(p))
    if (PROTOCOL_FORWARD_DECL (p) && PROTOCOL_LIST (p))
      generate_protocol_references(PROTOCOL_LIST (p));

  for (p = protocol_chain; p; p = TREE_CHAIN(p))
    {
    tree nst_methods = PROTOCOL_NST_METHODS(p);
    tree cls_methods = PROTOCOL_CLS_METHODS(p);
    char buf[256];

    /* if protocol wasn't referenced, don't generate any code */        
    if (!PROTOCOL_FORWARD_DECL (p))
      continue;

    /* Make sure we link in the Protocol class. */
    add_class_reference (get_identifier (PROTOCOL_OBJECT_CLASS_NAME));

    while (nst_methods)
      {
      hack_method_prototype(nst_methods, _tmp_decl);
      encoding = encode_method_prototype(nst_methods, _tmp_decl);
      METHOD_ENCODING (nst_methods) = encoding;
      
      nst_methods = TREE_CHAIN (nst_methods);
      }

    while (cls_methods)
      {
      hack_method_prototype(cls_methods, _tmp_decl);
      encoding = encode_method_prototype(cls_methods, _tmp_decl);
      METHOD_ENCODING (cls_methods) = encoding;
      
      cls_methods = TREE_CHAIN (cls_methods);
      }
    generate_method_descriptors(p);

    if (PROTOCOL_LIST(p))
      refs_decl = generate_protocol_list(p);
    else
      refs_decl = 0;

    /* static struct objc_protocol _OBJC_PROTOCOL_<mumble>; */

    sc_spec = tree_cons (NULLT, ridpointers[(int) RID_STATIC], NULLT);
    decl_specs = tree_cons (NULLT, objc_protocol_template, sc_spec);

    decl = start_decl (
      		synth_id_with_class_suffix ("_OBJC_PROTOCOL", p),
		decl_specs, 1);

    protocol_name_expr = add_objc_string (PROTOCOL_NAME (p), class_names);

    if (refs_decl)
      {
      if (!cast_type2)
        cast_type2 = groktypename (build_tree_list(
			build_tree_list (NULLT,	objc_protocol_template), 
			build1 (INDIRECT_REF, NULLT, 
			  build1 (INDIRECT_REF, NULLT, NULLT))));
 
      refs_expr = build_unary_op (ADDR_EXPR, refs_decl, 0);
      TREE_TYPE(refs_expr) = cast_type2;
      }
    else
      refs_expr = build_int_2 (0, 0);

    /* _OBJC_INSTANCE_METHODS_decl/_OBJC_CLASS_METHODS_decl are set
       by generate_method_descriptors(), which is called above */
    initlist = build_protocol_initializer (protocol_name_expr, refs_expr,
					 _OBJC_INSTANCE_METHODS_decl, 
					 _OBJC_CLASS_METHODS_decl);
    finish_decl (decl, initlist, NULLT);

    /* Mark the decl as used to avoid "defined but not used" warning. */
    TREE_USED (decl) = 1;
    }
}

static tree
build_protocol_initializer (protocol_name, protocol_list,
			    instance_methods, class_methods)
     tree protocol_name;
     tree protocol_list;
     tree instance_methods;
     tree class_methods;
{
  tree initlist = NULLT, expr;
  static tree cast_type = 0;

  if (!cast_type)
    cast_type = groktypename (build_tree_list (
                   build_tree_list (NULLT, 
                     xref_tag (RECORD_TYPE, get_identifier (_TAG_CLASS))), 
                     build1 (INDIRECT_REF, NULLT, NULLT)));

  /* filling the "isa" in with one allows the runtime system to
     detect that the version change...should remove before final release */

  expr = build_int_2 (PROTOCOL_VERSION, 0);
  TREE_TYPE(expr) = cast_type;
  initlist = tree_cons (NULLT, expr, initlist);
  initlist = tree_cons (NULLT, protocol_name, initlist);
  initlist = tree_cons (NULLT, protocol_list, initlist);

  if (!instance_methods)
    initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);
  else
    {
      expr = build_unary_op (ADDR_EXPR, instance_methods, 0);
      initlist = tree_cons (NULLT, expr, initlist);
    }
  if (!class_methods)
    initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);
  else
    {
      expr = build_unary_op (ADDR_EXPR, class_methods, 0);
      initlist = tree_cons (NULLT, expr, initlist);
    }
  return build_nt (CONSTRUCTOR, NULLT, nreverse (initlist));
}

/* end code generation for protocols... */

/*
 *  struct objc_category {
 *    char *category_name;
 *    char *class_name;
 *    struct objc_method_list *instance_methods;
 *    struct objc_method_list *class_methods;
 *    struct objc_protocol_list *protocols;
 *  };
 */
static void
build_category_template ()
{
  tree decl_specs, field_decl, field_decl_chain;

  objc_category_template = start_struct (RECORD_TYPE,
					 get_identifier (_TAG_CATEGORY));
  /* char *category_name; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_CHAR]);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("category_name"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

  /* char *class_name; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_CHAR]);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("class_name"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_method_list *instance_methods; */

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
						 get_identifier (_TAG_METHOD_LIST)));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("instance_methods"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_method_list *class_methods; */

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
						 get_identifier (_TAG_METHOD_LIST)));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("class_methods"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_protocol **protocol_list; */

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
					  get_identifier (_TAG_PROTOCOL)));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("protocol_list"));
  field_decl = build1 (INDIRECT_REF, NULLT, field_decl);
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  finish_struct (objc_category_template, field_decl_chain);
}

/*
 *  struct objc_class {
 *    struct objc_class *isa;
 *    struct objc_class *super_class;
 *    char *name;
 *    long version;
 *    long info;
 *    long instance_size;
 *    struct objc_ivar_list *ivars;
 *    struct objc_method_list *methods;
 *    struct objc_cache *cache;
 *    struct objc_protocol_list *protocols;
 *  };
 */
static void
build_class_template ()
{
  tree decl_specs, field_decl, field_decl_chain;

  objc_class_template = start_struct (RECORD_TYPE, get_identifier (_TAG_CLASS));

  /* struct objc_class *isa; */

  decl_specs = build_tree_list (NULLT, objc_class_template);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("isa"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

  /* struct objc_class *super_class; */

  decl_specs = build_tree_list (NULLT, objc_class_template);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("super_class"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* char *name; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_CHAR]);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("name"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* long version; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_LONG]);
  field_decl = get_identifier ("version");
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* long info; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_LONG]);
  field_decl = get_identifier ("info");
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* long instance_size; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_LONG]);
  field_decl = get_identifier ("instance_size");
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_ivar_list *ivars; */

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
						 get_identifier (_TAG_IVAR_LIST)));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("ivars"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_method_list *methods; */

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
						 get_identifier (_TAG_METHOD_LIST)));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("methods"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_cache *cache; */

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
						 get_identifier ("objc_cache")));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("cache"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_protocol **protocol_list; */

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE,
					  get_identifier (_TAG_PROTOCOL)));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("protocol_list"));
  field_decl = build1 (INDIRECT_REF, NULLT, field_decl);
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  finish_struct (objc_class_template, field_decl_chain);
}

#ifdef OBJC_CLASS_REFS
/*
 *  struct _objc_class_ref {
 *    struct _objc_class *class;
 *  };
 */
static void
build_class_ref_template ()
{
  tree decl_specs, field_decl;

  objc_class_ref_template
    = start_struct (RECORD_TYPE, get_identifier ("_objc_class_ref"));

  /* struct objc_class *class; */

  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE, get_identifier (TAG_CLASS)));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("class"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  finish_struct (objc_class_ref_template, field_decl);
}
#endif /* OBJC_CLASS_REFS */

/*
 * generate appropriate forward declarations for an implementation
 */
static void
synth_forward_declarations ()
{
  tree sc_spec, decl_specs, factory_id, anId;

  /* extern struct objc_class _OBJC_CLASS_<my_name>; */

  anId = synth_id_with_class_suffix ("_OBJC_CLASS", implementation_context);

  sc_spec = build_tree_list (NULLT, ridpointers[(int) RID_EXTERN]);
  decl_specs = tree_cons (NULLT, objc_class_template, sc_spec);
  _OBJC_CLASS_decl = define_decl (anId, decl_specs);

  /* extern struct objc_class _OBJC_METACLASS_<my_name>; */

  anId = synth_id_with_class_suffix ("_OBJC_METACLASS", implementation_context);

  _OBJC_METACLASS_decl = define_decl (anId, decl_specs);

  /* pre-build the following entities - for speed/convenience. */

  anId = get_identifier ("super_class");
  _clsSuper_ref = build_component_ref (_OBJC_CLASS_decl, anId);
  __clsSuper_ref = build_component_ref (_OBJC_METACLASS_decl, anId);
}

static void
error_with_ivar (message, decl, rawdecl)
     char *message;
     tree decl;
     tree rawdecl;
{
  extern int errorcount;

#ifdef REPORT_EVENT
  bzero (errbuf, BUFSIZE);
  report_event (0, current_function_decl,
		DECL_SOURCE_FILE (decl), DECL_SOURCE_LINE (decl),
		"%s `%s'", message, gen_declaration (rawdecl, errbuf), 0);
#endif
  fprintf (stderr, "%s:%d: ",
	   DECL_SOURCE_FILE (decl), DECL_SOURCE_LINE (decl));
  bzero (errbuf, BUFSIZE);
  fprintf (stderr, "%s `%s'\n", message, gen_declaration (rawdecl, errbuf));
  errorcount++;
}

#define USERTYPE(t)	(TREE_CODE (t) == RECORD_TYPE || \
			 TREE_CODE (t) == UNION_TYPE ||  \
			 TREE_CODE (t) == ENUMERAL_TYPE)

static void
check_ivars (inter, imp)
     tree inter;
     tree imp;
{
  tree intdecls = CLASS_IVARS (inter);
  tree impdecls = CLASS_IVARS (imp);
  tree rawintdecls = CLASS_RAW_IVARS (inter);
  tree rawimpdecls = CLASS_RAW_IVARS (imp);

  while (1)
    {
      tree t1, t2;

      if (intdecls == 0 && impdecls == 0)
	break;
      if (intdecls == 0 || impdecls == 0)
	{
	  error ("inconsistent instance variable specification");
	  break;
	}
      t1 = TREE_TYPE (intdecls); t2 = TREE_TYPE (impdecls);

      if (!comptypes (t1, t2))
	{
	  if (DECL_NAME (intdecls) == DECL_NAME (impdecls))
	    {
	      error_with_ivar ("conflicting instance variable type",
			       impdecls, rawimpdecls);
	      error_with_ivar ("previous declaration of",
			       intdecls, rawintdecls);
	    }
	  else			/* both the type and the name don't match */
	    {
	      error ("inconsistent instance variable specification");
	      break;
	    }
	}
      else if (DECL_NAME (intdecls) != DECL_NAME (impdecls))
	{
	  error_with_ivar ("conflicting instance variable name",
			   impdecls, rawimpdecls);
	  error_with_ivar ("previous declaration of",
			   intdecls, rawintdecls);
	}
      intdecls = TREE_CHAIN (intdecls);
      impdecls = TREE_CHAIN (impdecls);
      rawintdecls = TREE_CHAIN (rawintdecls);
      rawimpdecls = TREE_CHAIN (rawimpdecls);
    }
}

/*
 * 	struct objc_super {
 *		id self;
 *		struct objc_class *class;
 *	};
 */
static tree
build_super_template ()
{
  tree record, decl_specs, field_decl, field_decl_chain;

  record = start_struct (RECORD_TYPE, get_identifier (_TAG_SUPER));

  /* struct objc_object *self; */

  decl_specs = build_tree_list (NULLT, objc_object_reference);
  field_decl = get_identifier ("self");
  field_decl = build1 (INDIRECT_REF, NULLT, field_decl);
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

  /* struct objc_class *class; */

  decl_specs = get_identifier (_TAG_CLASS);
  decl_specs = build_tree_list (NULLT, xref_tag (RECORD_TYPE, decl_specs));
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("class"));

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  finish_struct (record, field_decl_chain);

  /* `struct objc_super *' */
  super_type = groktypename (build_tree_list (build_tree_list (NULLT, record),
					      build1 (INDIRECT_REF, NULLT, NULLT)));
  return record;
}

/*
 * 	struct objc_ivar {
 *		char *ivar_name;
 *		char *ivar_type;
 *		int ivar_offset;
 *	};
 */
static tree
build_ivar_template ()
{
  tree objc_ivar_id, objc_ivar_record;
  tree decl_specs, field_decl, field_decl_chain;

  objc_ivar_id = get_identifier (_TAG_IVAR);
  objc_ivar_record = start_struct (RECORD_TYPE, objc_ivar_id);

  /* char *ivar_name; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_CHAR]);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("ivar_name"));

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

  /* char *ivar_type; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_CHAR]);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("ivar_type"));

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* int ivar_offset; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_INT]);
  field_decl = get_identifier ("ivar_offset");

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  finish_struct (objc_ivar_record, field_decl_chain);

  return objc_ivar_record;
}

/*
 * 	struct {
 *		int ivar_count;
 *		struct objc_ivar ivar_list[ivar_count];
 *	};
 */
static tree
build_ivar_list_template (list_type, size)
     tree list_type;
     int size;
{
  tree objc_ivar_list_id, objc_ivar_list_record;
  tree decl_specs, field_decl, field_decl_chain;

  objc_ivar_list_record = start_struct (RECORD_TYPE, NULLT);

  /* int ivar_count; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_INT]);
  field_decl = get_identifier ("ivar_count");

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

  /* struct objc_ivar ivar_list[]; */

  decl_specs = build_tree_list (NULLT, list_type);
  field_decl = build_nt (ARRAY_REF, get_identifier ("ivar_list"),
			 build_int_2 (size, 0));

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  finish_struct (objc_ivar_list_record, field_decl_chain);

  return objc_ivar_list_record;
}

/*
 * 	struct {
 *		int method_next;
 *		int method_count;
 *		struct objc_method method_list[method_count];
 *	};
 */
static tree
build_method_list_template (list_type, size)
     tree list_type;
     int size;
{
  tree objc_ivar_list_id, objc_ivar_list_record;
  tree decl_specs, field_decl, field_decl_chain;

  objc_ivar_list_record = start_struct (RECORD_TYPE, NULLT);

  /* int method_next; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_INT]);
  field_decl = get_identifier ("method_next");

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

  /* int method_count; */

  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_INT]);
  field_decl = get_identifier ("method_count");

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* struct objc_method method_list[]; */

  decl_specs = build_tree_list (NULLT, list_type);
  field_decl = build_nt (ARRAY_REF, get_identifier ("method_list"),
			 build_int_2 (size, 0));

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  finish_struct (objc_ivar_list_record, field_decl_chain);

  return objc_ivar_list_record;
}

static tree
build_ivar_list_initializer (field_decl, size)
     tree field_decl;
     int *size;
{
  tree initlist = NULLT;

  do
    {
    /* set name */
    if (DECL_NAME (field_decl))
      {
        initlist = tree_cons (NULLT, add_objc_string (DECL_NAME (field_decl), meth_var_names), initlist);
      }
    else
      {
        /* unnamed bit-field ivar (yuck). */
        initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);
      }

      /* set type */
      bzero (utlbuf, BUFSIZE);
      encode_field_decl (field_decl, utlbuf, utlbuf,
			 OBJC_ENCODE_DONT_INLINE_DEFS);

      initlist = tree_cons (NULLT, add_objc_string (get_identifier (utlbuf), meth_var_types), initlist);

      /* set offset */
#ifdef OBJCPLUS
    initlist = tree_cons(NULLT, build_int_2(DECL_OFFSET(field_decl)/
				DECL_SIZE_UNIT(field_decl), 0), initlist);
#else /* OBJCPLUS */
      initlist = tree_cons (NULLT,
			    build_int_2 (TREE_INT_CST_LOW (DECL_FIELD_BITPOS (field_decl)) / BITS_PER_UNIT, 0),
			    
			    initlist);
#endif /* OBJCPLUS */
      (*size)++;
      field_decl = TREE_CHAIN (field_decl);
    }
  while (field_decl);

  return build_nt (CONSTRUCTOR, NULLT, nreverse (initlist));
}

static tree
generate_ivars_list (type, name, size, list)
     tree type;
     char *name;
     int size;
     tree list;
{
  tree sc_spec, decl_specs, decl, initlist;

  sc_spec = tree_cons (NULLT, ridpointers[(int) RID_STATIC], NULLT);
  decl_specs = tree_cons (NULLT, type, sc_spec);

  decl = start_decl (synth_id_with_class_suffix (name, implementation_context),
				 decl_specs, 1);

  initlist = build_tree_list (NULLT, build_int_2 (size, 0));
  initlist = tree_cons (NULLT, list, initlist);

  finish_decl (decl, build_nt (CONSTRUCTOR, NULLT, nreverse (initlist)), NULLT);

  return decl;
}

static void
generate_ivar_lists ()
{
  tree initlist, ivar_list_template, chain;
  tree cast, variable_length_type;
  int size;

  if (!objc_ivar_template)
    objc_ivar_template = build_ivar_template ();

  cast = build_tree_list (build_tree_list (NULLT, xref_tag (RECORD_TYPE,
							    get_identifier (_TAG_IVAR_LIST))), NULLT);
  variable_length_type = groktypename (cast);

  /* only generate class variables for the root of the inheritance
     hierarchy since these will be the same for every class */

  if (CLASS_SUPER_NAME (implementation_template) == NULLT
      && (chain = TYPE_FIELDS (objc_class_template)))
    {
      size = 0;
      initlist = build_ivar_list_initializer (chain, &size);

      ivar_list_template = build_ivar_list_template (objc_ivar_template, size);

      _OBJC_CLASS_VARIABLES_decl =
	generate_ivars_list (ivar_list_template, "_OBJC_CLASS_VARIABLES",
			     size, initlist);
      /* cast! */
      TREE_TYPE (_OBJC_CLASS_VARIABLES_decl) = variable_length_type;
    }
  else
    _OBJC_CLASS_VARIABLES_decl = 0;

  chain = CLASS_IVARS (implementation_template);
  if (chain)
    {
      size = 0;
      initlist = build_ivar_list_initializer (chain, &size);

      ivar_list_template = build_ivar_list_template (objc_ivar_template, size);

      _OBJC_INSTANCE_VARIABLES_decl =
	generate_ivars_list (ivar_list_template, "_OBJC_INSTANCE_VARIABLES",
			     size, initlist);
      /* cast! */
      TREE_TYPE (_OBJC_INSTANCE_VARIABLES_decl) = variable_length_type;
    }
  else
    _OBJC_INSTANCE_VARIABLES_decl = 0;
}

static tree
build_dispatch_table_initializer (entries, size)
     tree entries;
     int *size;
{
  tree initlist = NULLT;

  do
    {
      initlist = tree_cons (NULLT, build_selector (METHOD_SEL_NAME (entries)), initlist);

      initlist = tree_cons (NULLT, add_objc_string (METHOD_ENCODING (entries), meth_var_types), initlist);

      initlist = tree_cons (NULLT, METHOD_DEFINITION (entries), initlist);

      (*size)++;
      entries = TREE_CHAIN (entries);
    }
  while (entries);

  return build_nt (CONSTRUCTOR, NULLT, nreverse (initlist));
}

/*
 * To accomplish method prototyping without generating all kinds of
 * inane warnings, the definition of the dispatch table entries were
 * changed from:
 *
 * 	struct objc_method { SEL _cmd; id (*_imp)(); };
 * to:
 * 	struct objc_method { SEL _cmd; void *_imp; };
 */
static tree
build_method_template ()
{
  tree _SLT_record;
  tree decl_specs, field_decl, field_decl_chain, parms;

  _SLT_record = start_struct (RECORD_TYPE, get_identifier (_TAG_METHOD));

#ifdef OBJC_SELS_R_INTS
  /* unsigned int _cmd; */
  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_UNSIGNED], NULLT);
  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_INT], decl_specs);
  field_decl = get_identifier ("_cmd");
#endif

#ifdef OBJC_SELS_R_STRUCT_PTRS
  /* struct objc_selector *_cmd; */
  decl_specs = tree_cons (NULLT,
			  xref_tag (RECORD_TYPE,
				    get_identifier (TAG_SELECTOR)),
			  NULLT);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("_cmd"));
#endif

  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  field_decl_chain = field_decl;

  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_CHAR], NULLT);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("method_types"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  /* void *_imp; */

  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_VOID], NULLT);
  field_decl = build1 (INDIRECT_REF, NULLT, get_identifier ("_imp"));
  field_decl = grokfield (input_filename, lineno, field_decl, decl_specs, NULLT);
  chainon (field_decl_chain, field_decl);

  finish_struct (_SLT_record, field_decl_chain);

  return _SLT_record;
}


static tree
generate_dispatch_table (type, name, size, list)
     tree type;
     char *name;
     int size;
     tree list;
{
  tree sc_spec, decl_specs, decl, initlist;

  sc_spec = tree_cons (NULLT, ridpointers[(int) RID_STATIC], NULLT);
  decl_specs = tree_cons (NULLT, type, sc_spec);

  decl = start_decl (synth_id_with_class_suffix (name, implementation_context), 
				decl_specs, 1);

  initlist = build_tree_list (NULLT, build_int_2 (0, 0));
  initlist = tree_cons (NULLT, build_int_2 (size, 0), initlist);
  initlist = tree_cons (NULLT, list, initlist);

  finish_decl (decl, build_nt (CONSTRUCTOR, NULLT, nreverse (initlist)), NULLT);

  return decl;
}

static void
generate_dispatch_tables ()
{
  tree initlist, chain, method_list_template;
  tree cast, variable_length_type;
  int size;

  if (!objc_method_template)
    objc_method_template = build_method_template ();

  cast = build_tree_list (build_tree_list (NULLT, xref_tag (RECORD_TYPE,
							    get_identifier (_TAG_METHOD_LIST))), NULLT);
  variable_length_type = groktypename (cast);

  chain = CLASS_CLS_METHODS (implementation_context);
  if (chain)
    {
      size = 0;
      initlist = build_dispatch_table_initializer (chain, &size);

      method_list_template = build_method_list_template (objc_method_template,
							 size);
      if (TREE_CODE (implementation_context) == CLASS_IMPLEMENTATION)
	_OBJC_CLASS_METHODS_decl = 
	    generate_dispatch_table (method_list_template,
				     "_OBJC_CLASS_METHODS", 
				     size, initlist);
      else
	/* we have a category */
	_OBJC_CLASS_METHODS_decl = 
	    generate_dispatch_table (method_list_template,
				     "_OBJC_CATEGORY_CLASS_METHODS", 
				     size, initlist);
      /* cast! */
      TREE_TYPE (_OBJC_CLASS_METHODS_decl) = variable_length_type;
    }
  else
    _OBJC_CLASS_METHODS_decl = 0;

  chain = CLASS_NST_METHODS (implementation_context);
  if (chain)
    {
      size = 0;
      initlist = build_dispatch_table_initializer (chain, &size);

      method_list_template = build_method_list_template (objc_method_template,
							 size);
      if (TREE_CODE (implementation_context) == CLASS_IMPLEMENTATION)
	_OBJC_INSTANCE_METHODS_decl = 
	    generate_dispatch_table (method_list_template,
				     "_OBJC_INSTANCE_METHODS", 
				     size, initlist);
      else
	/* we have a category */
	_OBJC_INSTANCE_METHODS_decl = 
	    generate_dispatch_table (method_list_template,
				     "_OBJC_CATEGORY_INSTANCE_METHODS", 
				     size, initlist);
      /* cast! */
      TREE_TYPE (_OBJC_INSTANCE_METHODS_decl) = variable_length_type;
    }
  else
    _OBJC_INSTANCE_METHODS_decl = 0;
}

static tree
generate_protocol_list(i_or_p)
    tree i_or_p;
{
      static tree cast_type = 0;
      tree initlist, decl_specs, sc_spec;
      tree refs_decl, expr_decl, lproto, e, plist;
      int size = 0;
      
      if (TREE_CODE(i_or_p) == CLASS_INTERFACE
          || TREE_CODE(i_or_p) == CATEGORY_INTERFACE)
        plist = CLASS_PROTOCOL_LIST(i_or_p);
      else if (TREE_CODE(i_or_p) == PROTOCOL_INTERFACE)
        plist = PROTOCOL_LIST(i_or_p);

      if (!cast_type)
        cast_type = groktypename (build_tree_list (
			build_tree_list (NULLT,	
			  xref_tag (RECORD_TYPE,
			  	get_identifier (_TAG_PROTOCOL))), 
			build1 (INDIRECT_REF, NULLT, NULLT)));

      /* compute size */     
      for (lproto = plist; lproto; lproto = TREE_CHAIN(lproto))
	if ((TREE_CODE(TREE_VALUE(lproto)) == PROTOCOL_INTERFACE) &&
            PROTOCOL_FORWARD_DECL(TREE_VALUE(lproto))) {
          size++;
        }

      /* build initializer */
      initlist = tree_cons (NULLT, build_int_2 (0, 0), NULLT);

      e = build_int_2 (size, 0);      
      TREE_TYPE (e) = cast_type; 
      initlist = tree_cons (NULLT, e, initlist);
      
      for (lproto = plist; lproto; lproto = TREE_CHAIN(lproto))
        {
        int offset;
        tree pval = TREE_VALUE(lproto);
	 
	if ((TREE_CODE(pval) == PROTOCOL_INTERFACE) && 
             PROTOCOL_FORWARD_DECL(pval))
            {
            e = build_unary_op (ADDR_EXPR, PROTOCOL_FORWARD_DECL(pval), 0);
            initlist = tree_cons (NULLT, e, initlist);
            }
        }

      /* static struct objc_protocol *refs[n]; */

      sc_spec = tree_cons (NULLT, ridpointers[(int) RID_STATIC], NULLT);
      decl_specs = tree_cons (NULLT, xref_tag (RECORD_TYPE,
			  	get_identifier (_TAG_PROTOCOL)), sc_spec);
				
      if (TREE_CODE(i_or_p) == PROTOCOL_INTERFACE)
        expr_decl = build_nt (ARRAY_REF, 
      			synth_id_with_class_suffix ("_OBJC_PROTOCOL_REFS", i_or_p),
		   	build_int_2 (size+2, 0));
      else if (TREE_CODE(i_or_p) == CLASS_INTERFACE)
        expr_decl = build_nt (ARRAY_REF, 
      			synth_id_with_class_suffix ("_OBJC_CLASS_PROTOCOLS", i_or_p),
		   	build_int_2 (size+2, 0));
      else if (TREE_CODE(i_or_p) == CATEGORY_INTERFACE)
        expr_decl = build_nt (ARRAY_REF, 
      			synth_id_with_class_suffix ("_OBJC_CATEGORY_PROTOCOLS", i_or_p),
		   	build_int_2 (size+2, 0));
			
      expr_decl = build1 (INDIRECT_REF, NULLT, expr_decl);

      refs_decl = start_decl (expr_decl, decl_specs, 1);

      finish_decl (refs_decl, build_nt (CONSTRUCTOR, NULLT, 
      				nreverse (initlist)), NULLT);

      return refs_decl;
}

static tree
build_category_initializer (cat_name, class_name,
			    instance_methods, class_methods, protocol_list)
     tree cat_name;
     tree class_name;
     tree instance_methods;
     tree class_methods;
     tree protocol_list;
{
  tree initlist = NULLT, expr;

  initlist = tree_cons (NULLT, cat_name, initlist);
  initlist = tree_cons (NULLT, class_name, initlist);

  if (!instance_methods)
    initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);
  else
    {
      expr = build_unary_op (ADDR_EXPR, instance_methods, 0);
      initlist = tree_cons (NULLT, expr, initlist);
    }
  if (!class_methods)
    initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);
  else
    {
      expr = build_unary_op (ADDR_EXPR, class_methods, 0);
      initlist = tree_cons (NULLT, expr, initlist);
    }

  /* protocol_list = */
  if (!protocol_list)
     initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);
  else
     {
	static tree cast_type2;
	
	if (!cast_type2)
	  cast_type2 = groktypename (build_tree_list(
			  build_tree_list (NULLT, xref_tag (RECORD_TYPE,
					    get_identifier (_TAG_PROTOCOL))), 
			  build1 (INDIRECT_REF, NULLT, 
			    build1 (INDIRECT_REF, NULLT, NULLT))));
    
	expr = build_unary_op (ADDR_EXPR, protocol_list, 0);
	TREE_TYPE(expr) = cast_type2;
	initlist = tree_cons (NULLT, expr, initlist);
     }

  return build_nt (CONSTRUCTOR, NULLT, nreverse (initlist));
}

/*
 *  struct objc_class {
 *    struct objc_class *isa;
 *    struct objc_class *super_class;
 *    char *name;
 *    long version;
 *    long info;
 *    long instance_size;
 *    struct objc_ivar_list *ivars;
 *    struct objc_method_list *methods;
 *    struct objc_cache *cache;
 *    struct objc_protocol_list *protocols;
 *  };
 */
static tree
build_shared_structure_initializer (isa, super, name, size, status,
				    dispatch_table, ivar_list, protocol_list)
     tree isa;
     tree super;
     tree name;
     tree size;
     int status;
     tree dispatch_table;
     tree ivar_list;
     tree protocol_list;
{
  tree initlist = NULLT, expr;

  /* isa = */
  initlist = tree_cons (NULLT, isa, initlist);

  /* super_class = */
  initlist = tree_cons (NULLT, super, initlist);

  /* name = */
  initlist = tree_cons (NULLT, name, initlist);

  /* version = */
  initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);

  /* info = */
  initlist = tree_cons (NULLT, build_int_2 (status), initlist);

  /* instance_size = */
  initlist = tree_cons (NULLT, size, initlist);

  /* objc_ivar_list = */
  if (!ivar_list)
    initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);
  else
    {
      expr = build_unary_op (ADDR_EXPR, ivar_list, 0);
      initlist = tree_cons (NULLT, expr, initlist);
    }

  /* objc_method_list = */
  if (!dispatch_table)
    initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);
  else
    {
      expr = build_unary_op (ADDR_EXPR, dispatch_table, 0);
      initlist = tree_cons (NULLT, expr, initlist);
    }

  /* method_cache = */
  initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);

  /* protocol_list = */
  if (!protocol_list)
     initlist = tree_cons (NULLT, build_int_2 (0, 0), initlist);
  else
     {
     static tree cast_type2;
     
     if (!cast_type2)
        cast_type2 = groktypename (build_tree_list(
			build_tree_list (NULLT,	xref_tag (RECORD_TYPE,
					  get_identifier (_TAG_PROTOCOL))), 
			build1 (INDIRECT_REF, NULLT, 
			  build1 (INDIRECT_REF, NULLT, NULLT))));
 
     expr = build_unary_op (ADDR_EXPR, protocol_list, 0);
     TREE_TYPE(expr) = cast_type2;
     initlist = tree_cons (NULLT, expr, initlist);
     }

  return build_nt (CONSTRUCTOR, NULLT, nreverse (initlist));
}

/*
 * static struct objc_category _OBJC_CATEGORY_<name> = { ... };
 */
static void
generate_category (cat)
     tree cat;
{
  tree sc_spec, decl_specs, decl;
  tree initlist, cat_name_expr, class_name_expr;
  tree protocol_decl, category;

  cat_name_expr = add_objc_string (CLASS_SUPER_NAME (cat), class_names);

  add_class_reference (CLASS_NAME (cat));
  class_name_expr = add_objc_string (CLASS_NAME (cat), class_names);

  category = CLASS_CATEGORY_LIST (implementation_template);
  
  /* find the category interface from the class it is associated with */
  while (category)
    {
      if (CLASS_SUPER_NAME (cat) == CLASS_SUPER_NAME (category))
	break;
      category = CLASS_CATEGORY_LIST (category);
    }

  if (category && CLASS_PROTOCOL_LIST (category))
    {
      generate_protocol_references (CLASS_PROTOCOL_LIST (category));
      protocol_decl = generate_protocol_list (category);
    }
  else
    protocol_decl = 0;

  sc_spec = tree_cons (NULLT, ridpointers[(int) RID_STATIC], NULLT);
  decl_specs = tree_cons (NULLT, objc_category_template, sc_spec);

  decl = start_decl (
      		synth_id_with_class_suffix ("_OBJC_CATEGORY", implementation_context),
		decl_specs, 1);

  initlist = build_category_initializer (
					 cat_name_expr, class_name_expr,
					 _OBJC_INSTANCE_METHODS_decl,
					 _OBJC_CLASS_METHODS_decl,
					 protocol_decl);

  finish_decl (decl, initlist, NULLT);
}

/*
 * static struct objc_class _OBJC_METACLASS_Foo={ ... };
 * static struct objc_class _OBJC_CLASS_Foo={ ... };
 */
static void
generate_shared_structures ()
{
  tree sc_spec, decl_specs, expr_decl, decl;
  tree name_expr, super_expr, root_expr;
  tree my_root_id = NULLT, my_super_id = NULLT;
  tree cast_type, initlist, protocol_decl;
  int offset;

  my_super_id = CLASS_SUPER_NAME (implementation_template);
  if (my_super_id)
    {
      add_class_reference (my_super_id);

      /* compute "my_root_id" - this is required for code generation.
       * the "isa" for all meta class structures points to the root of
       * the inheritance hierarchy (e.g. "__Object")...
       */
      my_root_id = my_super_id;
      do
	{
	  tree my_root_int = lookup_interface (my_root_id);

	  if (my_root_int && CLASS_SUPER_NAME (my_root_int))
	    my_root_id = CLASS_SUPER_NAME (my_root_int);
	  else
	    break;
	}
      while (1);
    }
  else				/* no super class */
    {
      my_root_id = CLASS_NAME (implementation_template);
    }

  cast_type = groktypename (build_tree_list (build_tree_list (NULLT,
							      objc_class_template), build1 (INDIRECT_REF, NULLT, NULLT)));

  name_expr = add_objc_string (CLASS_NAME (implementation_template), class_names);

  /* install class `isa' and `super' pointers at runtime */
  if (my_super_id)
    {
      super_expr = add_objc_string (my_super_id, class_names);
      super_expr = build_c_cast (cast_type, super_expr); /* cast! */
    }
  else
    super_expr = build_int_2 (0, 0);

  root_expr = add_objc_string (my_root_id, class_names);
  root_expr = build_c_cast (cast_type, root_expr); /* cast! */
  
  if (CLASS_PROTOCOL_LIST(implementation_template)) {
    generate_protocol_references(CLASS_PROTOCOL_LIST(implementation_template));
    protocol_decl = generate_protocol_list (implementation_template);
  } else
    protocol_decl = 0;
    
  /* static struct objc_class _OBJC_METACLASS_Foo = { ... }; */

  sc_spec = build_tree_list (NULLT, ridpointers[(int) RID_STATIC]);
//  sc_spec = tree_cons (NULLT, ridpointers[(int) RID_CONST], sc_spec);
  decl_specs = tree_cons (NULLT, objc_class_template, sc_spec);

  decl = start_decl (DECL_NAME (_OBJC_METACLASS_decl), decl_specs, 1);

  initlist = build_shared_structure_initializer (root_expr, super_expr, name_expr,
						 build_int_2 (TREE_INT_CST_LOW (TYPE_SIZE (objc_class_template)) / BITS_PER_UNIT, 0),
						 2 /*CLS_META*/,
						 _OBJC_CLASS_METHODS_decl,
						 _OBJC_CLASS_VARIABLES_decl, 
						 protocol_decl);
  finish_decl (decl, initlist, NULLT);

  /* static struct objc_class _OBJC_CLASS_Foo={ ... }; */

  decl = start_decl (DECL_NAME (_OBJC_CLASS_decl), decl_specs, 1);

  initlist = build_shared_structure_initializer (
						 build_unary_op (ADDR_EXPR, _OBJC_METACLASS_decl, 0),
						 super_expr, name_expr,
						 build_int_2 (TREE_INT_CST_LOW (TYPE_SIZE (CLASS_STATIC_TEMPLATE (implementation_template))) / BITS_PER_UNIT, 0),
						 1 /*CLS_FACTORY*/,
						 _OBJC_INSTANCE_METHODS_decl,
						 _OBJC_INSTANCE_VARIABLES_decl,
						 protocol_decl);

  finish_decl (decl, initlist, NULLT);
}

static tree
synth_id_with_class_suffix (preamble, ctxt)
     char *preamble;
     tree ctxt;
{
  if (TREE_CODE (ctxt) == CLASS_IMPLEMENTATION ||
      TREE_CODE (ctxt) == CLASS_INTERFACE)
    sprintf (utlbuf, "%s_%s", preamble,
	     IDENTIFIER_POINTER (CLASS_NAME (ctxt)));

  else if (TREE_CODE (ctxt) == CATEGORY_IMPLEMENTATION ||
	   TREE_CODE (ctxt) == CATEGORY_INTERFACE) 
    sprintf (utlbuf, "%s_%s_%s", preamble,
	     IDENTIFIER_POINTER (CLASS_NAME (ctxt)),
	     IDENTIFIER_POINTER (CLASS_SUPER_NAME (ctxt)));

  else if (TREE_CODE (ctxt) == PROTOCOL_INTERFACE)
    sprintf (utlbuf, "%s_%s", preamble,
	     IDENTIFIER_POINTER (PROTOCOL_NAME (ctxt)));

  return get_identifier (utlbuf);
}

static int
is_objc_type_qualifier (node)
     tree node;
{
  return (TREE_CODE (node) == IDENTIFIER_NODE
	  && (node == ridpointers [(int) RID_CONST]
	      || node == ridpointers [(int) RID_VOLATILE]
	      || node == ridpointers [(int) RID_IN]
	      || node == ridpointers [(int) RID_OUT]
	      || node == ridpointers [(int) RID_INOUT]
	      || node == ridpointers [(int) RID_BYCOPY]
	      || node == ridpointers [(int) RID_ONEWAY]));
}

/* If type is empty or only type qualifiers are present, add default
   type of id (otherwise grokdeclarator will default to int).  */

static tree
adjust_type_for_id_default (type)
     tree type;
{
  tree declspecs, chain;
  int type_spec_seen = 0;

  if (!type)
    return build_tree_list (build_tree_list (NULLT, objc_object_reference),
			    build1 (INDIRECT_REF, NULLT, NULLT));

  declspecs = TREE_PURPOSE (type);

  /* Determine if a typespec is present.  */
  for (chain = declspecs;
       chain;
       chain = TREE_CHAIN (chain))
    {
      if (!is_objc_type_qualifier (TREE_VALUE (chain)))
	return type;
    }

  return build_tree_list (tree_cons (NULLT, objc_object_reference, declspecs),
			  build1 (INDIRECT_REF, NULLT, NULLT));
}

/*
 *   usage:
 *		keyworddecl:
 *			selector ':' '(' typename ')' identifier
 *
 *   purpose:
 *		transform an Objective-C keyword argument into
 *		the C equivalent parameter declarator.
 *
 *   in:	key_name, an "identifier_node" (optional).
 *		arg_type, a  "tree_list" (optional).
 *		arg_name, an "identifier_node".
 *
 *   note:	it would be really nice to strongly type the preceding
 *		arguments in the function prototype; however, then i
 *		could not use the "accessor" macros defined in "tree.h".
 *
 *   out:	an instance of "keyword_decl".
 *
 */

tree
build_keyword_decl (key_name, arg_type, arg_name)
     tree key_name;
     tree arg_type;
     tree arg_name;
{
  tree keyword_decl;

  /* if no type is specified, default to "id" */
  arg_type = adjust_type_for_id_default (arg_type);

  keyword_decl = make_node (KEYWORD_DECL);

  TREE_TYPE (keyword_decl) = arg_type;
  KEYWORD_ARG_NAME (keyword_decl) = arg_name;
  KEYWORD_KEY_NAME (keyword_decl) = key_name;

  return keyword_decl;
}

/*
 *  given a chain of keyword_decl's, synthesize the full keyword selector.
 */
static tree
build_keyword_selector (selector)
     tree selector;
{
  int len = 0;
  tree key_chain, key_name;
  char *buf;

  for (key_chain = selector; key_chain; key_chain = TREE_CHAIN (key_chain))
    {
      if (TREE_CODE (selector) == KEYWORD_DECL)
	key_name = KEYWORD_KEY_NAME (key_chain);
      else if (TREE_CODE (selector) == TREE_LIST)
	key_name = TREE_PURPOSE (key_chain);

      if (key_name)
	len += IDENTIFIER_LENGTH (key_name) + 1;
      else			/* just a ':' arg */
	len++;
    }
  buf = (char *)alloca (len + 1);
  bzero (buf, len + 1);

  for (key_chain = selector; key_chain; key_chain = TREE_CHAIN (key_chain))
    {
      if (TREE_CODE (selector) == KEYWORD_DECL)
	key_name = KEYWORD_KEY_NAME (key_chain);
      else if (TREE_CODE (selector) == TREE_LIST)
	key_name = TREE_PURPOSE (key_chain);

      if (key_name)
	strcat (buf, IDENTIFIER_POINTER (key_name));
      strcat (buf, ":");
    }
  return get_identifier (buf);
}

/* used for declarations and definitions */

tree
build_method_decl (code, ret_type, selector, add_args)
     enum tree_code code;
     tree ret_type;
     tree selector;
     tree add_args;
{
  tree method_decl;

  /* if no type is specified, default to "id" */
  ret_type = adjust_type_for_id_default (ret_type);

  method_decl = make_node (code);
  TREE_TYPE (method_decl) = ret_type;

  METHOD_FILENAME(method_decl) = input_filename;
  METHOD_LINENUM(method_decl) = lineno;

  /*
   *  if we have a keyword selector, create an identifier_node that
   *  represents the full selector name (`:' included)...
   */
  if (TREE_CODE (selector) == KEYWORD_DECL)
    {
      METHOD_SEL_NAME (method_decl) = build_keyword_selector (selector);
      METHOD_SEL_ARGS (method_decl) = selector;
      METHOD_ADD_ARGS (method_decl) = add_args;
    }
  else
    {
      METHOD_SEL_NAME (method_decl) = selector;
      METHOD_SEL_ARGS (method_decl) = NULLT;
      METHOD_ADD_ARGS (method_decl) = NULLT;
    }

  return method_decl;
}

#define METHOD_DEF 0
#define METHOD_REF 1
/*
 * used by `build_message_expr ()' and `comp_method_types ()'.
 *
 * add_args is a tree_list node the following info on a parameter list:
 *
 *    The TREE_PURPOSE is a chain of decls of those parms.
 *    The TREE_VALUE is a list of structure, union and enum tags defined.
 *    The TREE_CHAIN is a list of argument types to go in the FUNCTION_TYPE.
 *    This tree_list node is later fed to `grokparms'.
 *
 *    VOID_AT_END nonzero means append `void' to the end of the type-list.
 *    Zero means the parmlist ended with an ellipsis so don't append `void'.
 */
static tree
get_arg_type_list (meth, context, superflag)
     tree meth;
     int context;
     int superflag;
{
  tree arglist, akey;

  /* receiver type */
  if (superflag)
    arglist = build_tree_list (NULLT, super_type);
  else
    {
      if (context == METHOD_DEF)
	arglist = build_tree_list (NULLT, TREE_TYPE (self_decl));
      else
	arglist = build_tree_list (NULLT, id_type);
    }

  /* selector type - will eventually change to `int' */
  chainon (arglist, build_tree_list (NULLT, _selector_type));

  /* build a list of argument types */
  for (akey = METHOD_SEL_ARGS (meth); akey; akey = TREE_CHAIN (akey))
    {
      tree arg_decl = groktypename_in_parm_context (TREE_TYPE (akey));
      chainon (arglist, build_tree_list (NULLT, TREE_TYPE (arg_decl)));
    }

  if (METHOD_ADD_ARGS (meth) == (tree)1)
    /*
     * we have a `, ...' immediately following the selector,
     * finalize the arglist...simulate get_parm_info (0)
     */
    ;
  else if (METHOD_ADD_ARGS (meth))
    {
      /* we have a variable length selector */
      tree add_arg_list = TREE_CHAIN (METHOD_ADD_ARGS (meth));
      chainon (arglist, add_arg_list);
    }
  else
    /* Finalize the arglist...simulate get_parm_info (1).  */
#ifdef OBJCPLUS
    chainon (arglist, void_list_node);
#else /* OBJCPLUS */
    chainon (arglist, build_tree_list (NULLT, void_type_node));
#endif /* OBJCPLUS */

  return arglist;
}

static tree
check_duplicates (hsh)
     hash hsh;
{
  tree meth = NULLT;

  if (hsh)
    {
      meth = hsh->key;

      if (hsh->list)
        {
	  /* we have two methods with the same name and different types */
	  attr loop;
	  char type;

	  type = (TREE_CODE (meth) == INSTANCE_METHOD_DECL) ? '-' : '+';

	  warning ("multiple declarations for method `%s'",
		   IDENTIFIER_POINTER (METHOD_SEL_NAME (meth)));

	  warn_with_method ("using", type, meth);
	  for (loop = hsh->list; loop; loop = loop->next)
	    warn_with_method ("also found", type, loop->value);
        }
    }
  return meth;
}

static tree
receiver_is_class_object (receiver)
      tree receiver;
{
  /* the receiver is a function call that returns an id...
   * ...check if it is a call to objc_getClass, if so, give it
   * special treatment.
   */
  tree exp = 0;

  if ((exp = TREE_OPERAND (receiver, 0)) && (TREE_CODE (exp) == ADDR_EXPR))
    {
      if ((exp = TREE_OPERAND (exp, 0)) && 
	  (TREE_CODE (exp) == FUNCTION_DECL) && exp == objc_getClass_decl)
	{
	  /* we have a call to objc_getClass! */
	  tree arg = 0;
    
	  if ((arg = TREE_OPERAND (receiver, 1)) &&
	      (TREE_CODE (arg) == TREE_LIST) &&
	      (arg = TREE_VALUE (arg)) &&
	      (TREE_CODE (arg) == NOP_EXPR) &&
	      (arg = TREE_OPERAND (arg, 0)) &&
	      (TREE_CODE (arg) == ADDR_EXPR) &&
	      (arg = TREE_OPERAND (arg, 0)) &&
	      (TREE_CODE (arg) == STRING_CST))
	    /* finally, we have the class name */
	    return get_identifier (TREE_STRING_POINTER (arg));
	}
    }
  return 0;
}

/* If we are currently building a message expr, this holds
   the identifier of the selector of the message.  This is
   used when printing warnings about argument mismatches. */


static tree building_objc_message_expr = 0;


tree
maybe_building_objc_message_expr ()
{
  return building_objc_message_expr;
}


/*
 *	(*(<abstract_decl>(*)())_msg)(receiver, selTransTbl[n], ...);
 *	(*(<abstract_decl>(*)())_msgSuper)(receiver, selTransTbl[n], ...);
 */
tree
build_message_expr (mess)
     tree mess;
{
  tree receiver = TREE_PURPOSE (mess), rtype, sel_name;
  tree args = TREE_VALUE (mess);
  tree params = NULLT;
  tree meth = NULLT, message_decl;
  tree retval;
  int statically_typed = 0, statically_allocated = 0;
  tree class_ident = 0;

  if (!doing_objc_thang)
    objc_fatal ();

  if (TREE_CODE (receiver) == ERROR_MARK)
    return error_mark_node;

  /* determine receiver type */
  rtype = TREE_TYPE (receiver);

  if (rtype == super_type)
    message_decl = _msgSuper_decl;
  else
    {
      message_decl = _msg_decl;

      if (TREE_STATIC_TEMPLATE (rtype))
	statically_allocated = 1;
      else if (TREE_CODE (rtype) == POINTER_TYPE
	       && TREE_STATIC_TEMPLATE (TREE_TYPE (rtype)))
	statically_typed = 1;
      /* classfix -smn */
      else if ((TREE_CODE (receiver) == CALL_EXPR) && (rtype == id_type) &&
		(class_ident = receiver_is_class_object (receiver)))
	;
      else if ((TYPE_MAIN_VARIANT (rtype) != id_type)
	       && (rtype != objc_class_type))
	{
	  bzero (errbuf, BUFSIZE);
	  error ("illegal receiver type `%s'",
		 gen_declaration (rtype, errbuf));
	  return error_mark_node;
	}
    }

  /* obtain the full selector name */
  if (TREE_CODE (args) == IDENTIFIER_NODE)
    /* a unary selector */
    sel_name = args;
  else if (TREE_CODE (args) == TREE_LIST)
    sel_name = build_keyword_selector (args);

  /* start building argument list - synthesizing the second argument */
  if (statically_allocated)
    params = build_tree_list (NULLT, build_unary_op (ADDR_EXPR, receiver, 0));
  else
    params = build_tree_list (NULLT, receiver);

  chainon (params, build_tree_list (NULLT,
	   build_selector_reference (sel_name)));
  if (TREE_CODE (args) == TREE_LIST)
    {
      tree chain = args, prev = NULLT;

      /* we have a keyword selector - check for comma expressions */
      while (chain)
	{
	  tree element = TREE_VALUE (chain);

	  /* we have a comma expression, must collapse... */
	  if (TREE_CODE (element) == TREE_LIST)
	    {
	      if (prev)
		TREE_CHAIN (prev) = element;
	      else
		args = element;
	    }
	  prev = chain;
	  chain = TREE_CHAIN (chain);
        }
      /* attach the rest of the argument list */
      chainon (params, args);
    }

  /* determine return type */

  if (rtype == super_type)
    {
      tree iface;

      if (CLASS_SUPER_NAME (implementation_template))
	{
	  iface = lookup_interface (CLASS_SUPER_NAME (implementation_template));
    
	  if (TREE_CODE (method_context) == INSTANCE_METHOD_DECL)
	    meth = lookup_instance_method_static (iface, sel_name);
	  else
	    meth = lookup_class_method_static (iface, sel_name);
    
	  if (iface && !meth)
	    {
	      warning ("`%s' does not respond to `%s'",
			IDENTIFIER_POINTER (CLASS_SUPER_NAME (implementation_template)),
			IDENTIFIER_POINTER (sel_name));
	    }
	}
      else
	{
	  error ("no super class declared in interface for `%s'",
		 IDENTIFIER_POINTER (CLASS_NAME (implementation_template)));
	  return error_mark_node;
	}

    }
  else if (statically_typed || statically_allocated)
    {
      tree ctype;

      if (statically_allocated)
        ctype = rtype;
      if (statically_typed)
        ctype = TREE_TYPE (rtype);

      /* `self' is now statically_typed...all methods should be visible
       * within the context of the implementation...
       */
      if ((implementation_context
	   && CLASS_NAME (implementation_context) == TYPE_NAME (ctype)))
	{
	  meth = lookup_instance_method_static (implementation_template, sel_name);

/* NEW!!! */
          if (!meth && TYPE_PROTOCOL_LIST (ctype))
	    meth = lookup_method_in_protocol_list (
			TYPE_PROTOCOL_LIST (ctype), sel_name, 0);

	  if (!meth && (implementation_template != implementation_context))
	    /* the method is not published in the interface...check locally */
	    meth = lookup_method (CLASS_NST_METHODS (implementation_context),
				  sel_name);
	}
      else
	{
	  tree iface = lookup_interface (TYPE_NAME (ctype));

	  if (iface)
	    meth = lookup_instance_method_static (iface, sel_name);
	  else
	    error ("Cannot find interface declaration for `%s'",
		   IDENTIFIER_POINTER (TYPE_NAME (ctype)));

          if (!meth && TYPE_PROTOCOL_LIST (ctype))
	    meth = lookup_method_in_protocol_list (
			TYPE_PROTOCOL_LIST (ctype), sel_name, 0);
	}

      if (!meth)
        warning ("`%s' does not respond to `%s'",
		 IDENTIFIER_POINTER (TYPE_NAME (ctype)),
		 IDENTIFIER_POINTER (sel_name));
    }
  else if (class_ident)
    {
      if ((implementation_context && 
	   CLASS_NAME (implementation_context) == class_ident))
	{
	  meth = lookup_class_method_static (implementation_template,
					     sel_name);
    
	  if (!meth && (implementation_template != implementation_context))
	    /* the method is not published in the interface...check locally */
	    meth = lookup_method (CLASS_CLS_METHODS (implementation_context),
				  sel_name);
	}
      else
	{
	  tree iface;
    
	  if (iface = lookup_interface (class_ident))
	    meth = lookup_class_method_static (iface, sel_name);
	}
  
      if (!meth)
	{
	  warning ("cannot find class (factory) method.");
	  warning ("return type for `%s' defaults to id",
		   IDENTIFIER_POINTER (sel_name));
	}
    }
  else if (rtype != id_type && TYPE_MAIN_VARIANT (rtype) == id_type)
    {
      /* An anonymous object that has been qualified with a protocol.  */
  
      tree protocol_list = TYPE_PROTOCOL_LIST (rtype);
      
      meth = lookup_method_in_protocol_list (protocol_list, sel_name, 0);

      if (!meth)
	{
          hash hsh;

	  warning ("method `%s' not implemented by protocol.",
		   IDENTIFIER_POINTER (sel_name));

          /* try and find the method signiture in the global pools! */

          if (!(hsh = hash_lookup (nst_method_hash_list, sel_name)))
	    hsh = hash_lookup (cls_method_hash_list, sel_name);

          if (!(meth = check_duplicates (hsh)))
	    warning ("return type defaults to id");
	}
    }
  else
    {
      hash hsh;

      /* we think we have an instance...loophole: extern id Object; */
      if (!(hsh = hash_lookup (nst_method_hash_list, sel_name)))
	{
	  /* for various loopholes...like sending messages to self in a
	   * factory context...
	   */
	  hsh = hash_lookup (cls_method_hash_list, sel_name);
	}
      if (!(meth = check_duplicates (hsh)))
	{
	  warning ("cannot find method.");
	  warning ("return type for `%s' defaults to id",
		   IDENTIFIER_POINTER (sel_name));
	}
    }

  building_objc_message_expr = sel_name;

  if (!meth)
    {
      retval = build_function_call (message_decl, params);
    }
  else				/* we have a method prototype */
    {
      tree arglist = NULLT;
      tree savret, savarg;

      arglist = get_arg_type_list (meth, METHOD_REF, message_decl == _msgSuper_decl);

      /* install argument types - normally set by "build_function_type ()". */
      savarg = TYPE_ARG_TYPES (TREE_TYPE (message_decl));
      TYPE_ARG_TYPES (TREE_TYPE (message_decl)) = arglist;

      /* install return type - "message_decl" is a function returning ret_type. */
      savret = TREE_TYPE (TREE_TYPE (message_decl));
      TREE_TYPE (TREE_TYPE (message_decl)) = groktypename (TREE_TYPE (meth));

#ifdef OBJCPLUS
      DECL_PRINT_NAME (message_decl) = 0;
#endif /* OBJCPLUS */

      /* check argments types supplied vs. ones in method prototype.
       * build_function_call () calls convert_arguments () to do the
       * type checking
       */
      retval = build_function_call (message_decl, params);

      /* restore previous return/argument types */
      TYPE_ARG_TYPES (TREE_TYPE (message_decl)) = savarg;
      TREE_TYPE (TREE_TYPE (message_decl)) = savret;
    }

  building_objc_message_expr = 0;

  return retval;
}

static void
build_protocol_reference (p)
     tree p;
{
  tree ref, decl, ident, ptype;
  struct obstack *save_current_obstack = current_obstack;
  struct obstack *save_rtl_obstack = rtl_obstack;

  rtl_obstack = current_obstack = &permanent_obstack;
  
  /* extern struct objc_protocol _OBJC_PROTOCOL_<mumble>; */

  ident = synth_id_with_class_suffix ("_OBJC_PROTOCOL", p);
  ptype = groktypename (build_tree_list(
                         build_tree_list(NULLT, objc_protocol_template), NULLT));

  if (IDENTIFIER_GLOBAL_VALUE (ident))
    decl = IDENTIFIER_GLOBAL_VALUE (ident); /* set by pushdecl() */
  else 
    {
      decl = build_decl (VAR_DECL, ident, ptype);
      TREE_EXTERNAL (decl) = 1;
      TREE_PUBLIC (decl) = 1;
      TREE_USED (decl) = 1;

      /* usually called from `rest_of_decl_compilation' */
      make_decl_rtl (decl, 0, 1); 
      /* our `extended/custom' pushdecl in c-decl.c */
      pushdecl_top_level (decl);  
    }
  current_obstack = save_current_obstack;
  rtl_obstack = save_rtl_obstack;

  PROTOCOL_FORWARD_DECL (p) = decl;
}

tree
build_protocol_expr (protoname)
     tree protoname;
{
  tree expr;
  tree p;

  if (!doing_objc_thang)
    objc_fatal ();

  p = lookup_protocol (protoname);

  if (!p) {
    error ("Cannot find protocol declaration for `%s'",
	   IDENTIFIER_POINTER (protoname));
    return error_mark_node;
  }

  if (!PROTOCOL_FORWARD_DECL (p)) 
    build_protocol_reference(p);

  expr = build_unary_op (ADDR_EXPR, PROTOCOL_FORWARD_DECL (p), 0);

  TREE_TYPE(expr) = protocol_type;

  return expr;
}

tree
build_selector_expr (selnamelist)
     tree selnamelist;
{
  tree selname;

  if (!doing_objc_thang)
    objc_fatal ();

  /* obtain the full selector name */
  if (TREE_CODE (selnamelist) == IDENTIFIER_NODE)
    /* a unary selector */
    selname = selnamelist;
  else if (TREE_CODE (selnamelist) == TREE_LIST)
    selname = build_keyword_selector (selnamelist);

  return build_selector_reference (selname);
}

tree
build_encode_expr (type)
     tree type;
{
  if (!doing_objc_thang)
    objc_fatal ();

  if (!utlbuf)
    utlbuf = (char *)xmalloc (BUFSIZE);
  bzero (utlbuf, BUFSIZE);

  encode_type (type, utlbuf, utlbuf + strlen (utlbuf),
	       OBJC_ENCODE_INLINE_DEFS);

  /* synthesize a string that represents the encoded struct/union */
  return my_build_string (strlen (utlbuf) + 1, utlbuf);
}

tree
build_ivar_reference (id)
     tree id;
{
  if (TREE_CODE (method_context) == CLASS_METHOD_DECL)
    TREE_TYPE (self_decl) = instance_type; /* cast */

  return build_component_ref (build_indirect_ref (self_decl, "->"), id);
}

#define HASH_ALLOC_LIST_SIZE	170
#define ATTR_ALLOC_LIST_SIZE	170
#define SIZEHASHTABLE 		257
#define HASHFUNCTION(key)	((int)key >> 2) 	/* divide by 4 */

static void
hash_init ()
{
  nst_method_hash_list = (hash *)xmalloc (SIZEHASHTABLE * sizeof (hash));
  cls_method_hash_list = (hash *)xmalloc (SIZEHASHTABLE * sizeof (hash));

  if (!nst_method_hash_list || !cls_method_hash_list)
    perror ("unable to allocate space in objc-tree.c");
  else
    {
      int i;

      for (i = 0; i < SIZEHASHTABLE; i++)
	{
	  nst_method_hash_list[i] = 0;
	  cls_method_hash_list[i] = 0;
	}
    }
}

static void
hash_enter (hashlist, method)
     hash *hashlist;
     tree method;
{
  static hash 	hash_alloc_list = 0;
  static int	hash_alloc_index = 0;
  hash obj;
  int slot = HASHFUNCTION (METHOD_SEL_NAME (method)) % SIZEHASHTABLE;

  if (!hash_alloc_list || hash_alloc_index >= HASH_ALLOC_LIST_SIZE)
    {
      hash_alloc_index = 0;
      hash_alloc_list = (hash)xmalloc (sizeof (struct hashedEntry) *
				      HASH_ALLOC_LIST_SIZE);
      if (!hash_alloc_list)
	perror ("unable to allocate in objc-tree.c");
    }
  obj = &hash_alloc_list[hash_alloc_index++];
  obj->list = 0;
  obj->next = hashlist[slot];
  obj->key = method;

  hashlist[slot] = obj;		/* append to front */
}

static hash
hash_lookup (hashlist, sel_name)
     hash *hashlist;
     tree sel_name;
{
  hash target;

  target = hashlist[HASHFUNCTION (sel_name) % SIZEHASHTABLE];

  while (target)
    {
      if (sel_name == METHOD_SEL_NAME (target->key))
	return target;

      target = target->next;
    }
  return 0;
}

static void
hash_add_attr (entry, value)
     hash entry;
     tree value;
{
  static attr 	attr_alloc_list = 0;
  static int	attr_alloc_index = 0;
  attr obj;

  if (!attr_alloc_list || attr_alloc_index >= ATTR_ALLOC_LIST_SIZE)
    {
      attr_alloc_index = 0;
      attr_alloc_list = (attr)xmalloc (sizeof (struct hashedAttribute) *
				      ATTR_ALLOC_LIST_SIZE);
      if (!attr_alloc_list)
	perror ("unable to allocate in objc-tree.c");
    }
  obj = &attr_alloc_list[attr_alloc_index++];
  obj->next = entry->list;
  obj->value = value;

  entry->list = obj;		/* append to front */
}

static tree
lookup_method (mchain, method)
     tree mchain;
     tree method;
{
  tree key;

  if (TREE_CODE (method) == IDENTIFIER_NODE)
    key = method;
  else
    key = METHOD_SEL_NAME (method);

  while (mchain)
    {
      if (METHOD_SEL_NAME (mchain) == key)
	return mchain;
      mchain = TREE_CHAIN (mchain);
    }
  return NULLT;
}

static tree
lookup_instance_method_static (interface, ident)
     tree interface;
     tree ident;
{
  tree inter = interface;
  tree meth = NULLT;

  while (inter)
    {
      if (meth = lookup_method (CLASS_NST_METHODS (inter), ident))
	return meth;

      if (CLASS_CATEGORY_LIST (inter))
	{
	  tree category = CLASS_CATEGORY_LIST (inter);

	  while (category)
	    {
	      if (meth = lookup_method (CLASS_NST_METHODS (category), ident))
		return meth;

/* NEW!!! */
	      /* Check for instance methods in protocols in categories.  */
	      if (CLASS_PROTOCOL_LIST (category))
		{
		  if (meth = lookup_method_in_protocol_list (
				CLASS_PROTOCOL_LIST (category), ident, 0))
		  return meth;
		}
      
	      category = CLASS_CATEGORY_LIST (category);
	    }
	}

      if (CLASS_PROTOCOL_LIST (inter))
        {
	  if (meth = lookup_method_in_protocol_list (
			CLASS_PROTOCOL_LIST (inter), ident, 0))
          return meth;
        }

      inter = lookup_interface (CLASS_SUPER_NAME (inter));
    }

  return meth;
}

static tree
lookup_class_method_static (interface, ident)
     tree interface;
     tree ident;
{
  tree inter = interface;
  tree meth = NULLT;
  tree root_inter = NULLT;

  while (inter)
    {
      if (meth = lookup_method (CLASS_CLS_METHODS (inter), ident))
	return meth;

      if (CLASS_CATEGORY_LIST (inter))
	{
	  tree category = CLASS_CATEGORY_LIST (inter);
    
	  while (category)
	    {
	      if (meth = lookup_method (CLASS_CLS_METHODS (category), ident))
		return meth;

/* NEW!!! */
	      /* Check for class methods in protocols in categories.  */
	      if (CLASS_PROTOCOL_LIST (category))
		{
		  if (meth = lookup_method_in_protocol_list (
				CLASS_PROTOCOL_LIST (category), ident, 1))
		  return meth;
		}

	      category = CLASS_CATEGORY_LIST (category);
	    }
	}

/* NEW!!! */
      /* Check for class methods in protocols.  */
      if (CLASS_PROTOCOL_LIST (inter))
        {
	if (meth = lookup_method_in_protocol_list (
			CLASS_PROTOCOL_LIST (inter), ident, 1))
          return meth;
        }

      root_inter = inter;
      inter = lookup_interface (CLASS_SUPER_NAME (inter));
    }

/* NEW!!! */
  /* Simulate wrap around.  */
  return lookup_instance_method_static (root_inter, ident);
}

tree
add_class_method (class, method)
     tree class;
     tree method;
{
  tree mth;
  hash hsh;

  if (!(mth = lookup_method (CLASS_CLS_METHODS (class), method)))
    {
      /* put method on list in reverse order */
      TREE_CHAIN (method) = CLASS_CLS_METHODS (class);
      CLASS_CLS_METHODS (class) = method;
    }
  else
    {
      if (TREE_CODE (class) == CLASS_IMPLEMENTATION)
	error ("duplicate definition of class method `%s'.",
	       IDENTIFIER_POINTER (METHOD_SEL_NAME (mth)));
      else
        {
	  /* check types, if different complain */
	  if (!comp_proto_with_proto (method, mth))
	    error ("duplicate declaration of class method `%s'.",
		   IDENTIFIER_POINTER (METHOD_SEL_NAME (mth)));
        }
    }

  if (!(hsh = hash_lookup (cls_method_hash_list, METHOD_SEL_NAME (method))))
    {
      /* install on a global chain */
      hash_enter (cls_method_hash_list, method);
    }
  else
    {
      /* check types, if different add to a list */
      if (!comp_proto_with_proto (method, hsh->key))
        hash_add_attr (hsh, method);
    }
  return method;
}

tree
add_instance_method (class, method)
     tree class;
     tree method;
{
  tree mth;
  hash hsh;

  if (!(mth = lookup_method (CLASS_NST_METHODS (class), method)))
    {
      /* put method on list in reverse order */
      TREE_CHAIN (method) = CLASS_NST_METHODS (class);
      CLASS_NST_METHODS (class) = method;
    }
  else
    {
      if (TREE_CODE (class) == CLASS_IMPLEMENTATION)
	error ("duplicate definition of instance method `%s'.",
	       IDENTIFIER_POINTER (METHOD_SEL_NAME (mth)));
      else
        {
	  /* check types, if different complain */
	  if (!comp_proto_with_proto (method, mth))
	    error ("duplicate declaration of instance method `%s'.",
		   IDENTIFIER_POINTER (METHOD_SEL_NAME (mth)));
        }
    }

  if (!(hsh = hash_lookup (nst_method_hash_list, METHOD_SEL_NAME (method))))
    {
      /* install on a global chain */
      hash_enter (nst_method_hash_list, method);
    }
  else
    {
      /* check types, if different add to a list */
      if (!comp_proto_with_proto (method, hsh->key))
        hash_add_attr (hsh, method);
    }
  return method;
}

static tree
add_class (class)
     tree class;
{
  /* put interfaces on list in reverse order */
  TREE_CHAIN (class) = interface_chain;
  interface_chain = class;
  return interface_chain;
}

static void
add_category (class, category)
      tree class;
      tree category;
{
  /* put categories on list in reverse order */
  CLASS_CATEGORY_LIST (category) = CLASS_CATEGORY_LIST (class);
  CLASS_CATEGORY_LIST (class) = category;
}

/* called after parsing each instance variable declaration. Necessary to
 * preserve typedefs and implement public/private...
 */

/* PUBLIC is 1 for public, 0 for protected, and 2 for private. */

tree
add_instance_variable (class, public, declarator, declspecs, width)
     tree class;
     int public;
     tree declarator;
     tree declspecs;
     tree width;
{
  tree field_decl, raw_decl;

  raw_decl = build_tree_list (declspecs	/*purpose*/, declarator/*value*/);

  if (CLASS_RAW_IVARS (class))
    chainon (CLASS_RAW_IVARS (class), raw_decl);
  else
    CLASS_RAW_IVARS (class) = raw_decl;

  field_decl = grokfield (input_filename, lineno,
			  declarator, declspecs, width);

  /* overload the public attribute, it is not used for FIELD_DECL's */
  switch (public)
    {
    case 0:
      TREE_PUBLIC (field_decl) = 0;
      TREE_PRIVATE (field_decl) = 0;
      TREE_PROTECTED (field_decl) = 1;
      break;
      
    case 1:
      TREE_PUBLIC (field_decl) = 1;
      TREE_PRIVATE (field_decl) = 0;
      TREE_PROTECTED (field_decl) = 0;
      break;
      
    case 2:
      TREE_PUBLIC (field_decl) = 0;
      TREE_PRIVATE (field_decl) = 1;
      TREE_PROTECTED (field_decl) = 0;
      break;
    }

  if (CLASS_IVARS (class))
    chainon (CLASS_IVARS (class), field_decl);
  else
    CLASS_IVARS (class) = field_decl;

  return class;
}

tree
is_ivar (decl_chain, ident)
     tree decl_chain;
     tree ident;
{
  for ( ; decl_chain; decl_chain = TREE_CHAIN (decl_chain))
    if (DECL_NAME (decl_chain) == ident)
      return decl_chain;
  return NULL_TREE;
}


/* True if the ivar is private and we are not in its implementation.  */

int
is_private (decl)
     tree decl;
{
  if (TREE_PRIVATE (decl)
      && ! is_ivar (CLASS_IVARS (implementation_template), DECL_NAME (decl)))
    {
      error ("instance variable `%s' is declared private",
	     IDENTIFIER_POINTER (DECL_NAME (decl)));
      return 1;
    }
  else
    return 0;
}


/* we have an instance variable reference, check to see if it is public...*/

int
is_public (expr, identifier)
     tree expr;
     tree identifier;
{
  tree basetype = TREE_TYPE (expr);
  enum tree_code code = TREE_CODE (basetype);
  tree decl;

  if (code == RECORD_TYPE)
    {
      if (TREE_STATIC_TEMPLATE (basetype))
	{
	  if (!lookup_interface (TYPE_NAME (basetype)))
	    {
	      error ("Cannot find interface declaration for `%s'",
		     IDENTIFIER_POINTER (TYPE_NAME (basetype)));
	      return 0;
	    }

	  if (decl = is_ivar (TYPE_FIELDS (basetype), identifier))
	    {
	      if (TREE_PUBLIC (decl))
		return 1;

	      /* important diffence between the Stepstone translator:
		 
		 all instance variables should be public within the context
		 of the implementation...
		 */
	      if (implementation_context
		  && (TREE_CODE (implementation_context) == CLASS_IMPLEMENTATION
		      || TREE_CODE (implementation_context) == CATEGORY_IMPLEMENTATION)
		  && CLASS_NAME (implementation_context) == TYPE_NAME (basetype))
		return (!is_private (decl));

	      error ("instance variable `%s' is declared %s",
		     IDENTIFIER_POINTER (identifier),
		     TREE_PRIVATE (decl) ? "private" : "protected");
	      return 0;
	    }
	}
      else if (implementation_context && (basetype == objc_object_reference))
	{
	  TREE_TYPE (expr) = _PRIVATE_record;
	  warning ("static access to object of type `id'");
	  warning ("please change to type `%s *'",
		   IDENTIFIER_POINTER (CLASS_NAME (implementation_context)));
	}
    }
  return 1;
}

/* implement @defs (<classname>) within struct bodies. */

tree
get_class_ivars (interface)
     tree interface;
{
  if (!doing_objc_thang)
    objc_fatal ();

  /* Make sure we copy the leaf ivars in case @defs is used in a local
     context.  Otherwise finish_struct will overwrite the layout info
     using temporary storage.  */
  return build_ivar_chain (interface, 1);
}

tree
get_class_reference (ident)
     tree ident;
{
  tree params;
  tree class_name_expr, initlist, constructor;

  add_class_reference (ident);

#ifdef OBJC_CLASS_REFS
  /* = ((struct _objc_class_ref) {(id) "CLASSNAME"})  */

  class_name_expr = add_objc_string (ident, class_names);
  class_name_expr = build_c_cast (objc_class_type, class_name_expr);

  initlist = build_tree_list (NULLT, class_name_expr);
  constructor = build (CONSTRUCTOR, objc_class_ref_template, NULLT,
		       nreverse (initlist));
  TREE_CONSTANT (constructor) = 1;
  TREE_STATIC (constructor) = 1;

  return build_component_ref (constructor, get_identifier ("class"));
#else /* OBJC_CLASS_REFS */
  params = build_tree_list (NULLT,
			    my_build_string (IDENTIFIER_LENGTH (ident) + 1,
					     IDENTIFIER_POINTER (ident)));

  return build_function_call (objc_getClass_decl, params);
#endif /* OBJC_CLASS_REFS */
}

/* make sure all entries in "chain" are also in "list" */

static int
check_methods (chain, list, mtype)
     tree chain;
     tree list;
     int mtype;
{
  int first = 1;

  while (chain)
    {
      if (!lookup_method (list, chain))
	{
	  if (first)
	    {
	      if (TREE_CODE (implementation_context) == CLASS_IMPLEMENTATION)
		warning ("incomplete implementation of class `%s'",
			 IDENTIFIER_POINTER (CLASS_NAME (implementation_context)));
	      else if (TREE_CODE (implementation_context) == CATEGORY_IMPLEMENTATION)
		warning ("incomplete implementation of category `%s'",
			 IDENTIFIER_POINTER (CLASS_SUPER_NAME (implementation_context)));
	      first = 0;
	    }
	  warning ("method definition for `%c%s' not found",
		   mtype, IDENTIFIER_POINTER (METHOD_SEL_NAME (chain)));
	}
      chain = TREE_CHAIN (chain);
    }

  return first;
}

tree
start_class (code, class_name, super_name, protocol_list)
     enum tree_code code;
     tree class_name;
     tree super_name;
     tree protocol_list;
{
  tree class, decl;

  if (!doing_objc_thang)
    objc_fatal ();

  class = make_node (code);

  CLASS_NAME (class) = class_name;
  CLASS_SUPER_NAME (class) = super_name;
  CLASS_CLS_METHODS (class) = NULL_TREE;

  if ((decl = lookup_name (class_name)) != 0)
    {
      error ("`%s' redeclared as different kind of symbol",
	     IDENTIFIER_POINTER (class_name));
      error_with_decl (decl, "previous declaration of `%s'");
    }

  if (code == CLASS_IMPLEMENTATION)
    {
      /* pre-build the following entities - for speed/convenience. */
      if (!self_id)
        self_id = get_identifier ("self");
      if (!_cmd_id)
        _cmd_id = get_identifier ("_cmd");

      if (!objc_super_template)
	objc_super_template = build_super_template ();

      method_slot = 0;		/* reset for multiple classes per file */

      implementation_context = class;

      /* lookup the interface for this implementation. */

      if (!(implementation_template = lookup_interface (class_name)))
        {
	  warning ("Cannot find interface declaration for `%s'",
		   IDENTIFIER_POINTER (class_name));
	  add_class (implementation_template = implementation_context);
        }

      /* if a super class has been specified in the implementation,
	 insure it conforms to the one specified in the interface */

      if (super_name
	  && (super_name != CLASS_SUPER_NAME (implementation_template)))
        {
	  error ("conflicting super class name `%s'",
		 IDENTIFIER_POINTER (super_name));
	  error ("previous declaration of `%s'",
		 IDENTIFIER_POINTER (CLASS_SUPER_NAME (implementation_template)));
        }
    }
  else if (code == CLASS_INTERFACE)
    {
      if (lookup_interface (class_name))
        warning ("duplicate interface declaration for class `%s'",
                 IDENTIFIER_POINTER (class_name));
      else
        add_class (class);

      if (protocol_list)
        {
	  lookup_and_install_protocols (protocol_list);
	  CLASS_PROTOCOL_LIST (class) = protocol_list;
	}
    }
  else if (code == CATEGORY_INTERFACE)
    {
      tree class_category_is_assoc_with;

      /* for a category, class_name is really the name of the class that
	 the following set of methods will be associated with...we must
	 find the interface so that can derive the objects template */

      if (!(class_category_is_assoc_with = lookup_interface (class_name)))
	{
	  error ("Cannot find interface declaration for `%s'",
		 IDENTIFIER_POINTER (class_name));
	  exit (1);
	}
      else
        add_category (class_category_is_assoc_with, class);

      if (protocol_list)
        {
	  lookup_and_install_protocols (protocol_list);
	  CLASS_PROTOCOL_LIST (class) = protocol_list;
	}
    }
  else if (code == CATEGORY_IMPLEMENTATION)
    {
      /* pre-build the following entities - for speed/convenience. */
      if (!self_id)
        self_id = get_identifier ("self");
      if (!_cmd_id)
        _cmd_id = get_identifier ("_cmd");

      if (!objc_super_template)
	objc_super_template = build_super_template ();

      method_slot = 0;		/* reset for multiple classes per file */

      implementation_context = class;

      /* for a category, class_name is really the name of the class that
	 the following set of methods will be associated with...we must
	 find the interface so that can derive the objects template */

      if (!(implementation_template = lookup_interface (class_name)))
        {
	  error ("Cannot find interface declaration for `%s'",
		 IDENTIFIER_POINTER (class_name));
	  exit (1);
        }
    }
  return class;
}

tree
continue_class (class)
     tree class;
{
  if (TREE_CODE (class) == CLASS_IMPLEMENTATION
      || TREE_CODE (class) == CATEGORY_IMPLEMENTATION)
    {
      struct imp_entry *impEntry;
      tree ivar_context;

      /* check consistency of the instance variables. */

      if (CLASS_IVARS (class))
	check_ivars (implementation_template, class);

      /* code generation */

#ifdef OBJCPLUS
      push_lang_context (lang_name_c);
#endif /* OBJCPLUS */

      ivar_context = build_private_template (implementation_template);

      if (!objc_class_template)
	build_class_template ();

      if (!(impEntry = (struct imp_entry *)xmalloc (sizeof (struct imp_entry))))
	perror ("unable to allocate in objc-tree.c");

      impEntry->next = imp_list;
      impEntry->imp_context = class;
      impEntry->imp_template = implementation_template;

      synth_forward_declarations ();
      impEntry->class_decl = _OBJC_CLASS_decl;
      impEntry->meta_decl = _OBJC_METACLASS_decl;

      /* append to front and increment count */
      imp_list = impEntry;
      if (TREE_CODE (class) == CLASS_IMPLEMENTATION)
	imp_count++;
      else
	cat_count++;

#ifdef OBJCPLUS
      pop_lang_context ();
#endif /* OBJCPLUS */

      return ivar_context;
    }
  else if (TREE_CODE (class) == CLASS_INTERFACE)
    {
#ifdef OBJCPLUS
    tree record;

    push_lang_context (lang_name_c);
    record = start_struct (RECORD_TYPE, CLASS_NAME (class));
#else /* OBJCPLUS */
      tree record = xref_tag (RECORD_TYPE, CLASS_NAME (class));
#endif /* OBJCPLUS */

      if (!TYPE_FIELDS (record))
	{
	  finish_struct (record, build_ivar_chain (class, 0));
	  CLASS_STATIC_TEMPLATE (class) = record;

	  /* mark this record as a class template - for static typing */
	  TREE_STATIC_TEMPLATE (record) = 1;
	}

#ifdef OBJCPLUS
      pop_lang_context ();
#endif /* OBJCPLUS */

      return NULLT;
    }
}

/*
 * this is called once we see the "@end" in an interface/implementation.
 */
tree
finish_class (class)
     tree class;
{
  if (TREE_CODE (class) == CLASS_IMPLEMENTATION)
    {
      /* all code generation is done in "finish_objc ()" */

      if (implementation_template != implementation_context)
	{
	  /* Ensure that all methods listed in the interface have bodies!  */
	  check_methods (CLASS_CLS_METHODS (implementation_template),
			 CLASS_CLS_METHODS (implementation_context), '+');
	  check_methods (CLASS_NST_METHODS (implementation_template),
			 CLASS_NST_METHODS (implementation_context), '-');

	  if (CLASS_PROTOCOL_LIST (implementation_template))
	    {
	      tree proto;
	  
	      for (proto = CLASS_PROTOCOL_LIST (implementation_template); 
		   proto; proto = TREE_CHAIN (proto))
		{
		  tree p = TREE_VALUE (proto);
	      
		  if (TREE_CODE (p) == PROTOCOL_INTERFACE)
		    {
		      int f1, f2;
		      
		      /* Ensure that all protocols have bodies! */
		      f1 = check_methods (PROTOCOL_CLS_METHODS (p),
				CLASS_CLS_METHODS (implementation_context),
				'+');
		      f2 = check_methods (PROTOCOL_NST_METHODS (p),
				CLASS_NST_METHODS (implementation_context),
				'-');
				
		      if (!f1 || !f2)
			warning ("class `%s' does not fully implement the `%s' protocol",
			    IDENTIFIER_POINTER (CLASS_NAME (implementation_context)),
			    IDENTIFIER_POINTER (PROTOCOL_NAME (p)));
      
		    }
		  else 
		    ; /* an identifier...if we could not find a protocol.  */
		}
	    }
	}
    }
  else if (TREE_CODE (class) == CATEGORY_IMPLEMENTATION)
    {
      tree category = CLASS_CATEGORY_LIST (implementation_template);
  
      /* find the category interface from the class it is associated with */
      while (category)
	{
	  if (CLASS_SUPER_NAME (class) == CLASS_SUPER_NAME (category))
	    break;
	  category = CLASS_CATEGORY_LIST (category);
	}
  
      if (category)
	{
	  /* ensure that all method listed in the interface contain bodies! */
	  check_methods (CLASS_CLS_METHODS (category),
			 CLASS_CLS_METHODS (implementation_context), '+');
	  check_methods (CLASS_NST_METHODS (category),
			 CLASS_NST_METHODS (implementation_context), '-');

/* NEW!!! */
	  if (CLASS_PROTOCOL_LIST (category))
	    {
	      tree proto;
	  
	      for (proto = CLASS_PROTOCOL_LIST (category); 
		   proto; proto = TREE_CHAIN (proto))
		{
		  tree p = TREE_VALUE (proto);
	      
		  if (TREE_CODE (p) == PROTOCOL_INTERFACE)
		    {
		      int f1, f2;
		      
		      /* Ensure that all protocols have bodies! */
		      f1 = check_methods (PROTOCOL_CLS_METHODS (p),
				CLASS_CLS_METHODS (implementation_context),
				'+');
		      f2 = check_methods (PROTOCOL_NST_METHODS (p),
				CLASS_NST_METHODS (implementation_context),
				'-');
				
		      if (!f1 || !f2)
			warning ("category `%s' does not fully implement the `%s' protocol",
			    IDENTIFIER_POINTER (CLASS_SUPER_NAME (implementation_context)),
			    IDENTIFIER_POINTER (PROTOCOL_NAME (p)));
      
		    }
		  else 
		    ; /* an identifier...if we could not find a protocol.  */
		}
	    }
	}
    }
  else if (TREE_CODE (class) == CLASS_INTERFACE)
    {
      tree decl_specs;

      /* extern struct objc_object *_<my_name>; */

      sprintf (utlbuf, "_%s", IDENTIFIER_POINTER (CLASS_NAME (class)));

      decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_EXTERN]);
      decl_specs = tree_cons (NULLT, objc_object_reference, decl_specs);
      define_decl (build1 (INDIRECT_REF, NULLT, get_identifier (utlbuf)), decl_specs);
    }
}

static tree
add_protocol (protocol)
     tree protocol;
{
  /* put protocol on list in reverse order */
  TREE_CHAIN (protocol) = protocol_chain;
  protocol_chain = protocol;
  return protocol_chain;
}

tree
lookup_protocol (ident)
     tree ident;
{
  tree chain;
    
  for (chain = protocol_chain; chain; chain = TREE_CHAIN (chain))
    {
      if (ident == PROTOCOL_NAME (chain))
	return chain;
    }
  return NULLT;
}

tree
start_protocol (code, name, list)
     enum tree_code code;
     tree name;
     tree list;
{
  tree protocol;

  if (!doing_objc_thang)
    objc_fatal ();

  /* this is as good a place as any...need a "push_tag_toplevel()" */
  if (!objc_protocol_template)
    objc_protocol_template = build_protocol_template ();

  protocol = make_node (code);

  PROTOCOL_NAME (protocol) = name;
  PROTOCOL_LIST (protocol) = list;

  lookup_and_install_protocols (list);

  if (lookup_protocol (name))
    warning ("duplicate declaration for protocol `%s'",
	     IDENTIFIER_POINTER (name));
  else
    add_protocol (protocol);

  PROTOCOL_FORWARD_DECL (protocol) = NULL_TREE;

  return protocol;
}

void
finish_protocol (protocol)
	tree protocol;
{
}

static void
encode_type_qualifiers (declspecs, utlbuf)
     tree declspecs;
     char *utlbuf;
{
  tree spec;

  for (spec = declspecs; spec; spec = TREE_CHAIN (spec))
    {
      if (ridpointers[RID_CONST] == TREE_VALUE (spec))
	strcat (utlbuf, "r");
      else if (ridpointers[RID_IN] == TREE_VALUE (spec))
	strcat (utlbuf, "n");
      else if (ridpointers[RID_INOUT] == TREE_VALUE (spec))
	strcat (utlbuf, "N");
      else if (ridpointers[RID_OUT] == TREE_VALUE (spec))
	strcat (utlbuf, "o");
      else if (ridpointers[RID_BYCOPY] == TREE_VALUE (spec))
	strcat (utlbuf, "O");
      else if (ridpointers[RID_ONEWAY] == TREE_VALUE (spec))
	strcat (utlbuf, "V");
    }
}

static void
encode_pointer (type, str, curtype, format)
     tree type;
     char *str;
     char *curtype;
     int format;
{
  tree pointer_to = TREE_TYPE (type);

  if (TREE_CODE (pointer_to) == RECORD_TYPE)
    {
      if (TYPE_NAME (pointer_to)
	  && TREE_CODE (TYPE_NAME (pointer_to)) == IDENTIFIER_NODE)
	{
	  char *name = IDENTIFIER_POINTER (TYPE_NAME (pointer_to));

	  if ((strcmp (name, TAG_OBJECT) == 0) || /* '@' */
	      (TREE_STATIC_TEMPLATE (pointer_to)))
	    {
	      strcat (str, "@");
	      return;
	    }
	  else if (strcmp (name, TAG_CLASS) == 0) /* '#' */
	    {
	      strcat (str, "#");
	      return;
	    }
#ifdef OBJC_SELS_R_STRUCT_PTRS
	  else if (strcmp (name, TAG_SELECTOR) == 0) /* ':' */
	    {
	      strcat (str, ":");
	      return;
	    }
#endif
	}
    }
  else if (TREE_CODE (pointer_to) == INTEGER_TYPE
	   && TYPE_MODE (pointer_to) == QImode)
    {
      strcat (str, "*");
      return;
    }

  /* we have a type that does not get special treatment... */

  /* NeXT extension */
  strcat (str, "^");
  encode_type (pointer_to, str, curtype, format);
}

static void
encode_array (type, str, curtype, format)
     tree type;
     char *str;
     char *curtype;
     int format;
{
  tree anIntCst = TYPE_SIZE (type);
  tree array_of = TREE_TYPE (type);

  /* An incomplete array is treated like a pointer.  */
  if (anIntCst == NULL)
    {
      /* split for obvious reasons.  North-Keys 30 Mar 1991 */
      encode_pointer (type, str, curtype, format);
      return;
    }
  
  sprintf (str + strlen (str), "[%d",
	   TREE_INT_CST_LOW (anIntCst)
	   / TREE_INT_CST_LOW (TYPE_SIZE (array_of)));
  encode_type (array_of, str, curtype, format);
  strcat (str, "]");
  return;
}

static void
encode_aggregate (type, str, curtype, format)
     tree type;
     char *str;
     char *curtype;
     int format;
{
  enum tree_code code = TREE_CODE (type);

  switch (code)
    {
    case RECORD_TYPE:
      {
	if (strlen (curtype) && curtype[strlen (curtype) - 1] == '^')
	  {
	    tree name = TYPE_NAME (type);

	    /* we have a reference - this is a NeXT extension */

            if ((strlen (curtype) == 1)
	        && (format == OBJC_ENCODE_INLINE_DEFS))
	      {
		/* output format of struct for first level only! */
		
		tree fields = TYPE_FIELDS (type);
		
                if (name && (TREE_CODE (name) == IDENTIFIER_NODE))
	          sprintf (str + strlen (str), "{%s=",
		  			IDENTIFIER_POINTER (name));
		else
		  strcat (str, "{?=");
		  
		for ( ; fields; fields = TREE_CHAIN (fields))
		  encode_field_decl (fields, str, curtype, format);
		strcat (str, "}");
	      }
            else if (name && (TREE_CODE (name) == IDENTIFIER_NODE))
	      sprintf (str + strlen (str), "{%s}",
		       IDENTIFIER_POINTER (name));
	    else	/* we have an untagged structure or a typedef */
	      sprintf (str + strlen (str), "{?}");
	  }
	else
	  {
	    tree name = TYPE_NAME (type);
	    tree fields = TYPE_FIELDS (type);

            if (format == OBJC_ENCODE_INLINE_DEFS)
              {
                if (name && (TREE_CODE (name) == IDENTIFIER_NODE))
	          sprintf (str + strlen (str), "{%s=",
		  			IDENTIFIER_POINTER (name));
		else
		  strcat (str, "{?=");
		  
		for ( ; fields; fields = TREE_CHAIN (fields))
		  encode_field_decl (fields, str, curtype, format);
		strcat (str, "}");
              }
            else
              {
                if (name && (TREE_CODE (name) == IDENTIFIER_NODE))
		  sprintf (str + strlen (str), "{%s}",
			   IDENTIFIER_POINTER (name));
		else	/* we have an untagged structure or a typedef */
		  sprintf (str + strlen (str), "{?}");
              }
	  }
	break;
      }
    case UNION_TYPE:
      {
	if (str[strlen (str)-1] == '^')
	  {
            if (TYPE_NAME (type)
		&& (TREE_CODE (TYPE_NAME (type)) == IDENTIFIER_NODE))
	      /* we have a reference - this is a NeXT extension */
	      sprintf (str + strlen (str), "(%s)",
		       IDENTIFIER_POINTER (TYPE_NAME (type)));
	    else		/* we have an untagged structure */
	      sprintf (str + strlen (str), "(?)");
	  }
	else
	  {
	    tree fields = TYPE_FIELDS (type);

            if (format == OBJC_ENCODE_INLINE_DEFS)
              {
		strcat (str, "(");
		for ( ; fields; fields = TREE_CHAIN (fields))
		  encode_field_decl (fields, str, curtype, format);
		strcat (str, ")");
              }
            else
              {
		if (TYPE_NAME (type) &&
		    (TREE_CODE (TYPE_NAME (type)) == IDENTIFIER_NODE))
		  /* we have a reference - this is a NeXT extension */
		  sprintf (str + strlen (str), "(%s)",
			   IDENTIFIER_POINTER (TYPE_NAME (type)));
		else		/* we have an untagged structure */
		  sprintf (str + strlen (str), "(?)");
              }
	  }
	break;
      }
    case ENUMERAL_TYPE:
      strcat (str, "i");
      break;
    }
}

/*
 *  support bitfields, the current version of Objective-C does not support
 *  them. the string will consist of one or more "b:n"'s where n is an
 *  integer describing the width of the bitfield. Currently, classes in
 *  the kit implement a method "-(char *)describeBitfieldStruct:" that
 *  simulates this...if they do not implement this method, the archiver
 *  assumes the bitfield is 16 bits wide (padded if necessary) and packed
 *  according to the GNU compiler. After looking at the "kit", it appears
 *  that all classes currently rely on this default behavior, rather than
 *  hand generating this string (which is tedious).
 */
static void
encode_bitfield (width, str, format)
     int width;
     char *str;
     int format;
{
  sprintf (str + strlen (str), "b%d", width);
}

/*
 *	format will be:
 *
 *	OBJC_ENCODE_INLINE_DEFS or OBJC_ENCODE_DONT_INLINE_DEFS
 */
static void
encode_type (type, str, curtype, format)
     tree type;
     char *str;
     char *curtype;
     int format;
{
  enum tree_code code = TREE_CODE (type);

  if (code == INTEGER_TYPE)
    {
      if (TREE_INT_CST_LOW (TYPE_MIN_VALUE (type)) == 0)
	{
	  /* unsigned integer types */

	  if (TYPE_MODE (type) == QImode) /* 'C' */
	    strcat (str, "C");
	  else if (TYPE_MODE (type) == HImode) /* 'S' */
	    strcat (str, "S");
	  else if (TYPE_MODE (type) == SImode)
	    {
	      if (type == long_unsigned_type_node)
		strcat (str, "L"); /* 'L' */
	      else
		strcat (str, "I"); /* 'I' */
	    }
	}
      else			/* signed integer types */
	{
	  if (TYPE_MODE (type) == QImode) /* 'c' */
	    strcat (str, "c");
	  else if (TYPE_MODE (type) == HImode) /* 's' */
	    strcat (str, "s");
	  else if (TYPE_MODE (type) == SImode) /* 'i' */
	    {
	      if (type == long_integer_type_node)
		strcat (str, "l"); /* 'l' */
	      else
		strcat (str, "i"); /* 'i' */
	    }
	}
    }
  else if (code == REAL_TYPE)
    {
      /* floating point types */

      if (TYPE_MODE (type) == SFmode) /* 'f' */
	strcat (str, "f");
      else if (TYPE_MODE (type) == DFmode
	       || TYPE_MODE (type) == TFmode) /* 'd' */
	strcat (str, "d");
    }

  else if (code == VOID_TYPE)	/* 'v' */
    strcat (str, "v");

  else if (code == ARRAY_TYPE)
    encode_array (type, str, curtype, format);

  else if (code == POINTER_TYPE)
    encode_pointer (type, str, curtype, format);

  else if (code == RECORD_TYPE || code == UNION_TYPE || code == ENUMERAL_TYPE)
    encode_aggregate (type, str, curtype, format);

  else if (code == FUNCTION_TYPE) /* '?' */
    strcat (str, "?");
}

static void
encode_field_decl (field_decl, str, curtype, format)
     tree field_decl;
     char *str;
     char *curtype;
     int format;
{
#ifdef OBJCPLUS
  if (TREE_PACKED(field_decl))
    encode_bitfield (DECL_SIZE_UNIT(field_decl), str, format);
#else /* OBJCPLUS */
  if (DECL_BIT_FIELD (field_decl))
    encode_bitfield (DECL_FRAME_SIZE (field_decl), str, format);
#endif /* OBJCPLUS */
  else
    encode_type (TREE_TYPE (field_decl), str, curtype, format);
}

static tree
expr_last (complex_expr)
     tree complex_expr;
{
  tree next;

  if (complex_expr)
    while (next = TREE_OPERAND (complex_expr, 0))
      complex_expr = next;
  return complex_expr;
}


/* The selector of the current method,
   or NULL if we aren't compiling a method.  */

tree
maybe_objc_method_name (decl)
      tree decl;
{
  if (method_context)
    return METHOD_SEL_NAME (method_context);
  else
    return 0;
}

/*
 *  Transform a method definition into a function definition as follows:
 *
 *  - synthesize the first two arguments, "self" and "_cmd".
 */

void
start_method_def (method)
     tree method;
{
  tree decl_specs;

  /* required to implement _msgSuper () */
  method_context = method;
  _OBJC_SUPER_decl = NULLT;

  pushlevel (0); 		/* must be called BEFORE "start_function ()" */

  /* generate prototype declarations for arguments..."new-style" */

  if (TREE_CODE (method_context) == INSTANCE_METHOD_DECL)
    decl_specs = build_tree_list (NULLT, _PRIVATE_record);
  else
    /* really a `struct objc_class *'...however we allow people to
       assign to self...which changes its type midstream.
       */
    decl_specs = build_tree_list (NULLT, objc_object_reference);

  push_parm_decl (build_tree_list (decl_specs,
				   build1 (INDIRECT_REF, NULLT, self_id)));

#ifdef OBJC_SELS_R_INTS
  decl_specs = build_tree_list (NULLT, ridpointers[(int) RID_UNSIGNED]);
  decl_specs = tree_cons (NULLT, ridpointers[(int) RID_INT], decl_specs);
  push_parm_decl (build_tree_list (decl_specs, _cmd_id));
#endif

#ifdef OBJC_SELS_R_STRUCT_PTRS
  decl_specs = build_tree_list (NULLT,
				xref_tag (RECORD_TYPE,
					  get_identifier (TAG_SELECTOR)));
  push_parm_decl (build_tree_list (decl_specs, 
				   build1 (INDIRECT_REF, NULLT, _cmd_id)));
#endif

  /* generate argument delclarations if a keyword_decl */
  if (METHOD_SEL_ARGS (method))
    {
      tree arglist = METHOD_SEL_ARGS (method);
      do
	{
	  tree arg_spec = TREE_PURPOSE (TREE_TYPE (arglist));
	  tree arg_decl = TREE_VALUE (TREE_TYPE (arglist));

	  if (arg_decl)
	    {
	      tree last_expr = expr_last (arg_decl);

	      /* unite the abstract decl with its name */
	      TREE_OPERAND (last_expr, 0) = KEYWORD_ARG_NAME (arglist);
	      push_parm_decl (build_tree_list (arg_spec, arg_decl));
#ifndef OBJCPLUS
	      /* unhook...restore the abstract declarator */
	      TREE_OPERAND (last_expr, 0) = NULLT;
#endif /* not OBJCPLUS */
	    }
	  else
	    push_parm_decl (build_tree_list (arg_spec, KEYWORD_ARG_NAME (arglist)));

	  arglist = TREE_CHAIN (arglist);
	}
      while (arglist);
    }

  if (METHOD_ADD_ARGS (method) > (tree)1)
    {
      /* we have a variable length selector - in "prototype" format */
      tree akey = TREE_PURPOSE (METHOD_ADD_ARGS (method));
      while (akey)
	{
	  /* this must be done prior to calling pushdecl (). pushdecl () is
	   * going to change our chain on us...
	   */
	  tree nextkey = TREE_CHAIN (akey);
	  pushdecl (akey);
	  akey = nextkey;
	}
    }
}

static void
error_with_method (message, mtype, method)
     char *message;
     char mtype;
     tree method;
{
  extern int errorcount;

#ifdef REPORT_EVENT
  bzero (errbuf, BUFSIZE);
  report_event (0, current_function_decl,
		METHOD_FILENAME (method), METHOD_LINENUM (method),
		"%s `%c%s'", message, mtype,
		gen_method_decl (method, errbuf));
#endif
  fprintf (stderr, "%s:%d: ",
	   METHOD_FILENAME (method), METHOD_LINENUM (method));
  bzero (errbuf, BUFSIZE);
  fprintf (stderr, "%s `%c%s'\n", message, mtype, gen_method_decl (method, errbuf));
  errorcount++;
}

static void
warn_with_method (message, mtype, method)
     char *message;
     char mtype;
     tree method;
{
#ifdef REPORT_EVENT
  bzero (errbuf, BUFSIZE);
  report_event (1, current_function_decl,
		METHOD_FILENAME (method), METHOD_LINENUM (method),
		"%s `%c%s'", message, mtype,
		gen_method_decl (method, errbuf));
#endif
  fprintf (stderr, "%s:%d: ",
	   METHOD_FILENAME (method), METHOD_LINENUM (method));
  bzero (errbuf, BUFSIZE);
  fprintf (stderr, "%s `%c%s'\n", message, mtype, gen_method_decl (method, errbuf));
}

/* return 1 if `method' is consistent with `proto' */

static int
comp_method_with_proto (method, proto)
     tree method, proto;
{
  static tree function_type = 0;

  /* create a function_type node once */
  if (!function_type)
    {
      struct obstack *ambient_obstack = current_obstack;
      
      current_obstack = &permanent_obstack;
      function_type = make_node (FUNCTION_TYPE);
      current_obstack = ambient_obstack;
    }

  /* install argument types - normally set by "build_function_type ()". */
  TYPE_ARG_TYPES (function_type) = get_arg_type_list (proto, METHOD_DEF, 0);

  /* install return type */
  TREE_TYPE (function_type) = groktypename (TREE_TYPE (proto));

  return comptypes (TREE_TYPE (METHOD_DEFINITION (method)), function_type);
}

/* return 1 if `proto1' is consistent with `proto2' */

static int
comp_proto_with_proto (proto1, proto2)
     tree proto1, proto2;
{
  static tree function_type1 = 0, function_type2 = 0;

  /* create a couple function_type node's once */
  if (!function_type1)
    {
      struct obstack *ambient_obstack = current_obstack;
      
      current_obstack = &permanent_obstack;
      function_type1 = make_node (FUNCTION_TYPE);
      function_type2 = make_node (FUNCTION_TYPE);
      current_obstack = ambient_obstack;
    }

  /* install argument types - normally set by "build_function_type ()". */
  TYPE_ARG_TYPES (function_type1) = get_arg_type_list (proto1, METHOD_REF, 0);
  TYPE_ARG_TYPES (function_type2) = get_arg_type_list (proto2, METHOD_REF, 0);

  /* install return type */
  TREE_TYPE (function_type1) = groktypename (TREE_TYPE (proto1));
  TREE_TYPE (function_type2) = groktypename (TREE_TYPE (proto2));

  return comptypes (function_type1, function_type2);
}

/*
 *  - generate an identifier for the function. the format is "_n_cls",
 *    where 1 <= n <= nMethods, and cls is the name the implementation we
 *    are processing.
 *  - install the return type from the method declaration.
 *  - if we have a prototype, check for type consistency.
 */
static void
really_start_method (method, parmlist)
     tree method, parmlist;
{
  tree sc_spec, ret_spec, ret_decl, decl_specs;
  tree method_decl, method_id;
  char buf[256];

  /* synth the storage class & assemble the return type */
  sc_spec = tree_cons (NULLT, ridpointers[(int) RID_STATIC], NULLT);
  ret_spec = TREE_PURPOSE (TREE_TYPE (method));
  decl_specs = chainon (sc_spec, ret_spec);

  if (TREE_CODE (implementation_context) == CLASS_IMPLEMENTATION)
#ifdef OBJC_GEN_METHOD_LABEL
    OBJC_GEN_METHOD_LABEL (buf,
      TREE_CODE (method) == INSTANCE_METHOD_DECL,
      IDENTIFIER_POINTER (CLASS_NAME (implementation_context)),
      NULL,
      IDENTIFIER_POINTER (METHOD_SEL_NAME (method)));
#else
    sprintf (buf, "_%d_%s", ++method_slot,
	     IDENTIFIER_POINTER (CLASS_NAME (implementation_context)));
#endif
  else				/* we have a category */
#ifdef OBJC_GEN_METHOD_LABEL
    OBJC_GEN_METHOD_LABEL (buf,
      TREE_CODE (method) == INSTANCE_METHOD_DECL,
      IDENTIFIER_POINTER (CLASS_NAME (implementation_context)),
      IDENTIFIER_POINTER (CLASS_SUPER_NAME (implementation_context)),
      IDENTIFIER_POINTER (METHOD_SEL_NAME (method)));
#else
    sprintf (buf, "_%d_%s_%s", ++method_slot,
	     IDENTIFIER_POINTER (CLASS_NAME (implementation_context)),
	     IDENTIFIER_POINTER (CLASS_SUPER_NAME (implementation_context)));
#endif

  method_id = get_identifier (buf);

  method_decl = build_nt (CALL_EXPR, method_id, parmlist, NULLT);

  /* check the delclarator portion of the return type for the method */
  if (ret_decl = TREE_VALUE (TREE_TYPE (method)))
    {
      /*
       * unite the complex decl (specified in the abstract decl) with the
       * function decl just synthesized...(int *), (int (*)()), (int (*)[]).
       */
      tree save_expr = expr_last (ret_decl);

      TREE_OPERAND (save_expr, 0) = method_decl;
      method_decl = ret_decl;
      /* fool the parser into thinking it is starting a function */
      start_function (decl_specs, method_decl, 0);
#ifdef OBJCPLUS
      /* must be called AFTER "start_function()" */
      store_parm_decls (); 	
#endif /* OBJCPLUS */
      /* unhook...this has the effect of restoring the abstract declarator */
      TREE_OPERAND (save_expr, 0) = NULLT;
    }
  else
    {
      TREE_VALUE (TREE_TYPE (method)) = method_decl;
      /* fool the parser into thinking it is starting a function */
      start_function (decl_specs, method_decl, 0);
#ifdef OBJCPLUS
      /* must be called AFTER "start_function()" */
      store_parm_decls (); 	
#endif /* OBJCPLUS */
      /* unhook...this has the effect of restoring the abstract declarator */
      TREE_VALUE (TREE_TYPE (method)) = NULLT;
    }

#ifdef OBJCPLUS
  /* set self_decl from the first argument...this global is used by 
   * build_ivar_reference().build_indirect_ref().
   */
  self_decl = DECL_ARGUMENTS(current_function_decl);
#endif /* OBJCPLUS */

  METHOD_DEFINITION (method) = current_function_decl;

  /* check consistency...start_function (), pushdecl (), duplicate_decls (). */

  if (implementation_template != implementation_context)
    {
      tree chain, proto;

      if (TREE_CODE (method) == INSTANCE_METHOD_DECL)
	proto = lookup_instance_method_static (implementation_template,
					       METHOD_SEL_NAME (method));
      else
	proto = lookup_class_method_static (implementation_template,
					    METHOD_SEL_NAME (method));

      if (proto)
	{
	  if (!comp_method_with_proto (method, proto))
	    {
	      fprintf (stderr, "%s: In method `%s'\n", input_filename,
		       IDENTIFIER_POINTER (METHOD_SEL_NAME (method)));
	      if (TREE_CODE (method) == INSTANCE_METHOD_DECL)
		{
		  warn_with_method ("conflicting types for", '-', method);
		  warn_with_method ("previous declaration of", '-', proto);
		}
	      else
		{
		  warn_with_method ("conflicting types for", '+', method);
		  warn_with_method ("previous declaration of", '+', proto);
		}
	    }
	}
    }
}

/*
 * the following routine is always called...this "architecture" is to
 * accommodate "old-style" variable length selectors.
 *
 *	- a:a b:b // prototype  ; id c; id d; // old-style
 */
void
continue_method_def ()
{
  tree parmlist;

  if (METHOD_ADD_ARGS (method_context) == (tree)1)
    /*
     * we have a `, ...' immediately following the selector.
     */
    parmlist = get_parm_info (0);
  else
    parmlist = get_parm_info (1); /* place a `void_at_end' */

#ifndef OBJCPLUS
  /* set self_decl from the first argument...this global is used by
   * build_ivar_reference ().build_indirect_ref ().
   */
  self_decl = TREE_PURPOSE (parmlist);
#endif /* not OBJCPLUS */

  poplevel (0, 0, 0);		/* must be called BEFORE "start_function ()" */

  really_start_method (method_context, parmlist);

#ifndef OBJCPLUS
  store_parm_decls ();		/* must be called AFTER "start_function ()" */
#endif /* not OBJCPLUS */
}

/* Called by the parser, from the `pushlevel' production.  */

void
add_objc_decls ()
{
  if (!_OBJC_SUPER_decl)
    {
      _OBJC_SUPER_decl = start_decl (get_identifier (_TAG_SUPER),
				     build_tree_list (NULLT,
						      objc_super_template),
				     0);

      finish_decl (_OBJC_SUPER_decl, NULLT, NULLT);

      /* this prevents `unused variable' warnings. */
      TREE_USED (_OBJC_SUPER_decl) = 1;
    }
}

/*
 *	_n_Method (id self, SEL sel, ...)
 *	{
 *		struct objc_super _S;
 *
 *		_msgSuper ((_S.self = self, _S.class = _cls, &_S), ...);
 *	}
 */
tree
get_super_receiver ()
{
  if (method_context)
    {
      tree super_expr, super_expr_list;

      /* set receiver to self */
      super_expr = build_component_ref (_OBJC_SUPER_decl, self_id);
      super_expr = build_modify_expr (super_expr, NOP_EXPR, self_decl);
      super_expr_list = build_tree_list (NULLT, super_expr);

      /* set class to begin searching */
      super_expr = build_component_ref (_OBJC_SUPER_decl, get_identifier ("class"));

      if (TREE_CODE (implementation_context) == CLASS_IMPLEMENTATION)
	{
	  /* [_cls, __cls]Super are "pre-built" in synth_foward_declarations () */

	  if (TREE_CODE (method_context) == INSTANCE_METHOD_DECL)
	    super_expr = build_modify_expr (super_expr, NOP_EXPR, _clsSuper_ref);
	  else
	    super_expr = build_modify_expr (super_expr, NOP_EXPR, __clsSuper_ref);
	}
      else			/* we have a category... */
	{
	  tree params, super_name = CLASS_SUPER_NAME (implementation_template);
	  tree funcCall;
	  tree super_class;

	  if (!super_name)  /* Barf if super used in a category of Object. */
	    {
	      error("no super class declared in interface for `%s'",
		    IDENTIFIER_POINTER (CLASS_NAME (implementation_template)));
	      return error_mark_node;
	    }

#ifdef OBJC_CLASS_REFS
	  super_class = get_class_reference (super_name);
	  if (TREE_CODE (method_context) == CLASS_METHOD_DECL)
	    super_class
	      = build_component_ref (build_indirect_ref (super_class, "->"),
				     get_identifier ("isa"));
	  TREE_TYPE (super_class) = TREE_TYPE (_clsSuper_ref);
	  super_expr = build_modify_expr (super_expr, NOP_EXPR, super_class);
#else /* OBJC_CLASS_REFS */
	  add_class_reference (super_name);

	  params = build_tree_list (NULLT,
				    my_build_string (IDENTIFIER_LENGTH (super_name) + 1,
						     IDENTIFIER_POINTER (super_name)));

	  if (TREE_CODE (method_context) == INSTANCE_METHOD_DECL)
	    funcCall = build_function_call (objc_getClass_decl, params);
	  else
	    funcCall = build_function_call (objc_getMetaClass_decl, params);

	  /* cast! */
	  TREE_TYPE (funcCall) = TREE_TYPE (_clsSuper_ref);
	  super_expr = build_modify_expr (super_expr, NOP_EXPR, funcCall);
#endif /* OBJC_CLASS_REFS */
	}
      chainon (super_expr_list, build_tree_list (NULL_TREE, super_expr));

      super_expr = build_unary_op (ADDR_EXPR, _OBJC_SUPER_decl, 0);
      chainon (super_expr_list, build_tree_list (NULL_TREE, super_expr));

      return build_compound_expr (super_expr_list);
    }
  else
    {
      error ("[super ...] must appear in a method context");
      return error_mark_node;
    }
}

static tree
encode_method_def (func_decl)
      tree func_decl;
{
  tree parms;
  int stack_size = 0;

  bzero (utlbuf, BUFSIZE);

  /* return type */
  encode_type (TREE_TYPE (TREE_TYPE (func_decl)), utlbuf,
	       utlbuf + strlen (utlbuf), OBJC_ENCODE_INLINE_DEFS);

  /* stack size */
  for (parms = DECL_ARGUMENTS (func_decl); parms;
       parms = TREE_CHAIN (parms))
    stack_size += TREE_INT_CST_LOW (TYPE_SIZE (DECL_ARG_TYPE (parms)))
		  / BITS_PER_UNIT;

  sprintf (&utlbuf[strlen (utlbuf)], "%d", stack_size);

  /* argument types */
  for (parms = DECL_ARGUMENTS (func_decl); parms;
       parms = TREE_CHAIN (parms))
    {
      int offset_in_bytes;
  
      /* type */ 
      encode_type (TREE_TYPE (parms), utlbuf, utlbuf + strlen (utlbuf),
		   OBJC_ENCODE_INLINE_DEFS);
  
      /* compute offset */
#ifdef OBJCPLUS
      if (DECL_OFFSET (parms) >= 0)
	{
	  offset_in_bytes = DECL_OFFSET (parms) / BITS_PER_UNIT;
#else /* OBJCPLUS */
      if (GET_CODE (DECL_INCOMING_RTL (parms)) == MEM)
        {
	  rtx addr = XEXP (DECL_INCOMING_RTL (parms), 0);
	  
	  /* ??? Here we assume that the parm address is indexed
	      off the frame pointer or arg pointer.
	      If that is not true, we produce meaningless results,
	      but do not crash.  */
	  if (GET_CODE (addr) == PLUS
	      && GET_CODE (XEXP (addr, 1)) == CONST_INT)
	    offset_in_bytes = INTVAL (XEXP (addr, 1));
	  else
	    offset_in_bytes = 0;
#endif /* OBJCPLUS */
	  
	  /* This is the case where the parm is passed as an int or double
	      and it is converted to a char, short or float and stored back
	      in the parmlist.  In this case, describe the parm
	      with the variable's declared type, and adjust the address
	      if the least significant bytes (which we are using) are not
	      the first ones.  */
#ifdef BYTES_BIG_ENDIAN
	  if (TREE_TYPE (parms) != DECL_ARG_TYPE (parms))
	    offset_in_bytes += (GET_MODE_SIZE (TYPE_MODE (DECL_ARG_TYPE (parms)))
				- GET_MODE_SIZE (GET_MODE (DECL_RTL (parms))));
#endif
	}
      else
        offset_in_bytes = 0;
      
#ifdef OBJCPLUS
      sprintf (&utlbuf[strlen (utlbuf)], "%d", offset_in_bytes);
#else /* OBJCPLUS */
      /* The "+ 8" is a total hack to account for the return pc and
         saved fp on the 68k.  We should redefine this format! */
      sprintf (&utlbuf[strlen (utlbuf)], "%d", offset_in_bytes + 8);
#endif /* OBJCPLUS */
    }

  return get_identifier (utlbuf);
}

void
finish_method_def ()
{
  METHOD_ENCODING (method_context) =
    encode_method_def (current_function_decl);

  finish_function (0);

  /* this must be done AFTER finish_function, since the optimizer may
     find "may be used before set" errors.  */
  method_context = NULLT;	/* required to implement _msgSuper () */
}

int
lang_report_error_function (decl)
      tree decl;
{
  if (method_context)
    {
      fprintf (stderr, "In method `%s'\n",
	       IDENTIFIER_POINTER (METHOD_SEL_NAME (method_context)));
      return 1;
    }
  else
    return 0;
}

static int
is_complex_decl (type)
     tree type;
{
  return (TREE_CODE (type) == ARRAY_TYPE
	  || TREE_CODE (type) == FUNCTION_TYPE
	  || (TREE_CODE (type) == POINTER_TYPE
	      && TYPE_MAIN_VARIANT (type) != id_type));
}


/* Code to convert a decl node into text for a declaration in C.  */

static char tmpbuf[256];

static void
adorn_decl (decl, str)
     tree decl;
     char *str;
{
  enum tree_code code = TREE_CODE (decl);

  if (code == ARRAY_REF)
    {
      tree anIntCst = TREE_OPERAND (decl, 1);

      if (anIntCst && TREE_CODE (anIntCst) == INTEGER_CST)
	sprintf (str + strlen (str), "[%d]", TREE_INT_CST_LOW (anIntCst));
      else
        strcat (str, "[]");
    }
  else if (code == ARRAY_TYPE)
    {
      tree anIntCst = TYPE_SIZE (decl);
      tree array_of = TREE_TYPE (decl);

      if (anIntCst && TREE_CODE (anIntCst) == INTEGER_CST)
	sprintf (str + strlen (str), "[%d]",
	       TREE_INT_CST_LOW (anIntCst)/TREE_INT_CST_LOW (TYPE_SIZE (array_of)));
      else
        strcat (str, "[]");
    }
  else if (code == CALL_EXPR)
    {
      tree chain = TREE_PURPOSE (TREE_OPERAND (decl, 1));

      strcat (str, "(");
      while (chain)
	{
	  gen_declaration (chain, str);
	  chain = TREE_CHAIN (chain);
	  if (chain)
	    strcat (str, ", ");
	}
      strcat (str, ")");
    }
  else if (code == FUNCTION_TYPE)
    {
      tree chain  = TYPE_ARG_TYPES (decl); /* a list of types */

      strcat (str, "(");
      while (chain && TREE_VALUE (chain) != void_type_node)
	{
	  gen_declaration (TREE_VALUE (chain), str);
	  chain = TREE_CHAIN (chain);
	  if (chain && TREE_VALUE (chain) != void_type_node)
	    strcat (str, ", ");
	}
      strcat (str, ")");
    }
  else if (code == INDIRECT_REF)
    {
      strcpy (tmpbuf, "*");
      if (TREE_TYPE (decl) && TREE_CODE (TREE_TYPE (decl)) == TREE_LIST)
        {
	  tree chain;
	  
	  for (chain = nreverse (copy_list (TREE_TYPE (decl)));
	       chain;
	       chain = TREE_CHAIN (chain))
	    {
	      if (TREE_CODE (TREE_VALUE (chain)) == IDENTIFIER_NODE)
		{
		  strcat (tmpbuf, " ");
		  strcat (tmpbuf, IDENTIFIER_POINTER (TREE_VALUE (chain)));
		}
	    }
	  if (str[0])
	    strcat (tmpbuf, " ");
	}
      strcat (tmpbuf, str);
      strcpy (str, tmpbuf);
    }
  else if (code == POINTER_TYPE)
    {
      strcpy (tmpbuf, "*");
      if (TREE_READONLY (decl) || TYPE_VOLATILE (decl))
        {
	  if (TREE_READONLY (decl))
	    strcat (tmpbuf, " const");
	  if (TYPE_VOLATILE (decl))
	    strcat (tmpbuf, " volatile");
	  if (str[0])
	    strcat (tmpbuf, " ");
	}
      strcat (tmpbuf, str);
      strcpy (str, tmpbuf);
    }
}

static char *
gen_declarator (decl, buf, name)
     tree decl;
     char *buf;
     char *name;
{
  if (decl)
    {
      enum tree_code code = TREE_CODE (decl);
      char *str;
      tree op;
      int wrap = 0;

      switch (code)
	{
	case ARRAY_REF: case INDIRECT_REF: case CALL_EXPR:
	  {
	    op = TREE_OPERAND (decl, 0);

	    /* we have a pointer to a function or array...(*)(), (*)[] */
	    if ((code == ARRAY_REF || code == CALL_EXPR) &&
		(op && TREE_CODE (op) == INDIRECT_REF))
	      wrap = 1;

	    str = gen_declarator (op, buf, name);

	    if (wrap)
	      {
		strcpy (tmpbuf, "("); strcat (tmpbuf, str); strcat (tmpbuf, ")");
		strcpy (str, tmpbuf);
	      }

	    adorn_decl (decl, str);
	    break;
	  }
	case ARRAY_TYPE: case FUNCTION_TYPE: case POINTER_TYPE:
	  {
	    str = strcpy (buf, name);

	    /* this clause is done iteratively...rather than recursively */
	    do
	      {
		op = is_complex_decl (TREE_TYPE (decl))
		  ? TREE_TYPE (decl)
		    : NULLT;

		adorn_decl (decl, str);

		/* we have a pointer to a function or array...(*)(), (*)[] */
		if ((code == POINTER_TYPE) &&
		    (op && (TREE_CODE (op) == FUNCTION_TYPE
			    || TREE_CODE (op) == ARRAY_TYPE)))
		  {
		    strcpy (tmpbuf, "("); strcat (tmpbuf, str); strcat (tmpbuf, ")");
		    strcpy (str, tmpbuf);
		  }

		decl = is_complex_decl (TREE_TYPE (decl))
		  ? TREE_TYPE (decl)
		    : NULLT;
	      }
	    while (decl && (code = TREE_CODE (decl)));

	    break;
	  }
	case IDENTIFIER_NODE:
	  /* will only happen if we are processing a "raw" expr-decl. */
	  return strcpy (buf, IDENTIFIER_POINTER (decl));
	}

      return str;
    }
  else				/* we have an abstract declarator or a _DECL node */
    {
      return strcpy (buf, name);
    }
}

static void
gen_declspecs (declspecs, buf, raw)
     tree declspecs;
     char *buf;
     int raw;
{
  if (raw)
    {
      tree chain;

      for (chain = nreverse (copy_list (declspecs)); chain; chain = TREE_CHAIN (chain))
	{
	  tree aspec = TREE_VALUE (chain);

	  if (TREE_CODE (aspec) == IDENTIFIER_NODE)
	    strcat (buf, IDENTIFIER_POINTER (aspec));
	  else if (TREE_CODE (aspec) == RECORD_TYPE)
	    {
	      if (TYPE_NAME (aspec))
		{
		  tree protocol_list = TYPE_PROTOCOL_LIST (aspec);
		  
		  if (!TREE_STATIC_TEMPLATE (aspec))
		    strcat (buf, "struct ");
		  strcat (buf, IDENTIFIER_POINTER (TYPE_NAME (aspec)));
/* NEW!!! */
		  if (protocol_list)
		    {
		      tree chain = protocol_list;
		      
		      strcat (buf, " <");
		      while (chain)
			{
			  strcat (buf, IDENTIFIER_POINTER (PROTOCOL_NAME (TREE_VALUE (chain))));
			  chain = TREE_CHAIN (chain);
			  if (chain)
			    strcat (buf, ", ");
			}
		      strcat (buf, ">");
		    }
		}
	      else
		strcat (buf, "untagged struct");
	    }
	  else if (TREE_CODE (aspec) == UNION_TYPE)
	    {
	      if (TYPE_NAME (aspec))
		{
		  strcat (buf, "union ");
		  strcat (buf, IDENTIFIER_POINTER (TYPE_NAME (aspec)));
		}
	      else
		strcat (buf, "untagged union");
	    }
	  else if (TREE_CODE (aspec) == ENUMERAL_TYPE)
	    {
	      if (TYPE_NAME (aspec))
		{
		  strcat (buf, "enum ");
		  strcat (buf, IDENTIFIER_POINTER (TYPE_NAME (aspec)));
		}
	      else
		strcat (buf, "untagged enum");
	    }
	  else if (TREE_CODE (aspec) == TYPE_DECL && DECL_NAME (aspec))
	    {
	      strcat (buf, IDENTIFIER_POINTER (DECL_NAME (aspec)));
	    }
/* NEW!!! */
	  else if (TREE_CODE (aspec) == POINTER_TYPE
		   && TYPE_MAIN_VARIANT (aspec) == id_type)
	    {
	      tree protocol_list = TYPE_PROTOCOL_LIST (aspec);
	      
	      strcat (buf, "id");
	      if (protocol_list)
		{
		  tree chain = protocol_list;
		  
		  strcat (buf, " <");
		  while (chain)
		    {
		      strcat (buf, IDENTIFIER_POINTER (PROTOCOL_NAME (TREE_VALUE (chain))));
		      chain = TREE_CHAIN (chain);
		      if (chain)
			strcat (buf, ", ");
		    }
		  strcat (buf, ">");
		}
	    }
	  if (TREE_CHAIN (chain))
	    strcat (buf, " ");
	}
    }
  else
    {
    /* type qualifiers */

    if (TREE_READONLY (declspecs))
      strcat (buf, "const ");
    if (TYPE_VOLATILE (declspecs))
      strcat (buf, "volatile ");

    switch (TREE_CODE (declspecs))
      {
	/* type specifiers */

      case INTEGER_TYPE:	/* signed integer types */

	declspecs = TYPE_MAIN_VARIANT (declspecs);

        if (declspecs == short_integer_type_node) /* 's' */
          strcat (buf, "short");
        else if (declspecs == integer_type_node) /* 'i' */
          strcat (buf, "int");
        else if (declspecs == long_integer_type_node) /* 'l' */
          strcat (buf, "long");
        else if (declspecs == long_long_integer_type_node)
          strcat (buf, "long long");
        else if (declspecs == signed_char_type_node || /* 'c' */
  	         declspecs == char_type_node)
          strcat (buf, "char");

        /* unsigned integer types */

        else if (declspecs == short_unsigned_type_node)	/* 'S' */
          strcat (buf, "unsigned short");
        else if (declspecs == unsigned_type_node) /* 'I' */
          strcat (buf, "unsigned int");
        else if (declspecs == long_unsigned_type_node) /* 'L' */
          strcat (buf, "unsigned long");
        else if (declspecs == long_long_unsigned_type_node) /* 'L' */
          strcat (buf, "unsigned long long");
        else if (declspecs == unsigned_char_type_node) /* 'C' */
          strcat (buf, "unsigned char");
	break;

      case REAL_TYPE:		/* floating point types */

	declspecs = TYPE_MAIN_VARIANT (declspecs);

        if (declspecs == float_type_node) /* 'f' */
          strcat (buf, "float");
        else if (declspecs == double_type_node)	/* 'd' */
          strcat (buf, "double");
	else if (declspecs == long_double_type_node) /* 'd' */
          strcat (buf, "long double");
	break;

      case RECORD_TYPE:
	if (TYPE_NAME (declspecs) &&
	    (TREE_CODE (TYPE_NAME (declspecs)) == IDENTIFIER_NODE))
	  {
	    tree protocol_list = TYPE_PROTOCOL_LIST (declspecs);

	    if (!TREE_STATIC_TEMPLATE (declspecs))
	      strcat (buf, "struct ");
	    strcat (buf, IDENTIFIER_POINTER (TYPE_NAME (declspecs)));
/* NEW!!! */
	    if (protocol_list)
	      {
		tree chain = protocol_list;
		
		strcat (buf, " <");
		while (chain)
		  {
		    strcat (buf, IDENTIFIER_POINTER (PROTOCOL_NAME (TREE_VALUE (chain))));
		    chain = TREE_CHAIN (chain);
		    if (chain)
		      strcat (buf, ", ");
		  }
		strcat (buf, ">");
	      }
	  }
	else
	  strcat (buf, "untagged struct");
	break;
      case UNION_TYPE:
	if (TYPE_NAME (declspecs) &&
	    (TREE_CODE (TYPE_NAME (declspecs)) == IDENTIFIER_NODE))
	  {
	    strcat (buf, "union ");
	    strcat (buf, IDENTIFIER_POINTER (TYPE_NAME (declspecs)));
	  }
	else
	  strcat (buf, "untagged union");
	break;
      case ENUMERAL_TYPE:
	if (TYPE_NAME (declspecs) &&
	    (TREE_CODE (TYPE_NAME (declspecs)) == IDENTIFIER_NODE))
	  {
	    strcat (buf, "enum ");
	    strcat (buf, IDENTIFIER_POINTER (TYPE_NAME (declspecs)));
	  }
	else
	  strcat (buf, "untagged enum");
	break;
      case VOID_TYPE:
	strcat (buf, "void");
	break;
/* NEW!!! */
      case POINTER_TYPE:
	{
	  tree protocol_list = TYPE_PROTOCOL_LIST (declspecs);
	  
	  strcat (buf, "id");
	  if (protocol_list)
	    {
	      tree chain = protocol_list;
	      
	      strcat (buf, " <");
	      while (chain)
		{
		  strcat (buf, IDENTIFIER_POINTER (PROTOCOL_NAME (TREE_VALUE (chain))));
		  chain = TREE_CHAIN (chain);
		  if (chain)
		    strcat (buf, ", ");
		}
	      strcat (buf, ">");
	    }
	}
      }
    }
}

static char *
gen_declaration (atype_or_adecl, buf)
     tree atype_or_adecl;
     char *buf;
{
  char declbuf[256];

  if (TREE_CODE (atype_or_adecl) == TREE_LIST)
    {
      tree declspecs;		/* "identifier_node", "record_type" */
      tree declarator;		/* "array_ref", "indirect_ref", "call_expr"... */

      /* we have a "raw", abstract delclarator (typename) */
      declarator = TREE_VALUE (atype_or_adecl);
      declspecs  = TREE_PURPOSE (atype_or_adecl);

      gen_declspecs (declspecs, buf, 1);
      if (declarator)
        {
	  strcat (buf, " ");
	  strcat (buf, gen_declarator (declarator, declbuf, ""));
	}
    }
  else
    {
      tree atype;
      tree declspecs;	/* "integer_type", "real_type", "record_type"... */
      tree declarator;	/* "array_type", "function_type", "pointer_type". */

      if (TREE_CODE (atype_or_adecl) == FIELD_DECL
	  || TREE_CODE (atype_or_adecl) == PARM_DECL
	  || TREE_CODE (atype_or_adecl) == FUNCTION_DECL)
	atype = TREE_TYPE (atype_or_adecl);
      else
	atype = atype_or_adecl;	/* assume we have a *_type node */

      if (is_complex_decl (atype))
	{
	  tree chain;

	  /* get the declaration specifier...it is at the end of the list */
	  declarator = chain = atype;
	  do
	    chain = TREE_TYPE (chain); /* not TREE_CHAIN (chain); */
	  while (is_complex_decl (chain));
	  declspecs = chain;
	}
      else
	{
	  declspecs = atype;
	  declarator = NULLT;
	}

      gen_declspecs (declspecs, buf, 0);

      if (TREE_CODE (atype_or_adecl) == FIELD_DECL
	  || TREE_CODE (atype_or_adecl) == PARM_DECL
	  || TREE_CODE (atype_or_adecl) == FUNCTION_DECL)
	{
	  const char *decl_name;
	  
	  if (DECL_NAME (atype_or_adecl))
	    decl_name = IDENTIFIER_POINTER (DECL_NAME (atype_or_adecl));
	  else
	    decl_name = "";

	  if (declarator)
	    {
	      strcat (buf, " ");
	      strcat (buf, gen_declarator (declarator, declbuf, decl_name));
	    }
	  else if (decl_name[0])
	    {
	      strcat (buf, " ");
	      strcat (buf, decl_name);
	    }
	}
      else if (declarator)
        {
	  strcat (buf, " ");
	  strcat (buf, gen_declarator (declarator, declbuf, ""));
	}
    }
  return buf;
}

#define RAW_TYPESPEC(meth) (TREE_VALUE (TREE_PURPOSE (TREE_TYPE (meth))))

static char *
gen_method_decl (method, buf)
     tree method;
     char *buf;
{
  tree chain;

  if (RAW_TYPESPEC (method) != objc_object_reference)
    {
      strcpy (buf, "(");
      gen_declaration (TREE_TYPE (method), buf);
      strcat (buf, ")");
    }

  chain = METHOD_SEL_ARGS (method);
  if (chain)
    {				/* we have a chain of keyword_decls */
      do
        {
	  if (KEYWORD_KEY_NAME (chain))
	    strcat (buf, IDENTIFIER_POINTER (KEYWORD_KEY_NAME (chain)));

	  strcat (buf, ":");
	  if (RAW_TYPESPEC (chain) != objc_object_reference)
	    {
	      strcat (buf, "(");
	      gen_declaration (TREE_TYPE (chain), buf);
	      strcat (buf, ")");
	    }
	  strcat (buf, IDENTIFIER_POINTER (KEYWORD_ARG_NAME (chain)));
	  if (chain = TREE_CHAIN (chain))
	    strcat (buf, " ");
        }
      while (chain);

      if (METHOD_ADD_ARGS (method) == (tree)1)
        strcat (buf, ", ...");
      else if (METHOD_ADD_ARGS (method))
        { 
	  /* We have a tree list node as generate by get_parm_info ().  */
	  chain  = TREE_PURPOSE (METHOD_ADD_ARGS (method));
          /* know we have a chain of parm_decls */
          while (chain)
            {
	      strcat (buf, ", ");
	      gen_declaration (chain, buf);
	      chain = TREE_CHAIN (chain);
            }
	}
    }
  else				/* we have a unary selector */
    {
      strcat (buf, IDENTIFIER_POINTER (METHOD_SEL_NAME (method)));
    }

  return buf;
}

void
gen_prototype (fp, decl)
     FILE *fp;
     tree decl;
{
  /* we have a function definition - generate prototype */
  bzero (errbuf, BUFSIZE);
  gen_declaration (decl, errbuf);
  fprintf (fp, "%s;\n", errbuf);
}
/*
 *  debug info...
 */
static void
dump_interface (fp, chain)
     FILE *fp;
     tree chain;
{
  char *buf = (char *)xmalloc (256);
  char *my_name = IDENTIFIER_POINTER (CLASS_NAME (chain));
  tree ivar_decls = CLASS_RAW_IVARS (chain);
  tree nst_methods = CLASS_NST_METHODS (chain);
  tree cls_methods = CLASS_CLS_METHODS (chain);

  fprintf (fp, "\n@interface %s", my_name);

  if (CLASS_SUPER_NAME (chain))
    {
      char *super_name = IDENTIFIER_POINTER (CLASS_SUPER_NAME (chain));
      fprintf (fp, " : %s\n", super_name);
    }
  else
    fprintf (fp, "\n");

  if (ivar_decls)
    {
      fprintf (fp, "{\n");
      do
	{
	  bzero (buf, 256);
	  fprintf (fp, "\t%s;\n", gen_declaration (ivar_decls, buf));
	  ivar_decls = TREE_CHAIN (ivar_decls);
	}
      while (ivar_decls);
      fprintf (fp, "}\n");
    }

  while (nst_methods)
    {
      bzero (buf, 256);
      fprintf (fp, "- %s;\n", gen_method_decl (nst_methods, buf));
      nst_methods = TREE_CHAIN (nst_methods);
    }

  while (cls_methods)
    {
      bzero (buf, 256);
      fprintf (fp, "+ %s;\n", gen_method_decl (cls_methods, buf));
      cls_methods = TREE_CHAIN (cls_methods);
    }
  fprintf (fp, "\n@end");
}

void
init_objc ()
{
  /* Add the special tree codes of Objective C to the tables.  */

  tree_code_type
    = (char **) realloc (tree_code_type,
			 sizeof (char *) * LAST_OBJC_TREE_CODE);
  tree_code_length
    = (int *) realloc (tree_code_length,
		       sizeof (int) * LAST_OBJC_TREE_CODE);
  tree_code_name
    = (char **) realloc (tree_code_name,
			 sizeof (char *) * LAST_OBJC_TREE_CODE);
  bcopy (objc_tree_code_type,
	 tree_code_type + (int) FIRST_OBJC_TREE_CODE,
	 (((int) LAST_OBJC_TREE_CODE - (int) FIRST_OBJC_TREE_CODE)
	  * sizeof (char *)));
  bcopy (objc_tree_code_length,
	 tree_code_length + (int) FIRST_OBJC_TREE_CODE,
	 (((int) LAST_OBJC_TREE_CODE - (int) FIRST_OBJC_TREE_CODE)
	  * sizeof (int)));
  bcopy (objc_tree_code_name,
	 tree_code_name + (int) FIRST_OBJC_TREE_CODE,
	 (((int) LAST_OBJC_TREE_CODE - (int) FIRST_OBJC_TREE_CODE)
	  * sizeof (char *)));

  errbuf = (char *)xmalloc (BUFSIZE);
  utlbuf = (char *)xmalloc (BUFSIZE);
  hash_init ();
  synth_module_prologue ();
}

void
finish_objc ()
{
  struct imp_entry *impent;
  tree chain;

  if (implementation_context || class_names_chain ||
      meth_var_names_chain || meth_var_types_chain || sel_ref_chain)
    generate_objc_symtab_decl ();

  for (impent = imp_list; impent; impent = impent->next)
    {
      implementation_context = impent->imp_context;
      implementation_template = impent->imp_template;

      _OBJC_CLASS_decl = impent->class_decl;
      _OBJC_METACLASS_decl = impent->meta_decl;

      if (TREE_CODE (implementation_context) == CLASS_IMPLEMENTATION)
	{
	  /* all of the following reference the string pool...  */
	  generate_ivar_lists ();
	  generate_dispatch_tables ();
	  generate_shared_structures ();
	}
      else
	{
	  generate_dispatch_tables ();
	  generate_category (implementation_context);
	}
    }

  if (sel_ref_chain)
    build_selector_translation_table ();

  if (protocol_chain) 
    generate_protocols ();

  if (implementation_context || class_names_chain ||
      meth_var_names_chain || meth_var_types_chain || sel_ref_chain)
    {
      build_module_descriptor ();
    }

  /* dump the string table last */

  generate_strings ();

  /* dump the class references...this forces the appropriate classes
     to be linked into the executable image, preserving unix archive
     semantics...this can be removed when we move to a more dynamically
     linked environment
     */
  for (chain = cls_ref_chain; chain; chain = TREE_CHAIN (chain))
    {
      sprintf (utlbuf, ".reference .objc_class_name_%s",
	       IDENTIFIER_POINTER (TREE_VALUE (chain)));
      assemble_asm (my_build_string (strlen (utlbuf) + 1, utlbuf));
    }
  for (impent = imp_list; impent; impent = impent->next)
    {
      implementation_context = impent->imp_context;
      implementation_template = impent->imp_template;

      if (TREE_CODE (impent->imp_context) == CLASS_IMPLEMENTATION)
	{
	  sprintf (utlbuf, ".objc_class_name_%s=0",
		   IDENTIFIER_POINTER (CLASS_NAME (impent->imp_context)));
	  assemble_asm (my_build_string (strlen (utlbuf) + 1, utlbuf));

	  sprintf (utlbuf, ".globl .objc_class_name_%s",
		   IDENTIFIER_POINTER (CLASS_NAME (impent->imp_context)));
	  assemble_asm (my_build_string (strlen (utlbuf) + 1, utlbuf));
	}
      else if (TREE_CODE (impent->imp_context) == CATEGORY_IMPLEMENTATION)
	{
	  /* Do the same for categories.  Even though no references to these
	      symbols are generated automatically by the compiler, it gives
	      you a handle to pull them into an archive by hand. */
	  sprintf (utlbuf, ".objc_category_name_%s_%s=0",
		   IDENTIFIER_POINTER (CLASS_NAME (impent->imp_context)),
		   IDENTIFIER_POINTER (CLASS_SUPER_NAME (impent->imp_context)));
	  assemble_asm (my_build_string (strlen (utlbuf) + 1, utlbuf));

	  sprintf (utlbuf, ".globl .objc_category_name_%s_%s",
		   IDENTIFIER_POINTER (CLASS_NAME (impent->imp_context)),
		   IDENTIFIER_POINTER (CLASS_SUPER_NAME (impent->imp_context)));
	  assemble_asm (my_build_string (strlen (utlbuf) + 1, utlbuf));
	}
    }
  /*** this fixes a gross bug in the assembler...it `expects' #APP to have
   *** a matching #NO_APP, or it crashes (sometimes). app_disable () will
   *** insure this is the case. 5/19/89, s.naroff.
   ***/
  if (cls_ref_chain || imp_list)
    app_disable ();

  if (flag_gen_declaration)
    {
      add_class (implementation_context);
      dump_interface (gen_declaration_file, implementation_context);
    }
  if (warn_selector)
    {
      int slot;
      
      /* Run through the selector hash tables and print a warning for any
         selector which has multiple methods. */
      
      for (slot = 0; slot < SIZEHASHTABLE; slot++)
        {
	  hash hsh;
	  
	  for (hsh = cls_method_hash_list[slot]; hsh; hsh = hsh->next)
	    {
	      if (hsh->list)
	        {
	          tree meth = hsh->key;
	          char type = (TREE_CODE (meth) == INSTANCE_METHOD_DECL)
			      ? '-' : '+';
	          attr loop;
	      
		  warning ("potential selector conflict for method `%s'",
			   IDENTIFIER_POINTER (METHOD_SEL_NAME (meth)));
	          warn_with_method ("found", type, meth);
	          for (loop = hsh->list; loop; loop = loop->next)
	            warn_with_method ("found", type, loop->value);
	        }
	    }
	}
    
      for (slot = 0; slot < SIZEHASHTABLE; slot++)
        {
	  hash hsh;
	  
	  for (hsh = nst_method_hash_list[slot]; hsh; hsh = hsh->next)
	    {
	      if (hsh->list)
	        {
	          tree meth = hsh->key;
	          char type = (TREE_CODE (meth) == INSTANCE_METHOD_DECL)
			      ? '-' : '+';
	          attr loop;
	      
		  warning ("potential selector conflict for method `%s'",
			   IDENTIFIER_POINTER (METHOD_SEL_NAME (meth)));
	          warn_with_method ("found", type, meth);
	          for (loop = hsh->list; loop; loop = loop->next)
	            warn_with_method ("found", type, loop->value);
	        }
	    }
	}
    }
}

#ifdef DEBUG

static void
objc_debug (fp)
     FILE *fp;
{
  char *buf = (char *)xmalloc (256);

  {				/* dump function prototypes */
    tree loop = _OBJC_MODULES_decl;

    fprintf (fp, "\n\nfunction prototypes:\n");
    while (loop)
      {
	if (TREE_CODE (loop) == FUNCTION_DECL && DECL_INITIAL (loop))
	  {
	    /* we have a function definition - generate prototype */
            bzero (errbuf, BUFSIZE);
	    gen_declaration (loop, errbuf);
	    fprintf (fp, "%s;\n", errbuf);
	  }
	loop = TREE_CHAIN (loop);
      }
  }
  {				/* dump global chains */
    tree loop;
    int i, index = 0, offset = 0;
    hash hashlist;

    for (i = 0; i < SIZEHASHTABLE; i++)
      {
	if (hashlist = nst_method_hash_list[i])
	  {
	    fprintf (fp, "\n\nnst_method_hash_list[%d]:\n", i);
	    do
	      {
		bzero (buf, 256);
		fprintf (fp, "-%s;\n", gen_method_decl (hashlist->key, buf));
		hashlist = hashlist->next;
	      }
	    while (hashlist);
	  }
      }
    for (i = 0; i < SIZEHASHTABLE; i++)
      {
	if (hashlist = cls_method_hash_list[i])
	  {
	    fprintf (fp, "\n\ncls_method_hash_list[%d]:\n", i);
	    do
	      {
		bzero (buf, 256);
		fprintf (fp, "-%s;\n", gen_method_decl (hashlist->key, buf));
		hashlist = hashlist->next;
	      }
	    while (hashlist);
	  }
      }
    fprintf (fp, "\nsel_refdef_chain:\n");
    for (loop = sel_refdef_chain; loop; loop = TREE_CHAIN (loop))
      {
	fprintf (fp, "(index: %4d offset: %4d) %s\n", index, offset,
		 IDENTIFIER_POINTER (TREE_VALUE (loop)));
	index++;
	/* add one for the '\0' character */
	offset += IDENTIFIER_LENGTH (TREE_VALUE (loop)) + 1;
      }
    fprintf (fp, "\n (max_selector_index: %4d.\n", max_selector_index);
  }
}
#endif

void
print_lang_statistics ()
{
}
